
#include "compat/compat.h"
#include "clock_m.h"
#include "clock_m_phc.h"
#include "include/syn1588_sys.h"
#include "syn1588nic_reg.h"

#include <linux/circ_buf.h>
#include <linux/bitmap.h>
#include <linux/math64.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/wait.h>


#ifndef MAX
#define MAX(X, Y) ((X) > (Y) ? (X) : (Y))
#endif
#ifndef MIN
#define MIN(X, Y) ((X) < (Y) ? (X) : (Y))
#endif

const u64 STEPS_PER_SECOND = 1000ull*1000*1000*(1u<<16);

struct io_matrix_source;
struct io_matrix_sink {
	enum syn1588_ioctl_sink_type type;
	struct io_matrix_source *connected_source;
	int reg_offset; //  offset in the iomatrix register

	struct io_matrix_sink *next; // used in connection handling
};
struct io_matrix_source {
	enum syn1588_ioctl_source_type type;
	struct io_matrix_sink *connected_sink_list_head;
	u32 reg_mask_gpio; // mask for the iomatrix register
	u32 reg_mask_event; // mask for the iomatrix register
};

// period interfacce

enum clock_m_period_res {
	CLOCK_M_PERIOD_RES_PERIOD0,
	CLOCK_M_PERIOD_RES_PERIOD1,
	CLOCK_M_PERIOD_RES_PERIOD2,
	CLOCK_M_PERIOD_RES_PERIOD3,
	__CLOCK_M_PERIOD_RES_NBITS
};
enum clock_m_res_event_res {
	CLOCK_M_RES_EVENT0,
	CLOCK_M_RES_EVENT1,
	__CLOCK_M_RES_EVENT_NBITS
};
enum clock_m_res_trigger_res {
	CLOCK_M_RES_TRIGGER0,
	CLOCK_M_RES_TRIGGER1,
	__CLOCK_M_RES_TRIGGER_NBITS
};

struct clock_m_period {
	struct clock_m_device *clock_m;
	int (*set_enable)(struct clock_m_period *, bool value);
	int (*setup)(struct clock_m_period *, struct syn1588_ioctl_duration high_time, struct syn1588_ioctl_duration low_time, bool is_rising_edge, enum syn1588_ioctl_period_start_condition start_condition);

	struct io_matrix_source source;
	u32 resources;
	struct syn1588_ioctl_duration current_high_time;
	struct syn1588_ioctl_duration current_low_time;
};

enum clock_m_event_state {
	EVENT_STATE_NONE = 0,
	EVENT_STATE_IDLE,
	EVENT_STATE_RUNNING,
};

static const size_t CLOCK_M_EVENT_BUFFER_SIZE = 128;

struct clock_m_event {
	struct clock_m_device *clock_m;
	int (*clear)(struct clock_m_event *);
	int (*setup)(struct clock_m_event *);

	syn1588_ioctl_timepoint (*read_ts)(struct clock_m_event *);

	struct io_matrix_sink sink;

	enum clock_m_event_state state;
	u32 resources;
	struct circ_buf timestamps_circ;
	wait_queue_head_t *wait_queue;
};

static const size_t CLOCK_M_TRIGGER_BUFFER_SIZE = 128;


struct circ_write_buf {
	void *buf;
	int head;
	int tail_written;
	int tail;
};

struct clock_m_trigger {
	struct clock_m_device *clock_m;
	int (*clear)(struct clock_m_trigger *);
	int (*setup)(struct clock_m_trigger *);

	void (*write_ts)(struct clock_m_trigger *, struct syn1588_ioctl_trigger_entry);

	struct io_matrix_source source;
	u32 fifo_size;

	u32 resources;
	struct circ_write_buf timestamps_circ;
};

struct clock_m_io_port {
	struct clock_m_device *clock_m;

	struct io_matrix_sink sink;
	struct io_matrix_source source;

	u32 resources;
};

struct syn1588_io_matrix {
	// connection matrix
	spinlock_t lock;
};
struct syn1588_io {
	struct clock_m_device *clock_m;

	struct clock_m_period periods[__SYN1588_IOCTL_PERIOD_NBITS];
	u32 period_resource_usage;
	spinlock_t period_lock;

	struct clock_m_event events[__SYN1588_IOCTL_EVENT_NBITS];
	u32 event_resource_usage;
	spinlock_t event_lock;

	struct clock_m_trigger triggers[__SYN1588_IOCTL_TRIGGER_NBITS];
	u32 trigger_resource_usage;
	spinlock_t trigger_lock;

	struct clock_m_io_port io_ports[__SYN1588_IOCTL_IO_PORT_NBITS];
	u32 io_port_resource_usage;
	spinlock_t io_port_lock;

	struct syn1588_io_matrix io_matrix;

};

static u64 mod_u64(u64 a, u64 b) {
	return do_div(a, b);
}

static u64 gcd_u64(u64 a, u64 b)
{
	u64 r;

	if (a < b)
		swap(a, b);
	while ((r = mod_u64(a, b)) != 0) {
		a = b;
		b = r;
	}
	return b;
}

static int convert_duration(const struct syn1588_ioctl_duration duration, u64 *result, struct syn1588_ioctl_duration *remainer) {
	if(duration.denominator == 0 || duration.numerator == 0) {
		return -EINVAL;
	} else {
		const u64 seconds = div64_u64(duration.numerator, duration.denominator);
		const u64 numerator_residual = duration.numerator - seconds * duration.denominator;
		const u64 sub_ns = mul_u64_u64_div_u64(STEPS_PER_SECOND, numerator_residual , duration.denominator);
		const u64 sub_ns_sum = sub_ns + STEPS_PER_SECOND*seconds;
		u64 gcd = 1;
		*result = sub_ns & 0xffffffffffffull;
		*result |= (seconds && 0xffff) << 16;

		if(remainer == NULL) {
			return 0;
		}

		/**
		 * see oregano - AN019
		 *
		 * P...excact period has `P.NUM` and `P.DEN`
		 * P = (n*Pa + m*Pb) / (n +m)
		 * Pa... sub_ns_sum
		 * Pb... Pa + 1`step`
		 * P = Pa * m / (n+m)
		 *
		 * remainer_num = P.NUM * STEPS_PER_SECOND - Pa * P.DEN
		 * g = gcd(remainer_num, P.DEN)
		 * n+m = P.DEN / g
		 * m = remainer_num / g
		 *
		 * remainer.numerator = n+m
		 * remainer.denominator = m
		 *
		 * n = (n+m) - m
		 *
		 * FRAC_NUM = n+m
		 * FRAC_DENNUM = n
		 *
		*/

		// minimize num / dem
		gcd =  gcd_u64(duration.numerator, duration.denominator);
		remainer->numerator = div64_u64(duration.numerator, gcd);
		remainer->denominator = div64_u64(duration.denominator, gcd);

		gcd =  gcd_u64(STEPS_PER_SECOND, sub_ns_sum);
		remainer->numerator = remainer->numerator * div64_u64(STEPS_PER_SECOND, gcd) - remainer->denominator * div64_u64(sub_ns_sum, gcd);
		if(remainer->numerator == 0) {
			return 0;
		}

		gcd =  gcd_u64(remainer->numerator, remainer->denominator);
		remainer->numerator = div64_u64(remainer->numerator, gcd);
		remainer->denominator = div64_u64(remainer->denominator, gcd);

		return 0;
	}
}

static int clock_m_period0_setup(
	struct clock_m_period *period,
	struct syn1588_ioctl_duration high_time,
	struct syn1588_ioctl_duration low_time,
	bool is_rising_edge,
	enum syn1588_ioctl_period_start_condition start_condition
) {
	struct clock_m_device *clock_m = period->clock_m;
	struct device *dev = clock_m->dev;
	u64 time_hw = 0;
	u32 half_duration_l, half_duration_h, eventctrl;
	int error;
	struct syn1588_ioctl_duration remain_ratio;
	if(high_time.numerator != low_time.numerator || high_time.denominator != low_time.denominator) {
		dev_err(dev, "no custom duty allowed\n");
		return -EOPNOTSUPP;
	}
	if(start_condition == SYN1588_IOCTL_PERIOD_START_1PPS) {
		dev_err(dev, "period can not start with the 1pps signal\n");
		return -EOPNOTSUPP;
	}
	error = convert_duration(high_time, &time_hw, &remain_ratio);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}

	//TODO check max time
	half_duration_l = time_hw;
	half_duration_h = time_hw >> 32;

	iowrite32(half_duration_l, clock_m->iobase + PERIODTIME0_L);
	iowrite32(half_duration_h, clock_m->iobase + PERIODTIME0_H);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	eventctrl &= ~(
		EVENTCTRL_PERIOD0_ENABLED |
		EVENTCTRL_PERIOD0_INITIAL_VALUE |
		EVENTCTRL_PERIOD0_OUTPUT_ENABLED |
		EVENTCTRL_PERIOD0_OUTPUT_VALUE |
		EVENTCTRL_PERIOD01_DUTY |
		EVENTCTRL_PERIOD0_START |
		(EVENTCTRL_PERIOD0_START << 1u)
	);
	switch(start_condition) {
	case SYN1588_IOCTL_PERIOD_START_ENABLE:
		eventctrl |= 0x00;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER0:
		eventctrl |= (EVENTCTRL_PERIOD0_START << 1) | EVENTCTRL_PERIOD0_START;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER1:
		eventctrl |= EVENTCTRL_PERIOD0_START;
		break;
	default:
		spin_unlock(&clock_m->reg_lock_eventctrl);
		return -EINVAL;
	}
	if(is_rising_edge) {
		eventctrl |= EVENTCTRL_PERIOD0_INITIAL_VALUE;
	} else {
		eventctrl |= EVENTCTRL_PERIOD0_OUTPUT_VALUE;
	}
	if(clock_m->clock_m_capabilities.precisionperiod) {
		u32 frac_num = remain_ratio.numerator;
		u32 frac_denum = remain_ratio.denominator - remain_ratio.denominator;
		struct syn1588_ioctl_duration duration = {
			numerator: time_hw* remain_ratio.denominator + remain_ratio.numerator*STEPS_PER_SECOND,
			denominator: STEPS_PER_SECOND*remain_ratio.denominator,
		};
		if(frac_num > 1 && frac_denum > 1 && frac_num < BIT(25) && frac_denum < BIT(24) ) {
			iowrite32(frac_num, clock_m->iobase + FRAC_NUM);
			iowrite32(frac_denum, clock_m->iobase + FRAC_DENUM);
			eventctrl |= GENLOCK_CTRL_PRECISION0ENBL;
			period->current_high_time = duration;
			period->current_low_time = duration;
		} else {
			struct syn1588_ioctl_duration duration = {
				numerator: time_hw,
				denominator: STEPS_PER_SECOND,
			};
			period->current_high_time = duration;
			period->current_low_time = duration;
		}
	} else {
		struct syn1588_ioctl_duration duration = {
			numerator: time_hw,
			denominator: STEPS_PER_SECOND,
		};
		period->current_high_time = duration;
		period->current_low_time = duration;
	}
	iowrite32(eventctrl, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static int clock_m_period0_set_enable(struct clock_m_period *period, bool enable)
{
	struct clock_m_device *clock_m = period->clock_m;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_PERIOD0_ENABLED | EVENTCTRL_PERIOD0_OUTPUT_ENABLED;

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	if(enable) {
		iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	} else {
		iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	}
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_period1_setup(struct clock_m_period *period, struct syn1588_ioctl_duration high_time, struct syn1588_ioctl_duration low_time, bool is_rising_edge, enum syn1588_ioctl_period_start_condition start_condition)
{
	struct clock_m_device *clock_m = period->clock_m;
	struct device *dev = clock_m->dev;
	u64 time_hw = 0;
	u32 half_duration_l, half_duration_h;
	struct syn1588_ioctl_duration remain_ratio;
	int error;
	u32 eventctrl;
	if(high_time.numerator != low_time.numerator || high_time.denominator != low_time.denominator) {
		dev_err(dev, "no custom duty allowed\n");
		return -EOPNOTSUPP;
	}
	if(start_condition == SYN1588_IOCTL_PERIOD_START_1PPS) {
		dev_err(dev, "period can not start with the 1pps signal\n");
		return -EOPNOTSUPP;
	}
	error = convert_duration(high_time, &time_hw, &remain_ratio);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}
	//TODO check max time
	half_duration_l = time_hw;
	half_duration_h = time_hw >> 32;

	iowrite32(half_duration_l, clock_m->iobase + PERIODTIME1_L);
	iowrite32(half_duration_h, clock_m->iobase + PERIODTIME1_H);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	eventctrl &= ~(
		EVENTCTRL_PERIOD1_ENABLED |
		EVENTCTRL_PERIOD1_INITIAL_VALUE |
		EVENTCTRL_PERIOD1_OUTPUT_ENABLED |
		EVENTCTRL_PERIOD1_OUTPUT_VALUE |
		EVENTCTRL_PERIOD1_START |
		(EVENTCTRL_PERIOD1_START << 1u)
	);
	switch(start_condition) {
	case SYN1588_IOCTL_PERIOD_START_ENABLE:
		eventctrl |= 0x00;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER0:
		eventctrl |= (EVENTCTRL_PERIOD1_START << 1) | EVENTCTRL_PERIOD1_START;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER1:
		eventctrl |= EVENTCTRL_PERIOD1_START;
		break;
	default:
		spin_unlock(&clock_m->reg_lock_eventctrl);
		return -EINVAL;
	}

	eventctrl |= EVENTCTRL_PERIOD1_OUTPUT_ENABLED;
	if(is_rising_edge) {
		eventctrl |= EVENTCTRL_PERIOD1_INITIAL_VALUE;
	} else {
		eventctrl |= EVENTCTRL_PERIOD1_OUTPUT_VALUE;
	}
	if(clock_m->clock_m_capabilities.precisionperiod) {
		struct syn1588_ioctl_duration duration = {
			numerator: time_hw* remain_ratio.denominator + remain_ratio.numerator*STEPS_PER_SECOND,
			denominator: STEPS_PER_SECOND*remain_ratio.denominator,
		};
		// period1 can only generate 27 / 13.5 / ...MHz
		// left over ratio is 17 / 27
		if(remain_ratio.numerator == 17 && remain_ratio.denominator == 27) {
			eventctrl |= GENLOCK_CTRL_PRECISION1ENBL;
			period->current_high_time = duration;
			period->current_low_time = duration;
		} else {
			struct syn1588_ioctl_duration duration = {
				numerator: time_hw,
				denominator: STEPS_PER_SECOND,
			};
			period->current_high_time = duration;
			period->current_low_time = duration;
		}

	} else {
		struct syn1588_ioctl_duration duration = {
			numerator: time_hw,
			denominator: STEPS_PER_SECOND,
		};
		period->current_high_time = duration;
		period->current_low_time = duration;
	}
	iowrite32(eventctrl, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_period1_set_enable(struct clock_m_period *period, bool enable)
{
	struct clock_m_device *clock_m = period->clock_m;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_PERIOD1_ENABLED | EVENTCTRL_PERIOD1_OUTPUT_ENABLED;

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	if(enable) {
		iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	} else {
		iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	}
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_period_duty0_setup(
	struct clock_m_period *period,
	struct syn1588_ioctl_duration high_time,
	struct syn1588_ioctl_duration low_time,
	bool is_rising_edge,
	enum syn1588_ioctl_period_start_condition start_condition
) {
	struct clock_m_device *clock_m = period->clock_m;
	struct device *dev = clock_m->dev;
	u64 high_time_hw = 0;
	u64 low_time_hw = 0;
	u32 low_duration_l, low_duration_h;
	u32 high_duration_l, high_duration_h;
	int error;
	u32 eventctrl;
	error = convert_duration(high_time, &high_time_hw, NULL);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}
	error = convert_duration(low_time, &low_time_hw, NULL);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}
	if(start_condition == SYN1588_IOCTL_PERIOD_START_1PPS) {
		dev_err(dev, "period can not start with the 1pps signal\n");
		return -EOPNOTSUPP;
	}

	//TODO check max time
	high_duration_l = high_time_hw;
	high_duration_h = high_time_hw >> 32;
	low_duration_l = low_time_hw;
	low_duration_h = low_time_hw >> 32;


	iowrite32(high_duration_l, clock_m->iobase + PERIODTIME0_L);
	iowrite32(high_duration_h, clock_m->iobase + PERIODTIME0_H);
	iowrite32(low_duration_l, clock_m->iobase + PERIODTIME1_L);
	iowrite32(low_duration_h, clock_m->iobase + PERIODTIME1_H);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	eventctrl &= ~(
		EVENTCTRL_PERIOD0_ENABLED |
		EVENTCTRL_PERIOD0_INITIAL_VALUE |
		EVENTCTRL_PERIOD0_OUTPUT_ENABLED |
		EVENTCTRL_PERIOD0_OUTPUT_VALUE |
		EVENTCTRL_PERIOD0_START |
		(EVENTCTRL_PERIOD0_START << 1u) |
		EVENTCTRL_PERIOD1_ENABLED |
		EVENTCTRL_PERIOD1_INITIAL_VALUE |
		EVENTCTRL_PERIOD1_OUTPUT_ENABLED |
		EVENTCTRL_PERIOD1_OUTPUT_VALUE |
		EVENTCTRL_PERIOD1_START |
		(EVENTCTRL_PERIOD1_START << 1u)
	);
	eventctrl |= EVENTCTRL_PERIOD01_DUTY;
	switch(start_condition) {
	case SYN1588_IOCTL_PERIOD_START_ENABLE:
		eventctrl |= 0x00;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER0:
		eventctrl |= (EVENTCTRL_PERIOD0_START << 1) | EVENTCTRL_PERIOD0_START;
		break;
	case SYN1588_IOCTL_PERIOD_START_TRIGGER1:
		eventctrl |= EVENTCTRL_PERIOD0_START;
		break;
	default:
		spin_unlock(&clock_m->reg_lock_eventctrl);
		return -EINVAL;
	}
	if(is_rising_edge) {
		eventctrl |= EVENTCTRL_PERIOD0_INITIAL_VALUE;
	} else {
		eventctrl |= EVENTCTRL_PERIOD0_OUTPUT_VALUE;
	}
	iowrite32(eventctrl, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	period->current_high_time.numerator = high_time_hw;
	period->current_high_time.denominator = STEPS_PER_SECOND;
	period->current_low_time.numerator = low_time_hw;
	period->current_low_time.denominator = STEPS_PER_SECOND;

	return 0;
}

static int clock_m_period_duty0_set_enable(struct clock_m_period *period, bool enable)
{
	struct clock_m_device *clock_m = period->clock_m;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_PERIOD0_ENABLED | EVENTCTRL_PERIOD0_OUTPUT_ENABLED;

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	if(enable) {
		iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	} else {
		iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	}
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_period2_setup(struct clock_m_period *period, struct syn1588_ioctl_duration high_time, struct syn1588_ioctl_duration low_time, bool is_rising_edge, enum syn1588_ioctl_period_start_condition start_condition)
{
	struct clock_m_device *clock_m = period->clock_m;
	struct device *dev = clock_m->dev;
	u64 time_hw = 0;
	u32 half_duration_l, half_duration_h;
	int error;
	if((high_time.numerator * 2) > high_time.denominator) {
		dev_err(dev, "not allowed longer period than 1s\n");
		return -EOPNOTSUPP;
	}
	if(high_time.numerator != low_time.numerator || high_time.denominator != low_time.denominator) {
		dev_err(dev, "no custom duty allowed\n");
		return -EOPNOTSUPP;
	}
	if(start_condition != SYN1588_IOCTL_PERIOD_START_1PPS) {
		dev_err(dev, "light period can only start with the 1pps signal\n");
		return -EOPNOTSUPP;
	}
	if(is_rising_edge == false) {
		return -EOPNOTSUPP;
	}
	error = convert_duration(high_time, &time_hw, NULL);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}
	half_duration_l = time_hw;
	half_duration_h = time_hw >> 32;

	iowrite32(half_duration_l, clock_m->iobase + PERIODTIME2_L);
	iowrite32(half_duration_h, clock_m->iobase + PERIODTIME2_H);

	period->current_high_time.numerator = time_hw;
	period->current_high_time.denominator = STEPS_PER_SECOND;
	period->current_low_time = period->current_high_time;

	return 0;
}

static int clock_m_period2_set_enable(struct clock_m_period *period, bool enable)
{
	struct clock_m_device *clock_m = period->clock_m;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_PERIOD2_ENABLED;

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	if(enable) {
		iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	} else {
		iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	}
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_period3_setup(
	struct clock_m_period *period,
	struct syn1588_ioctl_duration high_time,
	struct syn1588_ioctl_duration low_time,
	bool is_rising_edge,
	enum syn1588_ioctl_period_start_condition start_condition
)
{
	struct clock_m_device *clock_m = period->clock_m;
	struct device *dev = clock_m->dev;
	u64 time_hw = 0;
	u32 half_duration_l, half_duration_h;
	int error;
	if((high_time.numerator * 2) > high_time.denominator) {
		dev_err(dev, "not allowed longer period than 1s\n");
		return -EOPNOTSUPP;
	}
	if(high_time.numerator != low_time.numerator || high_time.denominator != low_time.denominator) {
		dev_err(dev, "no custom duty allowed\n");
		return -EOPNOTSUPP;
	}
	if(start_condition != SYN1588_IOCTL_PERIOD_START_1PPS) {
		dev_err(dev, "light period can only start with the 1pps signal\n");
		return -EOPNOTSUPP;
	}
	if(is_rising_edge == false) {
		return -EOPNOTSUPP;
	}
	error = convert_duration(high_time, &time_hw, NULL);
	if(error != 0) {
		dev_err(dev, "could not convert duration to hw time\n");
		return error;
	}
	half_duration_l = time_hw;
	half_duration_h = time_hw >> 32;

	iowrite32(half_duration_l, clock_m->iobase + PERIODTIME3_L);
	iowrite32(half_duration_h, clock_m->iobase + PERIODTIME3_H);

	period->current_high_time.numerator = time_hw;
	period->current_high_time.denominator = STEPS_PER_SECOND;
	period->current_low_time = period->current_high_time;

	return 0;
}

static int clock_m_period3_set_enable(struct clock_m_period *period, bool enable)
{
	struct clock_m_device *clock_m = period->clock_m;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_PERIOD3_ENABLED;

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	if(enable) {
		iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	} else {
		iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	}
	spin_unlock(&clock_m->reg_lock_eventctrl);
	return 0;
}

static int clock_m_event0_setup(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_EVENT0_ENABLED;
	int i;

	for(i = 0; i < 32; i ++) {
		//empty fifo
		syn1588_ioctl_timepoint tmp = event->read_ts(event);
		if(tmp.sec == 0 && tmp.nsec == 0) {
			break;
		}
		if(i >= 31)  {
			dev_err(clock_m->dev, "could not clear event0 fifo\n");
			return -EFAULT;
		}
	}

	iren = ioread32(clock_m->iobase + IREN);
	set_bit(CLOCK_M_IRSC_EVENT0OCCURED, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);


	return 0;
}

static int clock_m_event0_clear(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_EVENT0_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	clear_bit(CLOCK_M_IRSC_EVENT0OCCURED, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static syn1588_ioctl_timepoint clock_m_event0_read_ts(struct clock_m_event *event) {
	struct clock_m_device *clock_m = event->clock_m;
	u32 value_l = ioread32(clock_m->iobase + EVENTTIME0_L);
	u32 value_h = ioread32(clock_m->iobase + EVENTTIME0_H);

	syn1588_ioctl_timepoint res = {
		.sec = value_h,
		.nsec = value_l,
	};
	return res;
}

static int clock_m_event1_setup(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_EVENT1_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	set_bit(CLOCK_M_IRSC_EVENT1OCCURED, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static int clock_m_event1_clear(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_EVENT1_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	clear_bit(CLOCK_M_IRSC_EVENT1OCCURED, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static syn1588_ioctl_timepoint clock_m_event1_read_ts(struct clock_m_event *event) {
	struct clock_m_device *clock_m = event->clock_m;
	u32 value_l = ioread32(clock_m->iobase + EVENTTIME1_L);
	u32 value_h = ioread32(clock_m->iobase + EVENTTIME1_H);

	syn1588_ioctl_timepoint res = {
		.sec = value_h,
		.nsec = value_l,
	};
	return res;
}

static int clock_m_trigger0_setup(struct clock_m_trigger *trigger) {
	struct clock_m_device *clock_m = trigger->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_TRIGGER0_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	set_bit(CLOCK_M_IRSC_TRIGGER0MATCH, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static int clock_m_trigger0_clear(struct clock_m_trigger *trigger) {
	struct clock_m_device *clock_m = trigger->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_TRIGGER0_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	clear_bit(CLOCK_M_IRSC_TRIGGER0MATCH, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static void clock_m_trigger0_write_ts(struct clock_m_trigger *trigger, struct syn1588_ioctl_trigger_entry entry) {
	struct clock_m_device *clock_m = trigger->clock_m;
	u32 value_l = entry.ts.nsec;
	u32 value_h = entry.ts.sec & 0xFFFFF; // value in sec lower 20bits

	value_h |= ((u32)entry.action) << 20;

	iowrite32(value_l, clock_m->iobase + TRIGTIME0_L);
	iowrite32(value_h, clock_m->iobase + TRIGTIME0_H);
}

static int clock_m_trigger1_setup(struct clock_m_trigger *trigger) {
	struct clock_m_device *clock_m = trigger->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_TRIGGER1_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	set_bit(CLOCK_M_IRSC_TRIGGER1MATCH, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl | eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}
static int clock_m_trigger1_clear(struct clock_m_trigger *trigger) {
	struct clock_m_device *clock_m = trigger->clock_m;
	unsigned long iren;
	u32 eventctrl;
	u32 eventctrl_mask = EVENTCTRL_TRIGGER1_ENABLED;

	iren = ioread32(clock_m->iobase + IREN);
	clear_bit(CLOCK_M_IRSC_TRIGGER1MATCH, &iren);
	iowrite32(iren, clock_m->iobase + IREN);

	spin_lock(&clock_m->reg_lock_eventctrl);
	eventctrl = ioread32(clock_m->iobase + EVENTCTRL);
	iowrite32(eventctrl & ~eventctrl_mask, clock_m->iobase + EVENTCTRL);
	spin_unlock(&clock_m->reg_lock_eventctrl);

	return 0;
}

static void clock_m_trigger1_write_ts(struct clock_m_trigger *trigger, struct syn1588_ioctl_trigger_entry entry) {
	struct clock_m_device *clock_m = trigger->clock_m;
	u32 value_l = entry.ts.nsec;
	u32 value_h = entry.ts.sec & 0xFFFFF; // value in sec lower 20bits

	value_h |= ((u32)entry.action) << 20;

	iowrite32(value_l, clock_m->iobase + TRIGTIME1_L);
	iowrite32(value_h, clock_m->iobase + TRIGTIME1_H);
}

int clock_m_io_probe(struct clock_m_device *clock_m) {
	struct device *dev = clock_m->dev;
	struct syn1588_io *syn1588_io = devm_kzalloc(dev, sizeof(*syn1588_io), GFP_KERNEL);
	if(IS_ERR_OR_NULL(syn1588_io)) {
		return -ENOMEM;
	}

	// set iomatrix to a defined state with 1pps active on X4
	iowrite32(0x00000009, clock_m->iobase + IOMATRIX);

	if(clock_m->clock_m_capabilities.period0) {
		struct clock_m_period *period = &syn1588_io->periods[SYN1588_IOCTL_PERIOD0];
		period->clock_m = clock_m;
		period->set_enable = clock_m_period0_set_enable;
		period->setup = clock_m_period0_setup;
		period->resources = BIT(CLOCK_M_PERIOD_RES_PERIOD0);
		period->source.reg_mask_gpio = 0x5;
		period->source.reg_mask_event = 0x5;
		period->source.type = SYN1588_IOCTL_SOURCE_PERIOD;
	}

	if(clock_m->clock_m_capabilities.period1) {
		struct clock_m_period *period = &syn1588_io->periods[SYN1588_IOCTL_PERIOD1];
		period->clock_m = clock_m;
		period->set_enable = clock_m_period1_set_enable;
		period->setup = clock_m_period1_setup;
		period->resources = BIT(CLOCK_M_PERIOD_RES_PERIOD1);
		period->source.reg_mask_gpio = 0x6;
		period->source.reg_mask_event = 0x6;
		period->source.type = SYN1588_IOCTL_SOURCE_PERIOD;
	}

	if(clock_m->clock_m_capabilities.dutyperiod) {
		struct clock_m_period *period = &syn1588_io->periods[SYN1588_IOCTL_PERIOD_DUTY0];
		period->clock_m = clock_m;
		period->set_enable = clock_m_period_duty0_set_enable;
		period->setup = clock_m_period_duty0_setup;
		period->resources = BIT(CLOCK_M_PERIOD_RES_PERIOD0) |  BIT(CLOCK_M_PERIOD_RES_PERIOD1);
		period->source.reg_mask_gpio = 0x5;
		period->source.reg_mask_event = 0x5;
		period->source.type = SYN1588_IOCTL_SOURCE_PERIOD;
	}

	if(clock_m->clock_m_capabilities.period2) {
		struct clock_m_period *period = &syn1588_io->periods[SYN1588_IOCTL_PERIOD2];
		period->clock_m = clock_m;
		period->set_enable = clock_m_period2_set_enable;
		period->setup = clock_m_period2_setup;
		period->resources = BIT(CLOCK_M_PERIOD_RES_PERIOD2);
		period->source.reg_mask_gpio = 0x1;
		period->source.reg_mask_event = 0xc;
		period->source.type = SYN1588_IOCTL_SOURCE_PERIOD;
	}

	if(clock_m->clock_m_capabilities.period3) {
		struct clock_m_period *period = &syn1588_io->periods[SYN1588_IOCTL_PERIOD3];
		period->clock_m = clock_m;
		period->set_enable = clock_m_period3_set_enable;
		period->setup = clock_m_period3_setup;
		period->resources = BIT(CLOCK_M_PERIOD_RES_PERIOD3);
		period->source.reg_mask_gpio = 0x2;
		period->source.reg_mask_event = 0xd;
		period->source.type = SYN1588_IOCTL_SOURCE_PERIOD;
	}

	if(clock_m->clock_m_capabilities.event0) {
		struct clock_m_event *event = &syn1588_io->events[SYN1588_IOCTL_EVENT0];
		event->clock_m = clock_m;
		event->setup = clock_m_event0_setup;
		event->clear = clock_m_event0_clear;
		event->read_ts = clock_m_event0_read_ts;
		event->resources = BIT(CLOCK_M_RES_EVENT0);
		event->state = EVENT_STATE_IDLE;
		event->sink.reg_offset = 16;
		event->sink.type = SYN1588_IOCTL_SINK_EVENT;
	}

	if(clock_m->clock_m_capabilities.event1) {
		struct clock_m_event *event = &syn1588_io->events[SYN1588_IOCTL_EVENT1];
		event->clock_m = clock_m;
		event->setup = clock_m_event1_setup;
		event->clear = clock_m_event1_clear;
		event->read_ts = clock_m_event1_read_ts;
		event->resources = BIT(CLOCK_M_RES_EVENT1);
		event->state = EVENT_STATE_IDLE;
		event->sink.reg_offset = 20;
		event->sink.type = SYN1588_IOCTL_SINK_EVENT;
	}

	if(clock_m->clock_m_capabilities.trigger0) {
		struct clock_m_trigger *trigger = &syn1588_io->triggers[SYN1588_IOCTL_TRIGGER0];
		trigger->clock_m = clock_m;
		trigger->setup = clock_m_trigger0_setup;
		trigger->clear = clock_m_trigger0_clear;
		trigger->write_ts = clock_m_trigger0_write_ts;
		trigger->fifo_size = 16;
		trigger->resources = BIT(CLOCK_M_RES_TRIGGER0);
		trigger->source.reg_mask_gpio = 0x7;
		trigger->source.reg_mask_event = 0x7;
		trigger->source.type = SYN1588_IOCTL_SOURCE_TRIGGER;
	}

	if(clock_m->clock_m_capabilities.trigger1) {
		struct clock_m_trigger *trigger = &syn1588_io->triggers[SYN1588_IOCTL_TRIGGER1];
		trigger->clock_m = clock_m;
		trigger->setup = clock_m_trigger1_setup;
		trigger->clear = clock_m_trigger1_clear;
		trigger->write_ts = clock_m_trigger1_write_ts;
		trigger->fifo_size = 1;
		trigger->resources = BIT(CLOCK_M_RES_TRIGGER1);
		trigger->source.reg_mask_gpio = 0x8;
		trigger->source.reg_mask_event = 0x8;
		trigger->source.type = SYN1588_IOCTL_SOURCE_TRIGGER;
	}

	if(true) {
		struct clock_m_io_port *io_port = &syn1588_io->io_ports[SYN1588_IOCTL_IO_PORT_X4];
		io_port->resources = BIT(SYN1588_IOCTL_IO_PORT_X4);
		io_port->source.reg_mask_event = 0x1;
		io_port->sink.reg_offset = 0;
		io_port->sink.type = SYN1588_IOCTL_SINK_IO_PORT;
		io_port->source.type = SYN1588_IOCTL_SOURCE_IO_PORT;
	}

	if(true) {
		struct clock_m_io_port *io_port = &syn1588_io->io_ports[SYN1588_IOCTL_IO_PORT_X5];
		io_port->resources = BIT(SYN1588_IOCTL_IO_PORT_X5);
		io_port->source.reg_mask_event = 0x2;
		io_port->sink.reg_offset = 12;
		io_port->sink.type = SYN1588_IOCTL_SINK_IO_PORT;
		io_port->source.type = SYN1588_IOCTL_SOURCE_IO_PORT;
	}

	if(true) {
		struct clock_m_io_port *io_port = &syn1588_io->io_ports[SYN1588_IOCTL_IO_PORT_X6];
		io_port->resources = BIT(SYN1588_IOCTL_IO_PORT_X6);
		io_port->source.reg_mask_event = 0x3;
		io_port->sink.reg_offset = 8;
		io_port->sink.type = SYN1588_IOCTL_SINK_IO_PORT;
		io_port->source.type = SYN1588_IOCTL_SOURCE_IO_PORT;
	}

	if(true) {
		struct clock_m_io_port *io_port = &syn1588_io->io_ports[SYN1588_IOCTL_IO_PORT_X7];
		io_port->resources = BIT(SYN1588_IOCTL_IO_PORT_X7);
		io_port->source.reg_mask_event = 0x4;
		io_port->sink.reg_offset = 4;
		io_port->sink.type = SYN1588_IOCTL_SINK_IO_PORT;
		io_port->source.type = SYN1588_IOCTL_SOURCE_IO_PORT;
	}


	clock_m->syn1588_io = syn1588_io;

	return 0;
}


static int alloc_period_res(struct clock_m_period *period)
{
	struct clock_m_device *clock_m = period->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	spin_lock(&syn1588_io->period_lock);
	if(syn1588_io->period_resource_usage & period->resources) {
		spin_unlock(&syn1588_io->period_lock);
		return -EBUSY;
	}
	syn1588_io->period_resource_usage |= period->resources;
	spin_unlock(&syn1588_io->period_lock);

	return 0;
}
static int free_period_res(struct clock_m_period *period)
{
	struct clock_m_device *clock_m = period->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;

	spin_lock(&syn1588_io->period_lock);
	syn1588_io->period_resource_usage &= ~period->resources;
	spin_unlock(&syn1588_io->period_lock);

	return 0;
}

int clock_m_period_setup (
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_period_id period_id,
	struct syn1588_ioctl_duration high_time,
	struct syn1588_ioctl_duration low_time,
	bool is_rising_edge,
	enum syn1588_ioctl_period_start_condition start_condition
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_period *period;
	int error = 0;

	if(period_id < 0 || period_id >= __SYN1588_IOCTL_PERIOD_NBITS) {
		dev_err(dev, "invalid period_id\n");
		return -EINVAL;
	}
	period = &syn1588_io->periods[period_id];

	if(period->resources == 0) {
		dev_err(dev, "period is not existing\n");
		return -ENODEV;
	}

	error = alloc_period_res(period);
	if(error) {
		dev_err(dev, "could not claim period resources\n");
		return error;
	}

	error = period->setup(period, high_time, low_time, is_rising_edge, start_condition);
	if (error) {
		free_period_res(period);
		return error;
	}

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_period_setup);

int clock_m_period_clear(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_period_id period_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_period *period;

	if(period_id < 0 || period_id >= __SYN1588_IOCTL_PERIOD_NBITS) {
		dev_err(dev, "invalid period_id\n");
		return -EINVAL;
	}
	period = &syn1588_io->periods[period_id];

	if(period->resources == 0) {
		dev_err(dev, "period is not existing\n");
		return -ENODEV;
	}

	(void)period->set_enable(period,false);
	free_period_res(period);

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_period_clear);

int clock_m_period_set_enable(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_period_id period_id,
	bool value
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_period *period;

	if(period_id < 0 || period_id >= __SYN1588_IOCTL_PERIOD_NBITS) {
		dev_err(dev, "invalid period_id\n");
		return -EINVAL;
	}
	period = &syn1588_io->periods[period_id];

	if(period->resources == 0) {
		dev_err(dev, "period is not existing\n");
		return -ENODEV;
	}

	return period->set_enable(period,value);
}
EXPORT_SYMBOL_GPL(clock_m_period_set_enable);

int clock_m_period_get_durations(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_period_id period_id,
	struct syn1588_ioctl_duration *high_time,
	struct syn1588_ioctl_duration *low_time
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_period *period;

	if(period_id < 0 || period_id >= __SYN1588_IOCTL_PERIOD_NBITS) {
		dev_err(dev, "invalid period_id\n");
		return -EINVAL;
	}
	period = &syn1588_io->periods[period_id];

	if(period->resources == 0) {
		dev_err(dev, "period is not existing\n");
		return -ENODEV;
	}

	*high_time = period->current_high_time;
	*low_time = period->current_low_time;

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_period_get_durations);

static int alloc_event_res(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	spin_lock(&syn1588_io->event_lock);
	if(syn1588_io->event_resource_usage & event->resources) {
		spin_unlock(&syn1588_io->event_lock);
		return -EBUSY;
	}
	syn1588_io->event_resource_usage |= event->resources;
	spin_unlock(&syn1588_io->event_lock);

	return 0;
}
static int free_event_res(struct clock_m_event *event)
{
	struct clock_m_device *clock_m = event->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;

	spin_lock(&syn1588_io->event_lock);
	syn1588_io->event_resource_usage &= ~event->resources;
	spin_unlock(&syn1588_io->event_lock);

	return 0;
}


int clock_m_event_setup (
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id,
	wait_queue_head_t *wait_queue
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event;
	int error = 0;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->resources == 0) {
		dev_err(dev, "event is not existing\n");
		return -ENODEV;
	}

	error = alloc_event_res(event);
	if(error) {
		dev_err(dev, "could not claim event resources\n");
		return error;
	}
	event->timestamps_circ.head = 0;
	event->timestamps_circ.tail = 0;
	event->timestamps_circ.buf = devm_kmalloc(dev, sizeof(syn1588_ioctl_timepoint)*CLOCK_M_EVENT_BUFFER_SIZE, GFP_KERNEL);
	if(IS_ERR_OR_NULL(event->timestamps_circ.buf)) {
		free_event_res(event);
		return PTR_ERR(event->timestamps_circ.buf);
	}
	error = event->setup(event);
	if (error) {
		free_event_res(event);
		return error;
	}
	event->wait_queue = wait_queue;
	event->state = EVENT_STATE_RUNNING;

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_event_setup);

int clock_m_event_free(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event;
	int error;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->resources == 0) {
		dev_err(dev, "event is not existing\n");
		return -ENODEV;
	}

	if( (syn1588_io->event_resource_usage & event->resources) != event->resources) {
		dev_err(dev, "event is not set up, so can not be cleard\n");
		return -EINVAL;
	}

	//disable
	event->state = EVENT_STATE_IDLE;
	error = event->clear(event);
	if(error) {
		return error;
	}
	event->wait_queue = NULL;
	devm_kfree(dev, event->timestamps_circ.buf);

	free_event_res(event);

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_event_free);

int clock_m_event_ts_available(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event;
	struct circ_buf *timestamps_circ = NULL;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->state != EVENT_STATE_RUNNING) {
		dev_err(dev, "event is not in running state\n");
		return -ENODEV;
	}

	timestamps_circ = &event->timestamps_circ;
	return CIRC_CNT(timestamps_circ->head, timestamps_circ->tail, CLOCK_M_EVENT_BUFFER_SIZE);
}
EXPORT_SYMBOL_GPL(clock_m_event_ts_available);

int clock_m_event_get_ts(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id,
	syn1588_ioctl_timepoint *buffer
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event;
	struct circ_buf *timestamps_circ = NULL;
	syn1588_ioctl_timepoint *timestamps = NULL;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->state != EVENT_STATE_RUNNING) {
		dev_err(dev, "event is not in running state\n");
		return -ENODEV;
	}
	timestamps_circ = &event->timestamps_circ;
	timestamps = (syn1588_ioctl_timepoint *)timestamps_circ->buf;

	if(timestamps_circ->head == timestamps_circ->tail) {
		return -EAGAIN;
	}
	*buffer = timestamps[timestamps_circ->tail];
	timestamps_circ->tail = (timestamps_circ->tail+1) % CLOCK_M_EVENT_BUFFER_SIZE;

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_event_get_ts);

int clock_m_event_read_ts(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event = NULL;
	struct circ_buf *timestamps_circ = NULL;
	syn1588_ioctl_timepoint *timestamps = NULL;
	bool buffer_was_empty = false;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->state != EVENT_STATE_RUNNING) {
		dev_dbg(dev, "new event function is not running\n");
		return -EINVAL;
	}
	timestamps_circ = &event->timestamps_circ;
	timestamps = (syn1588_ioctl_timepoint *)timestamps_circ->buf;

	if(timestamps_circ->head==timestamps_circ->tail) {
		buffer_was_empty = true;
	}

	while(true) {
		syn1588_ioctl_timepoint ts = event->read_ts(event);
		if(ts.sec == 0 && ts.nsec == 0) {
			break;
		}
		if(timestamps_circ->head!=timestamps_circ->tail) {
			int index_last = (timestamps_circ->head -1 + CLOCK_M_EVENT_BUFFER_SIZE) % CLOCK_M_EVENT_BUFFER_SIZE;
			syn1588_ioctl_timepoint ts_old = timestamps[index_last];
			if(ts_old.sec == ts.sec && ts_old.nsec == ts.nsec) {
				break;
			}
		}
		if(CIRC_SPACE(timestamps_circ->head, timestamps_circ->tail, CLOCK_M_EVENT_BUFFER_SIZE) >= 1) {
			timestamps[timestamps_circ->head] = ts;

			timestamps_circ->head = (timestamps_circ->head + 1) % CLOCK_M_EVENT_BUFFER_SIZE;
		}else {
			return -ENOMEM;
		}
	}

	if(buffer_was_empty && timestamps_circ->head!=timestamps_circ->tail) {
		// if buffer was empty trigger new irq
		if(event->wait_queue != NULL) {
			wake_up(event->wait_queue);
		} else {
			dev_warn(dev, "invalid wait_queue pointer\n");
		}
	}
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_event_read_ts);

int clock_m_event_clear(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_event_id event_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_event *event;
	struct circ_buf *timestamps_circ = NULL;

	if(event_id < 0 || event_id >= __SYN1588_IOCTL_EVENT_NBITS) {
		dev_err(dev, "invalid event_id\n");
		return -EINVAL;
	}
	event = &syn1588_io->events[event_id];

	if(event->state != EVENT_STATE_RUNNING) {
		dev_err(dev, "event is not in running state\n");
		return -ENODEV;
	}
	timestamps_circ = &event->timestamps_circ;

	// set tail to head to clear the buffer
	timestamps_circ->tail = timestamps_circ->head;

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_event_clear);

static int alloc_trigger_res(struct clock_m_trigger *trigger)
{
	struct clock_m_device *clock_m = trigger->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;

	spin_lock(&syn1588_io->trigger_lock);
	if(syn1588_io->trigger_resource_usage & trigger->resources) {
		spin_unlock(&syn1588_io->trigger_lock);
		return -EBUSY;
	}
	syn1588_io->trigger_resource_usage |= trigger->resources;
	spin_unlock(&syn1588_io->trigger_lock);

	return 0;
}
static int free_trigger_res(struct clock_m_trigger *trigger)
{
	struct clock_m_device *clock_m = trigger->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;

	spin_lock(&syn1588_io->trigger_lock);
	syn1588_io->trigger_resource_usage &= ~trigger->resources;
	spin_unlock(&syn1588_io->trigger_lock);

	return 0;
}

int clock_m_trigger_setup (
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_trigger_id trigger_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_trigger *trigger;
	int error = 0;

	if(trigger_id < 0 || trigger_id >= __SYN1588_IOCTL_TRIGGER_NBITS) {
		dev_err(dev, "invalid trigger_id\n");
		return -EINVAL;
	}
	trigger = &syn1588_io->triggers[trigger_id];

	if(trigger->resources == 0) {
		dev_err(dev, "trigger is not existing\n");
		return -ENODEV;
	}

	error = alloc_trigger_res(trigger);
	if(error) {
		dev_err(dev, "could not claim trigger resources\n");
		return error;
	}

	error = trigger->setup(trigger);
	if (error) {
		free_trigger_res(trigger);
		return error;
	}
	trigger->timestamps_circ.head = 0;
	trigger->timestamps_circ.tail_written = 0;
	trigger->timestamps_circ.tail = 0;
	trigger->timestamps_circ.buf = devm_kmalloc(dev, sizeof(struct syn1588_ioctl_trigger_entry)*CLOCK_M_TRIGGER_BUFFER_SIZE, GFP_KERNEL);

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_trigger_setup);

int clock_m_trigger_clear(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_trigger_id trigger_id
){
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_trigger *trigger;

	if(trigger_id < 0 || trigger_id >= __SYN1588_IOCTL_TRIGGER_NBITS) {
		dev_err(dev, "invalid trigger_id\n");
		return -EINVAL;
	}
	trigger = &syn1588_io->triggers[trigger_id];

	if(trigger->resources == 0) {
		dev_err(dev, "trigger is not existing\n");
		return -ENODEV;
	}

	trigger->clear(trigger);
	devm_kfree(dev, trigger->timestamps_circ.buf);

	free_trigger_res(trigger);

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_trigger_clear);

static void clock_m_trigger_cleanup_buffer(
	struct clock_m_trigger *trigger
) {
	struct clock_m_device *clock_m = trigger->clock_m;
	struct circ_write_buf *timestamps_circ = &trigger->timestamps_circ;
	struct syn1588_ioctl_trigger_entry *timestamps = timestamps_circ->buf;
	struct timespec64 current_hw_time;

	//TODO reread out the TRIGGERTIMEX register
	clock_m_gettime(clock_m, &current_hw_time);

	while(timestamps_circ->tail_written!=timestamps_circ->tail) {
		struct syn1588_ioctl_trigger_entry entry = timestamps[timestamps_circ->tail];
		if(entry.ts.sec > current_hw_time.tv_sec || entry.ts.nsec > current_hw_time.tv_nsec) {
			return;
		}
		timestamps_circ->tail = (timestamps_circ->tail+1) % CLOCK_M_TRIGGER_BUFFER_SIZE;
	}
}

/**
 * add new trigger points to the ring buffer
 * the list must be sorted ascending
*/
int clock_m_trigger_add_action(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_trigger_id trigger_id,
	struct syn1588_ioctl_trigger_entry *buffer
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_trigger *trigger = NULL;
	struct circ_write_buf *timestamps_circ = NULL;
	struct syn1588_ioctl_trigger_entry *timestamps = NULL;
	bool buffer_was_empty = false;

	if(trigger_id < 0 || trigger_id >= __SYN1588_IOCTL_TRIGGER_NBITS) {
		dev_err(dev, "invalid trigger_id\n");
		return -EINVAL;
	}
	trigger = &syn1588_io->triggers[trigger_id];
	clock_m_trigger_cleanup_buffer(trigger);

	timestamps_circ = &trigger->timestamps_circ;
	timestamps = timestamps_circ->buf;

	if(CIRC_SPACE(timestamps_circ->head, timestamps_circ->tail, CLOCK_M_TRIGGER_BUFFER_SIZE) == 0) {
		pr_err("trigger busy\n");
		return -EBUSY;
	}

	if(timestamps_circ->head==timestamps_circ->tail) {
		buffer_was_empty = true;
	}

	timestamps[timestamps_circ->head] = *buffer;
	timestamps_circ->head = (timestamps_circ->head+1) % CLOCK_M_TRIGGER_BUFFER_SIZE;


	if(buffer_was_empty) {
		clock_m_trigger_write_ts(clock_m, trigger_id);
	}
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_trigger_add_action);

int clock_m_trigger_write_ts(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_trigger_id trigger_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_trigger *trigger = NULL;
	struct circ_write_buf *timestamps_circ = NULL;
	struct syn1588_ioctl_trigger_entry *timestamps = NULL;

	if(trigger_id < 0 || trigger_id >= __SYN1588_IOCTL_TRIGGER_NBITS) {
		dev_err(dev, "invalid trigger_id\n");
		return -EINVAL;
	}
	trigger = &syn1588_io->triggers[trigger_id];
	clock_m_trigger_cleanup_buffer(trigger);

	timestamps_circ = &trigger->timestamps_circ;
	timestamps = timestamps_circ->buf;

	while(timestamps_circ->head!=timestamps_circ->tail_written) {
		struct syn1588_ioctl_trigger_entry ts = timestamps[timestamps_circ->tail_written];
		if(CIRC_CNT(timestamps_circ->tail_written, timestamps_circ->tail, CLOCK_M_TRIGGER_BUFFER_SIZE)>=trigger->fifo_size)
		{
			return -EBUSY;
		}
		timestamps_circ->tail_written = (timestamps_circ->tail_written+1) % CLOCK_M_TRIGGER_BUFFER_SIZE;
		trigger->write_ts(trigger, ts);
	}
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_trigger_write_ts);


static int alloc_io_port_res(struct clock_m_io_port *io_port)
{
	struct clock_m_device *clock_m = io_port->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	spin_lock(&syn1588_io->io_port_lock);
	if(syn1588_io->io_port_resource_usage & io_port->resources) {
		spin_unlock(&syn1588_io->io_port_lock);
		return -EBUSY;
	}
	syn1588_io->io_port_resource_usage |= io_port->resources;
	spin_unlock(&syn1588_io->io_port_lock);

	return 0;
}
static int free_io_port_res(struct clock_m_io_port *io_port)
{
	struct clock_m_device *clock_m = io_port->clock_m;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;

	spin_lock(&syn1588_io->io_port_lock);
	syn1588_io->io_port_resource_usage &= ~io_port->resources;
	spin_unlock(&syn1588_io->io_port_lock);

	return 0;
}

int clock_m_io_port_setup (
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_io_port_id io_port_id
) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_io_port *io_port;
	int error = 0;

	if(io_port_id < 0 || io_port_id >= __SYN1588_IOCTL_IO_PORT_NBITS) {
		dev_err(dev, "invalid io_port_id\n");
		return -EINVAL;
	}
	io_port = &syn1588_io->io_ports[io_port_id];

	if(io_port->resources == 0) {
		dev_err(dev, "io_port is not existing\n");
		return -ENODEV;
	}

	io_port->clock_m = clock_m;

	error = alloc_io_port_res(io_port);
	if(error) {
		dev_err(dev, "could not claim io_port resources\n");
		return error;
	}


	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_port_setup);

int clock_m_io_port_clear(
	struct clock_m_device *clock_m,
	enum syn1588_ioctl_io_port_id io_port_id
){
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct clock_m_io_port *io_port;

	if(io_port_id < 0 || io_port_id >= __SYN1588_IOCTL_IO_PORT_NBITS) {
		dev_err(dev, "invalid io_port_id\n");
		return -EINVAL;
	}
	io_port = &syn1588_io->io_ports[io_port_id];

	if(io_port->resources == 0) {
		dev_err(dev, "io_port is not existing\n");
		return -ENODEV;
	}

	free_io_port_res(io_port);

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_port_clear);


static bool valid_source(const struct syn1588_ioctl_source *source) {
	if(source->type < 0 || source->type >= __SYN1588_IOCTL_SOURCE_NBITS) {
		return false;
	}
	switch(source->type) {
	case SYN1588_IOCTL_SOURCE_PERIOD:
		return source->id >=0 && source->id < __SYN1588_IOCTL_PERIOD_NBITS;
	case SYN1588_IOCTL_SOURCE_TRIGGER:
		return source->id >=0 && source->id < __SYN1588_IOCTL_TRIGGER_NBITS;
	case SYN1588_IOCTL_SOURCE_IO_PORT:
		return source->id >=0 && source->id < __SYN1588_IOCTL_IO_PORT_NBITS;
	case SYN1588_IOCTL_SOURCE_IRIG_B:
	case SYN1588_IOCTL_SOURCE_1PPS:
	case SYN1588_IOCTL_SOURCE_PLL_FILTERED:
		return false; //TODO
	default:
		return false;
	}
}
static bool valid_sink(const struct syn1588_ioctl_sink *sink) {
	if(sink->type < 0 || sink->type >= __SYN1588_IOCTL_SINK_NBITS) {
		return false;
	}
	switch(sink->type) {
	case SYN1588_IOCTL_SINK_EVENT:
		return sink->id >=0 && sink->id < __SYN1588_IOCTL_EVENT_NBITS;
	case SYN1588_IOCTL_SINK_IO_PORT:
		return sink->id >=0 && sink->id < __SYN1588_IOCTL_IO_PORT_NBITS;
	case SYN1588_IOCTL_SINK_IRIG_B:
		return false; //TODO
	default:
		return false;
	}
}

static struct io_matrix_source * to_internal_source(struct clock_m_device *clock_m, const struct syn1588_ioctl_source *source_ioctl) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct io_matrix_source *source = NULL;

	if(!valid_source(source_ioctl)) {
		dev_err(dev, "invalid source\n");
		return ERR_PTR(-ENODEV);
	}

	if(source_ioctl->type == SYN1588_IOCTL_SOURCE_PERIOD) {
		struct clock_m_period *period;
		period = &syn1588_io->periods[source_ioctl->id];
		if(period->resources == 0) {
			//device not exsist
			return ERR_PTR(-ENODEV);
		}
		source = &period->source;
	}else if(source_ioctl->type == SYN1588_IOCTL_SOURCE_TRIGGER) {
		struct clock_m_trigger *trigger;
		trigger = &syn1588_io->triggers[source_ioctl->id];
		if(trigger->resources == 0) {
			//device not exsist
			return ERR_PTR(-ENODEV);
		}
		source = &trigger->source;
	}else if(source_ioctl->type == SYN1588_IOCTL_SOURCE_IO_PORT) {
		struct clock_m_io_port *io_port;
		if(source_ioctl->id < 0 || source_ioctl->id >= __SYN1588_IOCTL_IO_PORT_NBITS) {
			dev_err(dev, "invalid io_port_id\n");
			return ERR_PTR(-EINVAL);
		}
		io_port = &syn1588_io->io_ports[source_ioctl->id];
		// wenn using io port as source it can not be a sink
		if(io_port->sink.connected_source != NULL) {
			return ERR_PTR(-EBUSY);
		}
		if(io_port->resources == 0) {
			//device not exsist
			return ERR_PTR(-ENODEV);
		}
		source = &io_port->source;
	}else if(source_ioctl->type == SYN1588_IOCTL_SOURCE_IRIG_B) {
		//TODO
		return ERR_PTR(-ENODEV);
	}else  if(source_ioctl->type == SYN1588_IOCTL_SOURCE_1PPS) {
		return ERR_PTR(-ENODEV);
	}else if(source_ioctl->type == SYN1588_IOCTL_SOURCE_PLL_FILTERED) {
		return ERR_PTR(-ENODEV);
	} else {
		return ERR_PTR(-ENODEV);
	}
	return source;
}

static struct io_matrix_sink *to_internal_sink(struct clock_m_device *clock_m, const struct syn1588_ioctl_sink *sink_ioctl) {
	struct device *dev = clock_m->dev;
	struct syn1588_io * syn1588_io = clock_m->syn1588_io;
	struct io_matrix_sink *sink = NULL;

	if(!valid_sink(sink_ioctl)) {
		dev_err(dev, "invalid sink\n");
		return ERR_PTR(-ENODEV);
	}

	if(sink_ioctl->type == SYN1588_IOCTL_SINK_EVENT) {
		struct clock_m_event *event;
		event = &syn1588_io->events[sink_ioctl->id];
		if(event->resources == 0) {
			//device not exsist
			return ERR_PTR(-ENODEV);
		}
		if(event->state != EVENT_STATE_RUNNING) {
			return ERR_PTR(-EINVAL);
		}
		sink = &event->sink;
	}else if(sink_ioctl->type == SYN1588_IOCTL_SINK_IO_PORT) {
		struct clock_m_io_port *io_port;
		if(sink_ioctl->id < 0 || sink_ioctl->id >= __SYN1588_IOCTL_IO_PORT_NBITS) {
			dev_err(dev, "invalid io_port_id\n");
			return ERR_PTR(-EINVAL);
		}
		io_port = &syn1588_io->io_ports[sink_ioctl->id];
		// wenn using io port as sink it can not be a source
		if(io_port->source.connected_sink_list_head != NULL) {
			return ERR_PTR(-EBUSY);
		}
		if(io_port->resources == 0) {
			//device not exsist
			return ERR_PTR(-ENODEV);
		}
		sink = &io_port->sink;
	}else if(sink_ioctl->type == SYN1588_IOCTL_SINK_IRIG_B) {
		//TODO
		return ERR_PTR(-ENODEV);
	} else {
		return ERR_PTR(-ENODEV);
	}
	return sink;
}

static void io_matrix_connect(struct clock_m_device *clock_m, struct io_matrix_source *source, struct io_matrix_sink *sink) {
	struct device *dev = clock_m->dev;
	u32 reg;

	spin_lock(&clock_m->reg_lock_iomatrix);
	reg = ioread32(clock_m->iobase + IOMATRIX);
	reg = reg & ~( 0xFu << sink->reg_offset);

	if( source->type == SYN1588_IOCTL_SOURCE_IO_PORT) {
		// disable output on PORT IF ENABLED BY default or by Register access
		struct clock_m_io_port *io_port  = container_of(source, struct clock_m_io_port, source);
		reg = reg & ~( (u32)0xF << io_port->sink.reg_offset);
	}

	if(sink->type == SYN1588_IOCTL_SINK_IRIG_B) {
		// TODO
	} else if( sink->type == SYN1588_IOCTL_SINK_IO_PORT ) {
		reg = reg | ( source->reg_mask_gpio << sink->reg_offset);
	} else if( sink->type == SYN1588_IOCTL_SINK_EVENT ) {
		reg = reg | ( source->reg_mask_event << sink->reg_offset);
	} else {
		spin_unlock(&clock_m->reg_lock_iomatrix);
		dev_err(dev, "invalid sink type");
		return;
	}
	iowrite32(reg, clock_m->iobase + IOMATRIX);
	spin_unlock(&clock_m->reg_lock_iomatrix);

	dev_dbg(dev, "connected sink offset %d with mask %x\n", sink->reg_offset, source->reg_mask_gpio);

	sink->connected_source = source;
	if(source->connected_sink_list_head) {
		sink->next = source->connected_sink_list_head;
	}
	source->connected_sink_list_head = sink;

}
static void io_matrix_disconnect(struct clock_m_device *clock_m, struct io_matrix_source *source, struct io_matrix_sink *sink) {
	struct device *dev = clock_m->dev;
	struct io_matrix_sink *sink_list;
	u32 reg;

	spin_lock(&clock_m->reg_lock_iomatrix);
	reg = ioread32(clock_m->iobase + IOMATRIX);
	reg = reg & ~( 0xFu << sink->reg_offset);
	iowrite32(reg, clock_m->iobase + IOMATRIX);
	spin_unlock(&clock_m->reg_lock_iomatrix);

	if(source->connected_sink_list_head == sink) {
		source->connected_sink_list_head = sink->next;
	}else {
		for( sink_list = source->connected_sink_list_head ; sink_list != NULL ; sink_list = sink_list->next ) {
			if(sink_list->next == sink) {
				sink_list->next = sink_list->next->next;
			}
		}
	}
	sink->connected_source = NULL;

	dev_dbg(dev, "disconnected sink offset %d\n", sink->reg_offset);

}

int clock_m_io_matrix_connect(
	struct clock_m_device *clock_m,
	struct syn1588_ioctl_source source_ioctl,
	struct syn1588_ioctl_sink sink_ioctl
){
	struct io_matrix_source *source = to_internal_source(clock_m, &source_ioctl);
	struct io_matrix_sink *sink = to_internal_sink(clock_m, &sink_ioctl);

	if(IS_ERR_OR_NULL(sink)) {
		if(sink == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}
	if(IS_ERR_OR_NULL(source)) {
		if(source == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(source);
	}

	if(sink->connected_source != NULL) {
		return -EBUSY;
	}

	if(source->type == SYN1588_IOCTL_SOURCE_IO_PORT && sink->type == SYN1588_IOCTL_SINK_IO_PORT) {
		return -EPERM;
	}
	if(source->type != SYN1588_IOCTL_SOURCE_IO_PORT && sink->type == SYN1588_IOCTL_SINK_IRIG_B) {
		return -EPERM;
	}
	if(source->type == SYN1588_IOCTL_SOURCE_IRIG_B && sink->type != SYN1588_IOCTL_SINK_IO_PORT) {
		return -EPERM;
	}
	if(source->type == SYN1588_IOCTL_SOURCE_PLL_FILTERED && sink->type != SYN1588_IOCTL_SINK_IO_PORT) {
		return -EPERM;
	}

	if(sink->type == SYN1588_IOCTL_SINK_IRIG_B) {
		pr_err("IRIG_B not supported yet");
		return -EPERM;
	}

	io_matrix_connect(clock_m, source, sink);
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_matrix_connect);

int clock_m_io_matrix_disconnect(
	struct clock_m_device *clock_m,
	struct syn1588_ioctl_source source_ioctl,
	struct syn1588_ioctl_sink sink_ioctl
){
	struct device *dev = clock_m->dev;

	struct io_matrix_source *source = to_internal_source(clock_m, &source_ioctl);
	struct io_matrix_sink *sink = to_internal_sink(clock_m, &sink_ioctl);


	if(IS_ERR_OR_NULL(source)) {
		if(source == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(source);
	}
	if(IS_ERR_OR_NULL(sink)) {
		if(sink == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}

	if(sink->connected_source != source) {
		dev_warn(dev, "invalid iomatrix disconnect sink not connected");
		return -EINVAL;
	}

	io_matrix_disconnect(clock_m, source, sink);
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_matrix_disconnect);

int clock_m_io_matrix_remove_sink(
	struct clock_m_device *clock_m,
	struct syn1588_ioctl_sink sink_ioctl
){
	struct io_matrix_sink *sink = to_internal_sink(clock_m, &sink_ioctl);
	struct io_matrix_source *source;

	if(IS_ERR_OR_NULL(sink)) {
		if(sink == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}
	source = sink->connected_source;
	if(IS_ERR_OR_NULL(source)) {
		if(source == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}
	io_matrix_disconnect(clock_m, source, sink);
	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_matrix_remove_sink);

int clock_m_io_matrix_remove_source(
	struct clock_m_device *clock_m,
	struct syn1588_ioctl_source source_ioctl
){
	struct io_matrix_source *source = to_internal_source(clock_m, &source_ioctl);
	struct io_matrix_sink *sink = NULL;
	if(IS_ERR_OR_NULL(source)) {
		if(source == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}
	sink=source->connected_sink_list_head;
	if(IS_ERR_OR_NULL(sink)) {
		if(sink == NULL) {
			return -ENODEV;
		}
		return PTR_ERR(sink);
	}

	while(sink != NULL) {
		struct io_matrix_sink *sink_current = sink;
		sink = sink->next;
		io_matrix_disconnect(clock_m, source, sink_current);
	}

	return 0;
}
EXPORT_SYMBOL_GPL(clock_m_io_matrix_remove_source);

