
#include "clock_m.h"
#include "oregano-i2c.h"
#include "oregano_tse_mac.h"
#include "oregano_tse_mac_sfp.h"
#include "syn1588nic_local.h"
#include "version.h" // current version
#include "compat/compat.h"

#include <linux/bitops.h>
#include <linux/crc32.h>
#include <linux/delay.h>
#include <linux/dmapool.h>
#include <linux/etherdevice.h>
#include <linux/i2c.h>
#include <linux/net_tstamp.h>
#include <linux/module.h>
#include <linux/printk.h>
#include <linux/platform_device.h>
#include <linux/types.h>

#if defined(CONFIG_OF)
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/of_net.h>
#endif


#define DRV_NAME "oregano-mac"

//we will use the following build number throughout the module to set
// some register values of the TSE_MAC to work in the Meinberg HPS100 system (HPS100_*)
#define HPS100_PKTGEN_MIN_BUILD 4010

// values to set the minimum Inter Frame Gap for PktGen operation
#define HPS100_PKTGEN_IFG_GAP_100MBIT_VALUE									\
	25 //delay between packets in line clocks 25 => 1 us
#define HPS100_PKTGEN_IFG_GAP_1GBIT_VALUE									  \
	125 //delay between packets in line clocks 125 => 1 us




/// NIC private data structure (MAC device)
struct nic_priv_data {
	// only one of this struct is active
	// this is determinated by the i2c_interface
	struct oregano_mac_sfp sfp;
	struct mii_bus *mii_bus;

#if KERNEL_VERSION(2, 6, 36) > LINUX_VERSION_CODE
	// older version has no phydev in the net_device struct
	struct phy_device *phydev;
#endif

	struct device *clock_m; ///< pointer to clock_m structure
	struct net_device *net_dev;

	void __iomem *iobase;

	u8 bd_cnt; ///< number of DMA buffers per direction
	u32 mac_mode_reg;

	syn1588_phydelay_getter_t phydelay_getter;

	struct hwtstamp_config tstamp_config;

	struct workqueue_struct *workq;
	struct work_struct work_watchdog;

	struct tasklet_struct tl_rx; ///< bottom-half handler for this device
	struct tasklet_struct tl_tx; ///< bottom-half handler for this device

	int irqs[OREGANO_MAC_IRQ_AMOUNT];

	volatile unsigned long lost_irqs; ///< lost IRQ mask (IRQ handler -> BH handler)
	volatile unsigned long irh_irsrc; ///< virtual IRSRC register for IRQ handler
	volatile unsigned long bhh_irsrc; ///< virtual IRSRC register for bottom half handler

	char version[16]; ///< MAC version number

	struct syn1588_dma *dma; ///< DMA pool for ring buffers
	struct bd_addr rx_dma_buffer[BUFF_CNT]; ///< receiver buffer ring
	struct bd_addr tx_dma_buffer[BUFF_CNT]; ///< transmit buffer ring
	struct sk_buff *tx_skb_buffer[BUFF_CNT]; ///< remember associated skb
	spinlock_t tx_dma_bdp_lock; ///< TX ring pointer lock


	unsigned char rx_dma_bd_rp; ///< RX ring read pointer
	unsigned char tx_dma_bd_rp, tx_dma_bd_wp; ///< TX ring read/write pointer

	spinlock_t mdio_lock; ///< locks access to mdio
	struct semaphore mii_access; ///< mutual exclusive lock MII access

	struct timer_list watchdog; ///< NIC watchdog (checking linkstate)
	struct net_device_stats stats; ///< NIC usage statistics
	struct nic_extra_stats extra_stats; ///< NIC extra statistical information
	struct mii_if_info mii_if; ///< MII API hooks, info
	u32 mii_clkdiv; ///< MII clock divider

	// MAC capability bits
	bool timestamping_tx; ///< SO_TIMESTAMPING tx
	bool timestamping_rx; ///< SO_TIMESTAMPING rx

	DECLARE_BITMAP(vlan_grp_bitmap, 4096);
	spinlock_t vlan_grp_bitmap_lock;

	unsigned int msg_enable; ///< network debugging level

	/**
	 * see  `MAC_CAPABILITY_0` in `an_tsemac_register_map`
	 * for more information
	 */
	struct oregano_mac_capabilities {
		bool rx_timestamper_implemented;
		bool tx_timestamper_implemented;
		bool one_step_implemented;
		bool statistic_counter_32bit;
		bool vlan_4_selectable;
		bool vlan_group_implemented;
		bool statistic_counter_additional_implemented;
		bool base_1g;
		bool base_10g;
		bool base_25g;
		bool fec_available; // Forward Error Correction
		bool jumbo_paket_support;
		bool axi_interface;
		bool second_timestamper;
		bool ptp_timestamper;
		bool fiber_interface_status_implemented; // only for 10/25G nics
		bool ptp_security_implemented;
		bool variable_burst;
		bool sw_i2c_mdio; // no fake mdio implementation
		bool i2c_interface; // we have an i2c interface instant of mdio

		bool programmable_interface_gap;
	} mac_capabilities;

	/**
	 * alignment of the timestamp behind a rx package
	 */
	int timestamp_alignment;
	/*
	struct oregano_mac_capabilities{
		enum oregano_mac_type {
			EMAC_TSE = 0,
			EMAC_XMAC //10G / 25G
		} mac;
		bool has_pll;
		bool has_fpga_conf;
		bool has_video_dac;
		bool has_eeprom;
		bool has_ocxo;
	} mac_capabilities;
	*/

#ifdef DEBUG_BHH_TX_QUEUE
	syn1588_timespec netif_queue_stoptime; ///< time when queue was stopped

	unsigned long
		netif_queue_maxstop; ///< maximum observed time queue in stopped state
#endif
};

#if KERNEL_VERSION(2, 6, 36) <= LINUX_VERSION_CODE
#define nic_phydev(nic_dev) (nic_dev->phydev)
#define nic_phydev_set(nic_dev, value) (nic_dev->phydev = value)
#else
#define BACKPORT_NICDEV_PHYDEV
#define nic_phydev(nic_dev) (((struct nic_priv_data *)netdev_priv(nic_dev))->phydev)
#define nic_phydev_set(nic_dev, phydev) (((struct nic_priv_data *)netdev_priv(nic_dev))->phydev = phydev)
#endif

#if KERNEL_VERSION(5, 6, 0) > LINUX_VERSION_CODE && !defined(RHEL8_BACKPORT_PHY_IOCTL) && !defined(SUSE15SP3_PHY_DO_IOCTRL)
#include <linux/phy.h>
#include <linux/netdevice.h>
static inline int phy_do_ioctl(struct net_device *ndev, struct ifreq *ifr, int cmd)
{
	if (!nic_phydev(ndev))
		return -ENODEV;

	#if KERNEL_VERSION(2, 6, 36) < LINUX_VERSION_CODE
	return phy_mii_ioctl(nic_phydev(ndev), ifr, cmd);
	#else
	return phy_mii_ioctl(nic_phydev(ndev), if_mii(ifr), cmd);
	#endif
}
#endif

struct syn1588_stats {
	char stat_string[ETH_GSTRING_LEN];
	//int type;
	int sizeof_stat;
	int stat_offset;
};

#define SYN1588_STAT(m)		sizeof(((struct nic_priv_data *)0)->m), \
				offsetof(struct nic_priv_data, m)

static const struct syn1588_stats syn1588_gstrings_stats[] = {
	{"rx_packets", SYN1588_STAT(stats.rx_packets)},
	{"tx_packets", SYN1588_STAT(stats.tx_packets)},
	{"rx_bytes", SYN1588_STAT(stats.rx_bytes)},
	{"tx_bytes", SYN1588_STAT(stats.tx_bytes)},
	{"rx_errors", SYN1588_STAT(stats.rx_errors)},
	{"tx_errors", SYN1588_STAT(stats.tx_errors)},
	{"rx_dropped", SYN1588_STAT(stats.rx_dropped)},
	{"tx_dropped", SYN1588_STAT(stats.tx_dropped)},
	{"multicast", SYN1588_STAT(stats.multicast)},
	{"collisions", SYN1588_STAT(stats.collisions)},
	{"rx_length_errors", SYN1588_STAT(stats.rx_length_errors)},
	{"rx_over_errors", SYN1588_STAT(stats.rx_over_errors)},
	{"rx_crc_errors", SYN1588_STAT(stats.rx_crc_errors)},
	{"rx_frame_errors", SYN1588_STAT(stats.rx_frame_errors)},
	{"rx_fifo_errors", SYN1588_STAT(stats.rx_fifo_errors)},
	{"rx_missed_errors", SYN1588_STAT(stats.rx_missed_errors)},
	{"tx_aborted_errors", SYN1588_STAT(stats.tx_aborted_errors)},
	{"tx_carrier_errors", SYN1588_STAT(stats.tx_carrier_errors)},
	{"tx_fifo_errors", SYN1588_STAT(stats.tx_fifo_errors)},
	{"tx_heartbeat_errors", SYN1588_STAT(stats.tx_heartbeat_errors)},
	{"tx_window_errors", SYN1588_STAT(stats.tx_window_errors)},
	// device-specific stats
	{"rx_ctrl_packets", SYN1588_STAT(extra_stats.rx_ctrl_packets)},
	{"rx_vlan_packets", SYN1588_STAT(extra_stats.rx_vlan_packets)},
	{"mcast_dropped", SYN1588_STAT(extra_stats.mcast_dropped)},
	{"rx_swcrc_err", SYN1588_STAT(extra_stats.rx_swcrc_err)},
	{"rx_late_coll", SYN1588_STAT(extra_stats.rx_late_coll)},
	{"drv_rx_queue_full", SYN1588_STAT(extra_stats.drv_rx_qfull)},
	{"drv_tx_queue_full", SYN1588_STAT(extra_stats.drv_tx_qfull)},
	{"tx_pkt_deferred", SYN1588_STAT(extra_stats.tx_pkt_deferred)},
	{"vlan_err", SYN1588_STAT(extra_stats.vlan_err)},
	{"ts_rx", SYN1588_STAT(extra_stats.ts_rx)},
	{"ts_tx", SYN1588_STAT(extra_stats.ts_tx)},
	{"ts2_rx", SYN1588_STAT(extra_stats.ts2_rx)},
	{"ts2_tx", SYN1588_STAT(extra_stats.ts2_tx)},
	{"ts_ptp_rx", SYN1588_STAT(extra_stats.ts_ptp_rx)},
	{"ts_ptp_tx", SYN1588_STAT(extra_stats.ts_ptp_tx)},
	{"hw_rx_packets", SYN1588_STAT(extra_stats.hw_rx_packets)},
	{"hw_tx_packets", SYN1588_STAT(extra_stats.hw_tx_packets)},
	{"hw_tx_error", SYN1588_STAT(extra_stats.hw_tx_error)},
	{"hw_vlan_packets", SYN1588_STAT(extra_stats.hw_vlan_packets)},
	{"hw_broadcast_packets", SYN1588_STAT(extra_stats.hw_broadcast_packets)},
	{"hw_multicast_packets", SYN1588_STAT(extra_stats.hw_multicast_packets)}
};


#define SYN1588_STATS_LEN ARRAY_SIZE(syn1588_gstrings_stats)

static inline void __iomem *PCI_MAC_BASE_DIR(void __iomem *dev)
{
	return dev + MAC_BASE;
}
// Read some data from a port
static inline u32 nic_readl(struct nic_priv_data *dev, u32 offset)
{
	return ioread32(dev->iobase + offset);
}
// Read some data from a port
static inline u32 nic_directreadl(void __iomem *dev, u32 offset)
{
	return ioread32(PCI_MAC_BASE_DIR(dev) + offset);
}
// Write some data to a port
static inline void nic_writel(struct nic_priv_data *dev, u32 val, u32 offset)
{
	iowrite32(val, dev->iobase + offset);
}
// Write some data to a port
// with a read back check
// true if sucessfully otherwise false
__must_check static inline bool nic_writel_check(struct nic_priv_data *dev, u32 val, u32 offset)
{

	dev_dbg(&dev->net_dev->dev, "write with check to MAC address 0x%lX data: 0x%lX\n", (long unsigned int)offset,
		 (unsigned long)val);
	iowrite32(val, dev->iobase + offset);
	if(ioread32(dev->iobase + offset) != val) {
		return false;
	}
	return true;
}


static inline u32 nic_get_bd_stat_tx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE + nr * 16); // BD size is always 16 Byte
}
static inline void nic_set_bd_stat_tx(struct nic_priv_data *dev, int nr, u32 value) {
	nic_writel(dev, value, MAC_BD_BASE + nr * 16); // BD size is always 16 Byte
}

static inline u32 nic_get_bd_stat_rx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE +  (dev->bd_cnt + nr) * 16); // BD size is always 16 Byte
}
static inline void nic_set_bd_stat_rx(struct nic_priv_data *dev, int nr, u32 value) {
	nic_writel(dev, value, MAC_BD_BASE + (dev->bd_cnt + nr) * 16);  // BD size is always 16 Byte
}

static inline u32 nic_get_bd_addr_tx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE + nr * 16 + 4); // BD size is always 16 Byte
}
static inline void nic_set_bd_addr_tx(struct nic_priv_data *dev, int nr, u32 value) {
	nic_writel(dev, value, MAC_BD_BASE + nr * 16 + 4); // BD size is always 16 Byte
}

static inline u32 nic_get_bd_addr_rx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE +  (dev->bd_cnt + nr) * 16 + 4); // BD size is always 16 Byte
}
static inline void nic_set_bd_addr_rx(struct nic_priv_data *dev, int nr, u32 value) {
	nic_writel(dev, value, MAC_BD_BASE + (dev->bd_cnt + nr) * 16 + 4); // BD size is always 16 Byte
}

static inline u32 nic_get_bd_ts1_tx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE + nr * 16 + 8); // BD size is always 16 Byte
}
static inline u32 nic_get_bd_ts2_tx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE + nr * 16 + 12); // BD size is always 16 Byte
}

static inline u32 nic_get_bd_ts1_rx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE +  (dev->bd_cnt + nr) * 16 + 8); // BD size is always 16 Byte
}
static inline u32 nic_get_bd_ts2_rx(struct nic_priv_data *dev, int nr) {
	return nic_readl(dev, MAC_BD_BASE +  (dev->bd_cnt + nr) * 16 + 12); // BD size is always 16 Byte
}

#define TX_BD(nr) (MAC_BD_BASE + (nr)*16)
#define RX_BD(nr)															  \
	(MAC_BD_BASE + nic_dat->bd_cnt * 16 + (nr)*16)
#define RX_BDP_INC(pointer) (((pointer) + 1) % nic_dat->bd_cnt)
#define TX_BDP_INC(pointer) (((pointer) + 1) % nic_dat->bd_cnt)
#define TX_BDP_2INC(pointer) (((pointer) + 2) % nic_dat->bd_cnt)



u32 oregano_mac_readl(struct nic_priv_data *dev, u32 offset){
	return nic_readl(dev, offset);
}
EXPORT_SYMBOL_GPL(oregano_mac_readl);
void oregano_mac_writel(struct nic_priv_data *dev, u32 val, u32 offset){
	nic_writel(dev, val, offset);
}
EXPORT_SYMBOL_GPL(oregano_mac_writel);

__must_check struct net_device *nic_dev_alloc(struct device *parent_device, void __iomem *iobase, struct syn1588_dma *dma_pool, int irq, struct oregano_mac_platform_data *pdata, syn1588_phydelay_getter_t phydelay_getter);

void nic_dev_exit(struct net_device *net_dev);
int nic_dev_setconfig(struct net_device *);

void syn1588_nic_print_capabilities(struct net_device *ndev){
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	pr_info("nic_cap rx_timestamper_implemented: %d\n", nic_dat->mac_capabilities.rx_timestamper_implemented);
	pr_info("nic_cap tx_timestamper_implemented: %d\n", nic_dat->mac_capabilities.tx_timestamper_implemented);
	pr_info("nic_cap one_step_implemented: %d\n", nic_dat->mac_capabilities.one_step_implemented);
	pr_info("nic_cap statistic_counter_32bit: %d\n", nic_dat->mac_capabilities.statistic_counter_32bit);
	pr_info("nic_cap vlan_4_selectable: %d\n", nic_dat->mac_capabilities.vlan_4_selectable);
	pr_info("nic_cap vlan_group_implemented: %d\n", nic_dat->mac_capabilities.vlan_group_implemented);
	pr_info("nic_cap statistic_counter_additional_implemented: %d\n", nic_dat->mac_capabilities.statistic_counter_additional_implemented);
	pr_info("nic_cap base_1g: %d\n", nic_dat->mac_capabilities.base_1g);
	pr_info("nic_cap base_10g: %d\n", nic_dat->mac_capabilities.base_10g);
	pr_info("nic_cap base_25g: %d\n", nic_dat->mac_capabilities.base_25g);
	pr_info("nic_cap fec_available: %d\n", nic_dat->mac_capabilities.fec_available);
	pr_info("nic_cap jumbo_paket_support: %d\n", nic_dat->mac_capabilities.jumbo_paket_support);
	pr_info("nic_cap axi_interface: %d\n", nic_dat->mac_capabilities.axi_interface);
	pr_info("nic_cap second_timestamper: %d\n", nic_dat->mac_capabilities.second_timestamper);
	pr_info("nic_cap ptp_timestamper: %d\n", nic_dat->mac_capabilities.ptp_timestamper);
	if(nic_dat->mac_capabilities.base_10g || nic_dat->mac_capabilities.base_25g) {
		pr_info("nic_cap fiber_interface_status_implemented: %d\n", nic_dat->mac_capabilities.fiber_interface_status_implemented);
	}
	pr_info("nic_cap ptp_security_implemented: %d\n", nic_dat->mac_capabilities.ptp_security_implemented);
	pr_info("nic_cap variable_burst: %d\n", nic_dat->mac_capabilities.variable_burst);
	pr_info("nic_cap sw_i2c_mdio: %d\n", nic_dat->mac_capabilities.sw_i2c_mdio);
	pr_info("nic_cap i2c_interface: %d\n", nic_dat->mac_capabilities.i2c_interface);
	pr_info("nic_cap programmable_interface_gap: %d\n", nic_dat->mac_capabilities.programmable_interface_gap);
}

void syn1588_nic_read_capabilities(struct net_device *ndev){
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	u32 nic_capabilities;

	nic_capabilities = ioread32(nic_dat->iobase + MAC_CAPABILITY_0);
	if(nic_capabilities & MAC_CAP0_RXTS)
		nic_dat->mac_capabilities.rx_timestamper_implemented = true;
	if(nic_capabilities & MAC_CAP0_TXTS)
		nic_dat->mac_capabilities.tx_timestamper_implemented = true;
	if(nic_capabilities & MAC_CAP0_1STEP)
		nic_dat->mac_capabilities.one_step_implemented = true;
	if(nic_capabilities & MAC_CAP0_STAT_CNT_LEN)
		nic_dat->mac_capabilities.statistic_counter_32bit = true;
	if(nic_capabilities & MAC_CAP0_VLAN_ID_NUM)
		nic_dat->mac_capabilities.vlan_4_selectable = true;
	if(nic_capabilities & MAC_CAP0_VLAN_GRP)
		nic_dat->mac_capabilities.vlan_group_implemented = true;
	if(nic_capabilities & MAC_CAP0_EXT_STAT)
		nic_dat->mac_capabilities.statistic_counter_additional_implemented = true;
	if(nic_capabilities & MAC_CAP0_1G_MAC)
		nic_dat->mac_capabilities.base_1g = true;
	if(nic_capabilities & MAC_CAP0_10G_MAC)
		nic_dat->mac_capabilities.base_10g = true;
	if(nic_capabilities & MAC_CAP0_25G_MAC)
		nic_dat->mac_capabilities.base_25g = true;
	if(nic_capabilities & MAC_CAP0_FEC)
		nic_dat->mac_capabilities.fec_available = true;
	if(nic_capabilities & MAC_CAP0_JUMBO)
		nic_dat->mac_capabilities.jumbo_paket_support = true;
	if(nic_capabilities & MAC_CAP0_AXI)
		nic_dat->mac_capabilities.axi_interface = true;
	if(nic_capabilities & MAC_CAP0_TS2)
		nic_dat->mac_capabilities.second_timestamper = true;
	if(nic_capabilities & MAC_CAP0_FIBER)
		nic_dat->mac_capabilities.fiber_interface_status_implemented = true;
	if(nic_capabilities & MAC_CAP0_SECURITY)
		nic_dat->mac_capabilities.ptp_security_implemented = true;
	if(nic_capabilities & MAC_CAP0_VARIABLE_BURST)
		nic_dat->mac_capabilities.variable_burst = true;
	if(nic_capabilities & MAC_CAP0_SW_MDIO)
		nic_dat->mac_capabilities.sw_i2c_mdio = true;
	if(nic_capabilities & MAC_CAP0_I2C)
		nic_dat->mac_capabilities.i2c_interface = true;
	if(nic_capabilities & MAC_CAP0_PTP_TS)
		nic_dat->mac_capabilities.ptp_timestamper = true;
	if(nic_capabilities & MAC_CAP1_INTERFRAME_GAP)
		nic_dat->mac_capabilities.programmable_interface_gap = true;

	// build 840 and previous sfp workaround
	if( !nic_dat->mac_capabilities.base_1g &&
	   !nic_dat->mac_capabilities.base_10g && !nic_dat->mac_capabilities.base_25g
	) {
		pr_warn("apply syn1588Nic2.x build <=840, mac workaround\n");
		nic_dat->mac_capabilities.base_1g = true;
	}

	nic_dat->timestamp_alignment = 4;
	if(nic_dat->mac_capabilities.base_10g || nic_dat->mac_capabilities.base_25g) {
		nic_dat->timestamp_alignment = 8;
	}
}

/**
 * Handle network device receive interrupt
 * \param ndev pointer to network device structure to set multicast list for
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
int nic_dev_rx_irq(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct sk_buff *skb;
	struct bd_addr *cur_buf;
	volatile u32 bd_len_stats = 0x0; // buffer descriptor length and status
	int bd_scan_count = 0; // how many BD have been scanned
	struct clock_m_device *clock_m = dev_get_clock_m(nic_dat->clock_m);
	const bool ntp_mode_enabled = clock_m && clock_m->ntpModeEnabled;

	if(netif_msg_rx_status(nic_dat))
		pr_debug("*** nic_dev_rx_irq ***\n");

	// read current buffer descriptor status and continue if not empty
	for (bd_scan_count = 0; bd_scan_count < nic_dat->bd_cnt; bd_scan_count++) {
		cur_buf = &nic_dat->rx_dma_buffer[nic_dat->rx_dma_bd_rp];

		bd_len_stats = nic_get_bd_stat_rx(nic_dat, nic_dat->rx_dma_bd_rp);

		// check if current buffer descriptor is filled
		if (bd_len_stats & MAC_RX_BD_EMPTY) {
			break;
		}

		// TODO: check all error conditions
		// check status for errors
		if (bd_len_stats & MAC_RX_BD_MISS) {
			nic_dat->stats.multicast++;
		}
		if (bd_len_stats & MAC_RX_BD_CTRLFRAME) {
			nic_dat->extra_stats.rx_ctrl_packets++;
		}
		if (bd_len_stats & MAC_RX_BD_VLANFRAME) {
			nic_dat->extra_stats.rx_vlan_packets++;
		}

		// if packet is ok, then proceed
		if (bd_len_stats &
			(MAC_RX_BD_STATS & ~MAC_RX_BD_MISS & ~MAC_RX_BD_CTRLFRAME &
			 ~MAC_RX_BD_VLANFRAME & ~MAC_RX_BD_TIMESTAMP)) {
			if(netif_msg_rx_status(nic_dat))
				pr_debug("received packet contained error! packet lost.\n");
			nic_dat->stats.rx_errors++;
		} else {
			const unsigned short bd_size =
				((bd_len_stats & 0xFFFF0000) >> 16);

			const unsigned short pkt_size = bd_size;

			if(netif_msg_rx_status(nic_dat))
				pr_debug("received packet len %u bytes.\n", pkt_size);
#ifdef JUMBO_PACKET
			if (DMA_BD_SIZE < bd_size || ETH_ZLEN > pkt_size ||
				ETH_MAX_JUMBO_FRAME_SIZE < pkt_size)
#else
			if (DMA_BD_SIZE < bd_size || ETH_ZLEN > pkt_size ||
				(ETH_FRAME_LEN + ETH_VLANTAG_LEN) < pkt_size)
#endif
			{
				if(netif_msg_rx_status(nic_dat))
					pr_debug("packet length (%u) doesn't match DMA buffer size (%u):\npacket dropped.\n",
					 		pkt_size, DMA_BD_SIZE);
				nic_dat->stats.rx_dropped++;
			} else {
				skb = dev_alloc_skb(bd_size);
				if (IS_ERR_OR_NULL(skb)) {
					pr_warn("unable to allocate socket buffer! packet lost.\n");
					nic_dat->stats.rx_dropped++;
				} else {
					// set meta information
					skb->dev = ndev;
					skb->ip_summed = CHECKSUM_NONE; // don't check it, was done by HW

					// copy paket data to socket buffer
					memcpy(skb_put(skb, bd_size),
						   cur_buf->vaddr, bd_size);

					{
						// set protocol information based on packet data
						skb->protocol = eth_type_trans(
							skb, ndev);

						// If RX timestamping is enabled and BD indicates an attached
						// timestamp
						if (nic_dat->tstamp_config.rx_filter != HWTSTAMP_FILTER_NONE && (bd_len_stats & MAC_RX_BD_TIMESTAMP) > 0) {
							u64 timestamp = 0;
							u32 __iomem *ts_addr;
							u32 sec;
							unsigned int align;

							// align to next 32bit boundary
							align = (bd_size + nic_dat->timestamp_alignment-1) & ~(nic_dat->timestamp_alignment-1);
							ts_addr = cur_buf->vaddr + align;

							// convert s to ns
							sec = ioread32be(ts_addr);
							if (ntp_mode_enabled) {
								// Subtract 70 years offet between NTP and PTP timescale
								sec = sec - (u32)NTP_PTP_OFFSET;
							}
							timestamp = (u64)sec * 1000000000ULL;

							if (ntp_mode_enabled) {
								// NTP mode enabled, convert NTP format to PTP befor using it
								u64 temp = ioread32be(ts_addr + 1);

								temp = (temp *
									1000000000ULL) >>
									   32;
								timestamp +=
									(u32)temp;
							} else {
								timestamp += (u64)(ioread32be(ts_addr+1) % 1000000000U);
							}

							skb_hwtstamps(skb)->hwtstamp = ns_to_ktime(timestamp);
						}

						// schedule network layer
						netif_rx(skb);
						// update stats
						nic_dat->stats.rx_packets++;
						nic_dat->stats.rx_bytes +=
							pkt_size;
					}
				}
			}
		}

		if(netif_msg_rx_status(nic_dat))
			pr_debug("orig. RX BD (%u) status: 0x%X.\n",
				nic_dat->rx_dma_bd_rp, bd_len_stats);
		// reset status and length
		bd_len_stats = bd_len_stats & (~MAC_RX_BD_STATS & 0x0000FFFF);
		bd_len_stats |= MAC_RX_BD_EMPTY; // set empty flag again
		// write regenerated status and length to buffer descriptor
		nic_set_bd_stat_rx(nic_dat, nic_dat->rx_dma_bd_rp, bd_len_stats);

		if(netif_msg_rx_status(nic_dat))
			pr_debug("worked on RX BD (%d) status: 0x%X.\n",
			(int)nic_dat->rx_dma_bd_rp, bd_len_stats & 0x0000FFFF);

		// switch to next buffer descriptor in the ring
		nic_dat->rx_dma_bd_rp = RX_BDP_INC(nic_dat->rx_dma_bd_rp);
	}

	if (!bd_scan_count && netif_msg_rx_status(nic_dat)) {
		unsigned short count;

		pr_debug("no receive DMA buffer used, but MAC issued an IRQ!\n");
		// FIXME! remove!
		for (count = 0; count < nic_dat->bd_cnt; count++) {
			pr_debug("#%02u TXBD: 0x%08X, RXBD: 0x%08X\n", count,
				 nic_get_bd_stat_tx(nic_dat, count),
				 nic_get_bd_stat_rx(nic_dat, count));
		}
		return -ENODATA;
	}

	if (bd_scan_count > 1 && netif_msg_rx_status(nic_dat))
		pr_debug("worked on multiple (%u) RX DMA buffers!\n",
			 bd_scan_count);

	if (nic_dat->bd_cnt - 1 == bd_scan_count) {
		pr_info("high load on RX DMA buffer ring.\n");
	}

	return 0;
}


/**
 * Handle network device transmit interrupt.
 * A packet has been sent onto the line by the MAC
 * \param ndev pointer to network device structure to set multicast list for
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
int nic_dev_tx_irq(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	u32 bd_len_stats; // buffer descriptor length/status
	struct sk_buff *skb;
	int bd_scan_count;
	struct clock_m_device *clock_m = dev_get_clock_m(nic_dat->clock_m);
	const bool ntp_mode_enabled = clock_m && clock_m->ntpModeEnabled;


	if(netif_msg_tx_done(nic_dat))
		pr_debug("*** nic_dev_tx_irq ***\n");

	spin_lock(&nic_dat->tx_dma_bdp_lock);
	if (nic_dat->tx_dma_bd_rp == nic_dat->tx_dma_bd_wp) {
		spin_unlock(&nic_dat->tx_dma_bdp_lock);
		if(netif_msg_tx_done(nic_dat))
			pr_info("no transmit buffer used, but MAC issued an IRQ!\n");
		return -ENODATA;
	}

	for (bd_scan_count = 0; bd_scan_count < nic_dat->bd_cnt;
		 bd_scan_count++) {
		if (nic_dat->tx_dma_bd_rp == nic_dat->tx_dma_bd_wp) {
			// there are no more pending DMA buffers
			break;
		}

		// read stats of current buffer descriptor
		if(netif_msg_tx_done(nic_dat))
			pr_debug("reading TX BD (%u).\n",
				nic_dat->tx_dma_bd_rp);
		bd_len_stats = nic_get_bd_stat_tx(nic_dat, nic_dat->tx_dma_bd_rp);

		if (bd_len_stats & MAC_TX_BD_READY) {
			if (!bd_scan_count)
				// an interrrupt was fired but no BD is ready to be evaluated
				// happens on high load and early gabing an BD befor an interrupt arrives.
				pr_debug("tried to evaluate unsent transmit DMA buffer!\n");
			break;
		}

		skb = nic_dat->tx_skb_buffer[nic_dat->tx_dma_bd_rp];
		if (skb) {
			if ((nic_dat->tstamp_config.tx_type == HWTSTAMP_TX_ONESTEP_SYNC || nic_dat->tstamp_config.tx_type == HWTSTAMP_TX_ON) &&
#if (KERNEL_VERSION(2, 6, 37) <= LINUX_VERSION_CODE)
				(skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP) &&
#else
				unlikely(skb_tx(skb)->hardware) &&
#endif
				(bd_len_stats & MAC_TX_BD_TIMESTAMP)) {
				u32 tmp;
				struct skb_shared_hwtstamps hwtstamps;

				tmp = nic_get_bd_ts1_tx(nic_dat, nic_dat->tx_dma_bd_rp);
				if (ntp_mode_enabled) {
					// Subtract 70 years offset between NTP and PTP timescale
					tmp = tmp - (u32)NTP_PTP_OFFSET;
				}
				// convert s to ns
				hwtstamps.hwtstamp =
					ns_to_ktime((u64)tmp * 1000000000LL);
				tmp = nic_get_bd_ts2_tx(nic_dat, nic_dat->tx_dma_bd_rp);

				if (ntp_mode_enabled) {
					// NTP mode enabled, convert NTP format to PTP before using it
					tmp = (u32)(
						((u64)tmp * 1000000000ULL) >>
						32);
					hwtstamps.hwtstamp = ktime_add_ns(
						hwtstamps.hwtstamp, tmp);
				} else {
					// add ns, mask out the upper ultimate bits
					hwtstamps.hwtstamp = ktime_add_ns(
						hwtstamps.hwtstamp,
						(u64)(0x3FFFFFFF & tmp));
				}

				skb_tstamp_tx(skb, &hwtstamps);
			}
			// free associated skb
			dev_kfree_skb(skb);
			nic_dat->tx_skb_buffer[nic_dat->tx_dma_bd_rp] = 0x0;
		}

		// increment read pointer
		nic_dat->tx_dma_bd_rp = TX_BDP_INC(nic_dat->tx_dma_bd_rp);

		// update statistics
		if (bd_len_stats & MAC_TX_BD_RETLIM)
			nic_dat->stats.tx_aborted_errors++;
		if (bd_len_stats & MAC_TX_BD_CARRIER)
			nic_dat->stats.tx_carrier_errors++;
		if (bd_len_stats & MAC_TX_BD_LATECOL)
			nic_dat->stats.tx_window_errors++;
		if (bd_len_stats & MAC_TX_BD_DEFER)
			nic_dat->extra_stats.tx_pkt_deferred++;

		if (bd_len_stats & (MAC_TX_BD_LATECOL | MAC_TX_BD_RETLIM))
			nic_dat->stats.tx_errors++;

		nic_dat->stats.tx_packets++;
		nic_dat->stats.tx_bytes += ((bd_len_stats & 0xFFFF0000) >> 16);
		nic_dat->stats.collisions += (bd_len_stats >> 4) & 0x000F;
	}

	if (bd_scan_count > 1 && netif_msg_tx_done(nic_dat))
		pr_debug("worked on multiple (%u) TX DMA buffers!\n",
			 bd_scan_count);

	if (nic_dat->bd_cnt - 1 == bd_scan_count && netif_msg_tx_done(nic_dat))
		pr_info("high load on TX DMA buffer ring.\n");


	if (netif_queue_stopped(ndev) &&
		(nic_dat->tx_dma_bd_rp != TX_BDP_INC(nic_dat->tx_dma_bd_wp))) {
#ifdef DEBUG_BHH_TX_QUEUE
		syn1588_timespec netif_queue_waketime, delta;

		syn1588_ktime_get_real_ts(&netif_queue_waketime);
		delta = syn1588_timespec_sub(netif_queue_waketime,
						 nic_dat->netif_queue_stoptime);
		if (delta.tv_nsec >
			DEBUG_BHH_MINIMUM_REPORT_DELTA) { // ony print when
			// processing takes
			// longer than 10 us
			if (delta.tv_nsec > nic_dat->netif_queue_maxstop) {
				nic_dat->netif_queue_maxstop = delta.tv_nsec;
				pr_warn("netif_queue_sleep: " SYN1588_TIMESPEC_PRINTF_FORMAT
					   ".%09lu	max: %09lu\n",
					   delta.tv_sec, delta.tv_nsec,
					   nic_dat->netif_queue_maxstop);
			}
		}
#endif
		netif_wake_queue(ndev);
	}
	spin_unlock(&nic_dat->tx_dma_bdp_lock);

	return 0;
}


/**
 * Bottom half IRQ handler RX
 * \param pt_adapter pointer to PCIe adapter the IRQ was for
 */
void syn1588_BHH_RX(unsigned long pt_adapter)
{
	struct net_device *net_dev = (struct net_device *)pt_adapter;
	nic_dev_rx_irq(net_dev);
}

/**
 * Bottom half IRQ handler RX
 * \param pt_adapter pointer to PCIe adapter the IRQ was for
 */
void syn1588_BHH_TX(unsigned long pt_adapter)
{
	struct net_device *net_dev = (struct net_device *)pt_adapter;
	nic_dev_tx_irq(net_dev);
}


/**
 * Interrupt handler for the syn1588(R) NIC driver
 * Called by the system if there happened an interrupt on the devices interrupt
 * line
 * \param irq Number of the interrupt
 * \param dev_id Pointer to the local data structure
 * \return IRQ_HANDLED if IRQ was from PCIe-NIC and will be processed, otherwise
 * IRQ_NONE
 */
irqreturn_t nic_interrupt(int irq, void *dev_id)
{
	volatile u32 temp_nic_irsrc;
	struct net_device *ndev = dev_id;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	int count;

	temp_nic_irsrc = nic_readl(nic_dat, MAC_INT_SOURCE); // read NIC interrupt source register

	if ((!temp_nic_irsrc))
		return IRQ_NONE; // this call was due to a shared interrupt line


	if (temp_nic_irsrc != 0) {
		// clear interrupt source register (acknowledge read irq's)
		nic_writel(nic_dat, temp_nic_irsrc, MAC_INT_SOURCE);
		// copy interrupts that were read out
		for (count = 0; count < sizeof(u32) * BITS_PER_BYTE; count++)
			if (unlikely((0 != (temp_nic_irsrc & 0x1 << count)) &&
					 (0 !=
					  test_and_set_bit(count,
							   &nic_dat->irh_irsrc))))
				set_bit(count, &nic_dat->lost_irqs);

		if (temp_nic_irsrc & (MAC_INT_TXB | MAC_INT_TXE)) {
			tasklet_schedule(&nic_dat->tl_tx);
		}

		if(temp_nic_irsrc & MAC_INT_RXE) {
			nic_dat->stats.rx_errors++;
		}

		if (temp_nic_irsrc &
			(MAC_INT_RXF  | MAC_INT_BUSY)) {
			tasklet_schedule(&nic_dat->tl_rx);
		}
	}

	if(netif_msg_intr(nic_dat))
		pr_debug("*** syn1588(R) PCIe/Dual NIC IRQ handler 0x%X ***\n",
		temp_nic_irsrc);

	return IRQ_HANDLED;
}


static void update_vlan_grp(struct net_device *ndev);


/**
 * <!--********************************************************************************-->
 * \name ETHTOOL function prototypes
 * <!--********************************************************************************-->
 * @{
 */

static void syn1588_get_drvinfo(struct net_device *, struct ethtool_drvinfo *);
// static void syn1588_get_pauseparam(struct net_device *, struct
// ethtool_pauseparam*);
static int syn1588_get_regs_len(struct net_device *);
static void syn1588_get_regs(struct net_device *, struct ethtool_regs *regs,
				 void *p);
static u32 syn1588_get_msglevel(struct net_device *);
static void syn1588_set_msglevel(struct net_device *, u32);
static void syn1588_get_strings(struct net_device *, u32, u8 *);
static int syn1588_get_stats_count(struct net_device *);
static void syn1588_get_ethtool_stats(struct net_device *,
					  struct ethtool_stats *, u64 *);
static int syn1588_begin(struct net_device *);
static void syn1588_complete(struct net_device *);
#if (KERNEL_VERSION(2, 6, 32) < LINUX_VERSION_CODE)
static int syn1588_get_sset_count(struct net_device *, int sset);
#endif

static int syn1588_get_ts_info(struct net_device *ndev,
				   struct ethtool_ts_info *ts_info);
/*@}*/



#if (KERNEL_VERSION(5, 17, 0) <= LINUX_VERSION_CODE) || defined(RHEL_BACKPORT_ETHTOOL_RINGPARAM)
static void syn1588_get_ringparam(struct net_device *ndev,
				struct ethtool_ringparam *param,
				struct kernel_ethtool_ringparam *kernel_param,
				struct netlink_ext_ack *netlink_ext_ack)
{
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	(void)kernel_param;
	(void)netlink_ext_ack;
#else
static void syn1588_get_ringparam(struct net_device *ndev,
				  struct ethtool_ringparam *param)
{
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
#endif

	dev_dbg(dev, "*** syn1588_get_ringparam ***\n");
	param->rx_max_pending = MAC_TOTAL_BD / 2;
	//	u32 rx_mini_max_pending;
	//	u32 rx_jumbo_max_pending;
	param->tx_max_pending = MAC_TOTAL_BD / 2;

	param->rx_pending = nic_dat->bd_cnt;
	//	u32 rx_mini_pending;
	//	u32 rx_jumbo_pending;
	param->tx_pending = nic_dat->bd_cnt;
}


static int oregano_mac_get_module_info(struct net_device *ndev,struct ethtool_modinfo *modinfo);
static int oregano_mac_get_eeprom(struct net_device *ndev, struct ethtool_eeprom *cmd, u8 *buf);

static int oregano_mac_get_module_info(struct net_device *ndev,struct ethtool_modinfo *modinfo)
{
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	if(modinfo->cmd != ETHTOOL_GMODULEINFO) {
		dev_err(dev, "unknown command %d\n", modinfo->cmd);
		return -ENOTSUPP;
	}

	//TODO check if sfp mode
	if(nic_dat->mac_capabilities.i2c_interface) {
		if(oregano_mac_sfp_is_sff8472_compliance(&nic_dat->sfp)) {
			modinfo->type = ETH_MODULE_SFF_8472;
			modinfo->eeprom_len = ETH_MODULE_SFF_8472_LEN;
		}else {
			modinfo->type = ETH_MODULE_SFF_8079;
			modinfo->eeprom_len = ETH_MODULE_SFF_8079_LEN;
		}

	}

	return 0;
}

static int oregano_mac_get_eeprom(struct net_device *ndev, struct ethtool_eeprom *cmd, u8 *buf) {
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	u32 max_length = sizeof(nic_dat->sfp.eeprom) - cmd->offset;
	const u8 *eeprom_raw = oregano_mac_sfp_get_eeprom_raw(&nic_dat->sfp);

	if(cmd->cmd != ETHTOOL_GMODULEEEPROM) {
		dev_err(dev, "unknown command %d\n", cmd->cmd);
		return -ENOTSUPP;
	}
	memcpy(buf, eeprom_raw + cmd->offset, min(cmd->len, max_length));

	return 0;
}

#if (KERNEL_VERSION(4, 6, 0) > LINUX_VERSION_CODE)
static int syn1588_get_settings(struct net_device *ndev, struct ethtool_cmd *cmd)
{
	if(nic_phydev(ndev) == NULL) {
		return -ENODEV;
	}
	return phy_ethtool_gset(nic_phydev(ndev), cmd);
}
static int syn1588_set_settings(struct net_device *ndev, struct ethtool_cmd *cmd)
{
	if(nic_phydev(ndev) == NULL) {
		return -ENODEV;
	}
	return phy_ethtool_sset(nic_phydev(ndev), cmd);
}
#endif


/// Set functions for supported ETHTOOL operations
static struct ethtool_ops syn1588_ethtool_ops = {
#if (KERNEL_VERSION(4, 6, 0) <= LINUX_VERSION_CODE)
	.get_link_ksettings = phy_ethtool_get_link_ksettings,
	.set_link_ksettings = phy_ethtool_set_link_ksettings,
#else
	.get_settings = syn1588_get_settings,
	.set_settings = syn1588_set_settings,
#endif
	.get_drvinfo = syn1588_get_drvinfo,
	.get_regs_len = syn1588_get_regs_len,
	.get_regs = syn1588_get_regs,
	//	void (*get_wol)(struct net_device *, struct ethtool_wolinfo *);	// HW
	//	support currently N/A
	//	int (*set_wol)(struct net_device *, struct ethtool_wolinfo *);	 // HW
	//	support currently N/A
	.get_msglevel = syn1588_get_msglevel,
	.set_msglevel = syn1588_set_msglevel,
	.nway_reset = phy_ethtool_nway_reset,
	.get_link = ethtool_op_get_link,
	.get_module_info = oregano_mac_get_module_info,
	.get_module_eeprom = oregano_mac_get_eeprom,
	//	int (*get_eeprom_len)(struct net_device *);
	//	int (*get_eeprom)(struct net_device *, struct ethtool_eeprom *, u8 *);
	//	int (*set_eeprom)(struct net_device *, struct ethtool_eeprom *, u8 *);
	//	int (*get_coalesce)(struct net_device *, struct ethtool_coalesce *);
	//	int (*set_coalesce)(struct net_device *, struct ethtool_coalesce *);
	.get_ringparam = syn1588_get_ringparam,
	//	int (*set_ringparam)(struct net_device *, struct ethtool_ringparam *);
	//	.get_pauseparam = syn1588_get_pauseparam,
	//	int (*set_pauseparam)(struct net_device *, struct ethtool_pauseparam*);
	//	u32 (*get_rx_csum)(struct net_device *);	// HW support currently N/A
	//	int (*set_rx_csum)(struct net_device *, u32);	// HW support currently N/A
	//	u32 (*get_tx_csum)(struct net_device *);	// HW support currently N/A
	//	int (*set_tx_csum)(struct net_device *, u32);	// HW support currently N/A
	//	u32 (*get_sg)(struct net_device *);		// HW support currently N/A
	//	int (*set_sg)(struct net_device *, u32);	// HW support currently N/A
	//	u32 (*get_tso)(struct net_device *);		// HW support currently N/A
	//	int (*set_tso)(struct net_device *, u32);	// HW support currently N/A
	//.self_test = syn1588_self_test,
	.get_strings = syn1588_get_strings,
	//	int (*phys_id)(struct net_device *, u32);
	.get_ethtool_stats = syn1588_get_ethtool_stats,

#if (KERNEL_VERSION(2, 6, 32) < LINUX_VERSION_CODE)
	.get_sset_count = syn1588_get_sset_count,
#else
	//.self_test_count = syn1588_self_test_count,
	.get_stats_count = syn1588_get_stats_count,
#endif
	.begin = syn1588_begin,
	.complete = syn1588_complete,
//	u32 (*get_ufo)(struct net_device *);					 // HW support currently N/A
//	int (*set_ufo)(struct net_device *, u32);			// HW support currently N/A
#if (KERNEL_VERSION(4, 4, 0) < LINUX_VERSION_CODE) || defined(RHEL7_BACKPORT_COMPAT)
	.get_ts_info = syn1588_get_ts_info
#endif
};


#ifdef RHEL6_ETHTOOL_OPS_EXT_STRUCT
static struct ethtool_ops_ext syn1588_ethtool_ops_ext = {
	.size		= sizeof(struct ethtool_ops_ext),
	.get_ts_info	= syn1588_get_ts_info
};
#endif

static const char syn1588_gstrings_test[][ETH_GSTRING_LEN] = {
	"Register test	(offline)"
};

/// number of self-tests to be run
#define SYN1588_TEST_LEN sizeof(syn1588_gstrings_test) / ETH_GSTRING_LEN

void set_mac_phy_delay(struct net_device *nic, u32 tx_delay, u32 rx_delay)
{
	u32 reg = 0;
	struct nic_priv_data *nic_dat = netdev_priv(nic);
	struct device *dev = &nic->dev;

	// Set the MAC PHY registers
	reg = tx_delay;
	nic_writel(nic_dat, reg, MAC_TS_PHY_DLY_TX);
	if (reg != nic_readl(nic_dat, MAC_TS_PHY_DLY_TX))
		dev_warn(dev, "MAC PHY delay TX setting failed!\n");

	reg = rx_delay;
	nic_writel(nic_dat, reg, MAC_TS_PHY_DLY_RX);
	if (reg != nic_readl(nic_dat, MAC_TS_PHY_DLY_RX))
		dev_warn(dev, "MAC PHY delay RX setting failed!\n");
}

// prototype for get_phy_oui
void get_phy_oui(struct net_device *ndev, u32 *phy_oui1, u32 *phy_oui2);




/**
 * Write PHY delay values to the PHY delay registers
 * this is called on media change
 */
void set_phy_delays(struct net_device *ndev, const struct phy_device *phy)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	syn1588_phydelay_t p = {};
	struct clock_m_device *clock_m = dev_get_clock_m(nic_dat->clock_m);


	if(linkmode_test_bit_compat(ETHTOOL_LINK_MODE_10000baseSR_Full_BIT, phy->supported) && phy->speed == 10000) {
		p = nic_dat->phydelay_getter(ETHTOOL_LINK_MODE_10000baseSR_Full_BIT);

		pr_info("set phy delay for SGMII\n");
		set_mac_phy_delay(ndev, p.mac.tx, p.mac.rx);
		if(clock_m) {
			set_mcore_phy_delay(clock_m, p.clock_m.tx, p.clock_m.rx);
		}
		return;
	}
	if(linkmode_test_bit_compat(ETHTOOL_LINK_MODE_100baseT_Full_BIT, phy->supported) && phy->speed == 100) {
		p = nic_dat->phydelay_getter(ETHTOOL_LINK_MODE_100baseT_Full_BIT);

		pr_info("set phy delay for MII\n");
		set_mac_phy_delay(ndev, p.mac.tx, p.mac.rx);
		if(clock_m) {
			set_mcore_phy_delay(clock_m, p.clock_m.tx, p.clock_m.rx);
		}
		return;
	}

	if(linkmode_test_bit_compat(ETHTOOL_LINK_MODE_1000baseT_Full_BIT, phy->supported) && phy->speed == 1000) {
		p = nic_dat->phydelay_getter(ETHTOOL_LINK_MODE_1000baseT_Full_BIT);
		pr_info("set phy delay for GMII copper\n");
	} else if(linkmode_test_bit_compat(ETHTOOL_LINK_MODE_1000baseX_Full_BIT, phy->supported) && phy->speed == 1000) {
		p = nic_dat->phydelay_getter(ETHTOOL_LINK_MODE_1000baseX_Full_BIT);
		pr_info("set phy delay for GMII fiber\n");
	} else {
		pr_err("could not determit link delay");
		return;
	}

	set_mac_phy_delay(ndev, p.mac.tx << 16, p.mac.rx << 16);
	if(clock_m) {
		set_mcore_phy_delay(clock_m, p.clock_m.tx << 16, p.clock_m.rx << 16);
	}

}

/**
 * <!--********************************************************************************-->
 * \defgroup network_functions functions required for Ethernet network operation
 * <!--********************************************************************************-->
 */

/**
 * close network device
 * disable all MAC interrupts, turn off RX/TX, and release DMA buffers
 * \param ndev pointer to network device structure to be closed
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
static int nic_dev_close(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	struct phy_device *phydev = nic_phydev(ndev);

	dev_dbg(dev, "*** nic_dev_close ***\n");

	// remove watchdog
	if (del_timer_sync(&nic_dat->watchdog))
		dev_dbg(dev, "NIC_DEV_CLOSE: Had to wait for NIC watchdog.\n");

	// stop network subsystem
	netif_stop_queue(ndev);

	// disable MAC
	if (!nic_writel_check(nic_dat, 0x00, MAC_MODE))
		dev_err(dev, "NIC_DEV_CLOSE: Error disabling MAC!\n");
	// disable all interrupts from MAC
	if (!nic_writel_check(nic_dat, 0x0, MAC_INT_MASK))
		dev_err(dev, "NIC_DEV_CLOSE: Error disabling MAC IRQs!\n");
	dev_dbg(dev, "NIC_DEV_CLOSE: completed!\n");

	if(!IS_ERR_OR_NULL(phydev)) {
		phy_suspend_compat(phydev);
	}

	return 0;
}


/**
 * open network device
 * configure MAC, add timer for watchdog, enable
 * network subsystem
 * // create DMA pool and buffers is outsourced due to dual nic compatiblity
 * \param ndev pointer to network device structure to be opened
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
static int nic_dev_open(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	struct phy_device *phydev = nic_phydev(ndev);

	dev_dbg(dev, "*** nic_dev_open ***\n");

	if (!is_valid_ether_addr(ndev->dev_addr)) {
		dev_info(dev, "NIC_DEV_OPEN: Invalid MAC address on interface %s, not activated!\n",
			 ndev->name);
		return -EFAULT;
	}

	if (nic_dat->bd_cnt <= 0 || nic_dat->bd_cnt * 2 > MAC_TOTAL_BD) {
		dev_err(dev, "NIC_DEV_OPEN: DMA buffer descriptor count out of range!\n");
		return -EINVAL;
	}

	// initialize DMA buffer rings
	nic_dat->rx_dma_bd_rp = 0; // reset RX ring read pointer
	nic_dat->tx_dma_bd_rp = 0; // reset TX ring read pointer
	nic_dat->tx_dma_bd_wp = 0; // reset TX ring write pointer

	dev_info(dev, "NIC_DEV_OPEN:Configuring MAC for interface %s.\n",
		 ndev->name);
	// set number of buffer descriptors
	if (nic_setReg(ndev, MAC_BD_NUM, nic_dat->bd_cnt, 1)) {
		dev_err(dev,
			 "NIC_DEV_OPEN: Error setting buffer descriptor count!\n");
		(void)nic_dev_close(ndev);
		return -EIO;
	}

	if(phydev) {
		phy_resume_compat(phydev);
	}

	// configure MAC
	if (nic_dev_setconfig(ndev)) {
		dev_err(dev, "NIC_DEV_OPEN: Error setting initial MAC mode!\n");
		(void)nic_dev_close(ndev);
		return -EIO;
	}

	// enable all interrupts from MAC // TODO: control frames / dropped frames
	if (nic_setReg(ndev, MAC_INT_MASK,
			   MAC_INT_TXB | MAC_INT_TXE | MAC_INT_RXF | MAC_INT_RXE,
			   1)) {
		// MAC_INT_BUSY | MAC_INT_TXC | MAC_INT_RXC); // TODO: implement remaining
		// IRQs
		dev_err(dev, "NIC_DEV_OPEN: Error setting MAC interrupt mask!\n");
		(void)nic_dev_close(ndev);
		return -EIO;
	}

	// start watchdog checking linkstate
	nic_dat->watchdog.expires = jiffies + LINK_UPDATE;
	add_timer(&nic_dat->watchdog);

	// initalize VLAN registers
	if (nic_dat->mac_capabilities.vlan_group_implemented) {
		nic_writel(nic_dat, 0, MAC_VLAN_GRP_ADR);
	}
	nic_writel(nic_dat, 0, MAC_VLAN0);
	nic_writel(nic_dat, 0, MAC_VLAN1);
	nic_writel(nic_dat, 0, MAC_VLAN2);
	nic_writel(nic_dat, 0, MAC_VLAN3);

	netif_start_queue(ndev);
	return 0;
}

/**
 * transmit packet via network device
 * \param skb pointer to socket buffer (containing packet data)
 * \param ndev pointer to network device structure to be closed
 * \return NETDEV_TX_OK on success, NETDEV_TX_BUSY if buffer is full, -EMSGSIZE
 * on wrong packet size
 * \ingroup network_functions
 */
static int nic_dev_start_xmit(struct sk_buff *skb, struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	struct bd_addr *cur_buf;

	volatile u32 bd_stats; // buffer descriptor status to be set

	dev_dbg(dev, "*** nic_dev_start_xmit ***\n");

	if (skb->len > DMA_BD_SIZE) {
		dev_notice(dev,
			 "dropping data: packet is too large for transmit queue!\n");
		return -EMSGSIZE;
	}

	spin_lock_bh(&nic_dat->tx_dma_bdp_lock);
	if (nic_dat->tx_dma_bd_rp == TX_BDP_INC(nic_dat->tx_dma_bd_wp)) {
		spin_unlock_bh(&nic_dat->tx_dma_bdp_lock);
		dev_notice(dev, "cannot write to transmit DMA buffer ring: full!\n");
		nic_dat->extra_stats.drv_tx_qfull++;
		return NETDEV_TX_BUSY;
	}

	// get pointer to current buffer descriptor
	cur_buf = &nic_dat->tx_dma_buffer[nic_dat->tx_dma_bd_wp];

	// copy socket buffer data to DMA ring
	memcpy((unsigned char *)cur_buf->vaddr, skb->data, skb->len);

	// read buffer descriptor status
	bd_stats = nic_get_bd_stat_tx(nic_dat, nic_dat->tx_dma_bd_wp);
	// LOGN(TX_QUEUED, DEBUG, "TX BD (%u) stats read: 0x%X.\n",
	// nic_dat->tx_dma_bd_wp, bd_stats);

	bd_stats = MAC_TX_BD_READY |
		   MAC_TX_BD_IRQ; // Note: CRC is enabled globally

	if (skb->len < ETH_ZLEN) {
		dev_dbg(dev, "transmitting short packet. (%u bytes)\n", skb->len);
	}

	dev_dbg(dev, "TX BD (%u) stats set to 0x%X.\n",
		 nic_dat->tx_dma_bd_wp, (skb->len) << 16 | bd_stats);

	// If the current packet gets written to the last free TX BD, stop the
	// system TX queue from sending new packets to the driver.
	//
	// WARNING: The system TX queue has to be stopped BEFORE writing the last
	// TX BD. If the system TX queue is stopped AFTER writing the last TX BD,
	// we have a race condition with the TX_BHH. This could lead to a deadlock
	// of the system TX queue.
	if (nic_dat->tx_dma_bd_rp == TX_BDP_2INC(nic_dat->tx_dma_bd_wp)) {
		int count;

		// next time we will have to drop packet (FIFO is now full) => stop queue
		dev_warn(dev,  "HW transmit queue full! TXWP: %i, TXRP: %i\n",
			 nic_dat->tx_dma_bd_wp, nic_dat->tx_dma_bd_rp);
		for (count = 0; count < nic_dat->bd_cnt; count++) {
			dev_dbg(dev,
				 "#%02u TXBD: 0x%08X, RXBD: 0x%08X\n", count,
				 nic_get_bd_stat_tx(nic_dat, count),
				 nic_get_bd_stat_rx(nic_dat, count));
		}
#ifdef DEBUG_BHH_TX_QUEUE
		syn1588_ktime_get_real_ts(&nic_dat->netif_queue_stoptime);
#endif
		netif_stop_queue(ndev);
	}

	// write length and status (enable transmission by MAC)
	nic_set_bd_stat_tx(nic_dat, nic_dat->tx_dma_bd_wp, (skb->len) << 16 | bd_stats);

	if ((nic_dat->tstamp_config.tx_type == HWTSTAMP_TX_ONESTEP_SYNC || nic_dat->tstamp_config.tx_type == HWTSTAMP_TX_ON) &&
#if (KERNEL_VERSION(2, 6, 37) <= LINUX_VERSION_CODE)
		(skb_shinfo(skb)->tx_flags & SKBTX_HW_TSTAMP)) {

		skb_shinfo(skb)->tx_flags |= SKBTX_IN_PROGRESS;
#else
		// for kernel 2.6.30 to 2.6.36
		unlikely(skb_tx(skb)->hardware)) {

		skb_tx(skb)->in_progress = 1;
#endif
		nic_dat->tx_skb_buffer[nic_dat->tx_dma_bd_wp] =
			skb; // remember associated socket buffer
	} else {
		nic_dat->tx_skb_buffer[nic_dat->tx_dma_bd_wp] = 0x0;
		dev_kfree_skb(skb); // free associate socket buffer
	}

	// increment buffer descriptor write pointer
	nic_dat->tx_dma_bd_wp = TX_BDP_INC(nic_dat->tx_dma_bd_wp);
	spin_unlock_bh(&nic_dat->tx_dma_bdp_lock);

#if (KERNEL_VERSION(4, 7, 0) <= LINUX_VERSION_CODE)
	netif_trans_update(ndev); // store time of transmission start
#elif (KERNEL_VERSION(2, 7, 0) <= LINUX_VERSION_CODE)
	netdev_get_tx_queue(ndev, 0)->trans_start = jiffies;
#else
	ndev->trans_start = jiffies; // store time of transmission start
#endif

	return NETDEV_TX_OK;
}

/**
 * This function provides information about the device statistics
 * \param ndev pointer to network device structure to set multicast list for
 * \return pointer to network device statistics
 * \ingroup network_functions
 */
struct net_device_stats *nic_dev_get_stats(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;

	dev_dbg(dev, "*** nic_dev_get_stats ***\n");
	return &nic_dat->stats;
}
/**
 * This function sets the MAC address of the device
 * \param ndev pointer to network device structure to set multicast list for
 * \param p_addr pointer to socket address to be used by the adapter
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
static int oregano_mac_dev_set_mac_address(struct net_device *ndev, const u8 *macaddr)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;

	dev_dbg(dev, "*** nic_dev_set_mac_address ***\n");

	if (netif_running(ndev))
		return -EBUSY;

	if (!is_valid_ether_addr(macaddr)) {
		dev_warn(dev,
			 "supplied MAC address is not valid, keeping old one.\n");
		return -EFAULT; // return bad address
	}

	// write MAC address to MAC and set SW override bit
	if (!nic_writel_check(nic_dat, macaddr[0] << 8 | macaddr[1] | MAC_ETH_ADDR1_SW,
			   MAC_ETH_ADDR1) ||
		!nic_writel_check(nic_dat, macaddr[2] << 24 | macaddr[3] << 16 | macaddr[4] << 8 |
				   macaddr[5],
			   MAC_ETH_ADDR0)) {
		dev_err(dev, "error writing MAC address to register!\n");
		return -EIO;
	}
#if (KERNEL_VERSION(5, 15, 0) <= LINUX_VERSION_CODE)
	eth_hw_addr_set(ndev, macaddr);
#else
	memcpy(ndev->dev_addr, macaddr, ndev->addr_len);
#endif
	dev_info(dev,
		 "overriding HW MAC address with %02X:%02X:%02X:%02X:%02X:%02X.\n",
		 macaddr[0], macaddr[1], macaddr[2], macaddr[3], macaddr[4],
		 macaddr[5]);
	return 0;
}

static int nic_dev_set_mac_address(struct net_device *ndev, void *p_addr)
{
	struct device *dev = &ndev->dev;
	struct sockaddr *addr = p_addr;

	if (p_addr == NULL) {
		dev_warn(dev,
			 "cannot access supplied MAC address, pointer invalid!\n");
		return -EFAULT; // return bad address
	}
	return oregano_mac_dev_set_mac_address(ndev, addr->sa_data);
}

/**
 * This function sets the MAC address filtering by the hardware.
 * Supported MAC address filtering: promiscuous (all packets), all-multi (all
 * multicast packets),
 * hash filtering of 32 multicast address ranges, (none) only packets matching
 * MAC address pass
 * \param ndev pointer to network device structure to set multicast list for
 * \ingroup network_functions
 */
void nic_dev_set_multicast_list(struct net_device *ndev)
{
	struct device *dev = &ndev->dev;
	u32 mac_mode;
	u32 hash0_adr = 0x0;
	u32 hash1_adr = 0x0;
	int ret = 0;

	dev_dbg(dev, "*** nic_dev_set_multicast_list ***\n");

	// read current MAC mode
	if (nic_getReg(ndev, MAC_MODE, &mac_mode)) {
		dev_err(dev, "error getting MAC mode!\n");
		return;
	}

	// if kernel set promiscuous flag, enable corresponding MAC mode, and clear
	// mutlicast filter
	if (ndev->flags & IFF_PROMISC) {
		// switch to promiscuous mode (catch all packets)
		if (nic_setReg(ndev, MAC_MODE, mac_mode | MAC_MODE_PRO, 1))
			dev_err(dev, "error setting promiscous mode!\n");
		// goto out;
	} else {
		// promiscuous flag is not set, so clear this bit
		if (nic_setReg(ndev, MAC_MODE, mac_mode & ~MAC_MODE_PRO, 1))
			dev_err(dev, "error clearing promiscous mode!\n");
	}

	if (ndev->flags & IFF_ALLMULTI) {
		// Catch all multicast addresses, so set the filter to all 1's.
		hash0_adr = 0xFFFFFFFF;
		hash1_adr = 0xFFFFFFFF;
		goto out;
	}

	/*
	 * ndev->mc_count and ndev->mc_list are gone in version 2.6.35 of the kernel
	 *
	 *
	 * in kernel version 2.6.35 there is a file net/core/dev_addr_lists.c with
	 * helper functions to add and remove multicast addresses
	 */
#if (KERNEL_VERSION(2, 6, 34) < LINUX_VERSION_CODE)
	if (netdev_mc_count(ndev)) {
		unsigned long hash; // hash value calculated from address
		struct netdev_hw_addr *ha;

		netdev_for_each_mc_addr(ha, ndev) {
			u_char *addrs = ha->addr;

			dev_dbg(dev,
				 "DMI addr: %02X:%02X:%02X:%02X:%02X:%02X\n",
				 addrs[0], addrs[1], addrs[2], addrs[3], addrs[4],
				 addrs[5]);

			// check if address is multicast
			if (!(addrs[0] & 1)) {
				dev_warn(dev,
					 "unsupported unicast address for filtering!\n");
				continue;
			}
			hash = (ether_crc(ETH_ALEN, addrs)) >> 26;

			dev_dbg(dev, "DMI crc: 0x%X, hash: 0x%lx\n",
				 (ether_crc(ETH_ALEN, addrs)), hash);

			if (hash >= 32)
				hash1_adr |= 1 << (hash - 32);
			else
				hash0_adr |= 1 << hash;
		}
	}
#else
	if (ndev->mc_count) {
		// add the addresses in the list
		struct dev_mc_list *mc_list =
			ndev->mc_list; // pointer to multicast list of the device
		unsigned long hash; // hash value calculated from address
		int count; // loop counter

		for (count = 0; count < ndev->mc_count; count++) {
			dev_dbg(dev,
				 "DMI addr #%d: %02X:%02X:%02X:%02X:%02X:%02X\n",
				 count, mc_list->dmi_addr[0], mc_list->dmi_addr[1],
				 mc_list->dmi_addr[2], mc_list->dmi_addr[3],
				 mc_list->dmi_addr[4], mc_list->dmi_addr[5]);

			// check if address is multicast
			if (!(mc_list->dmi_addr[0] & 1)) {
				dev_warn(dev,
					 "unsupported unicast address for filtering!\n");
				continue;
			}
			hash = (ether_crc(ETH_ALEN, mc_list->dmi_addr)) >> 26;

			dev_dbg(dev, "DMI crc: 0x%X, hash: 0x%ld\n",
				 (ether_crc(ETH_ALEN, mc_list->dmi_addr)), hash);

			if (hash >= 32)
				hash1_adr |= 1 << (hash - 32);
			else
				hash0_adr |= 1 << hash;

			mc_list = mc_list->next;
		}
	}
#endif

out:
	// store hash values to MAC
	ret |= nic_setReg(ndev, MAC_ETH_HASH0_ADR, hash0_adr, 1);
	ret |= nic_setReg(ndev, MAC_ETH_HASH1_ADR, hash1_adr, 1);
	if (ret)
		dev_err(dev, "error setting individual address filtering!\n");
}

/**
 * called by the networking code when a packet transmission fails to complete
 * within a reasonable period (TX_TIMEOUT)
 * \param ndev pointer to network device structure of failed device
 * \ingroup network_functions
 */
void
#if (KERNEL_VERSION(5, 6, 0) <= LINUX_VERSION_CODE)
nic_dev_tx_timeout(struct net_device *ndev, unsigned int timeout)
#elif defined(RHEL8_BACKPORT_NDO_TX_TIMEOUT) || defined(SUSE15SP2_BACKPORT_NDO_TX_TIMEOUT)
nic_dev_tx_timeout(struct net_device *ndev, unsigned int timeout)
#else
nic_dev_tx_timeout(struct net_device *ndev)
#endif
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	int count;

	// TODO: implement recover code
	dev_crit(dev, "network layer reported packet transmission timed out!\n");
	for (count = 0; count < nic_dat->bd_cnt; count++) {
		dev_crit(dev, "#%02u TXBD: 0x%08X, RXBD: 0x%08X\n", count,
			nic_get_bd_stat_tx(nic_dat, count),
			nic_get_bd_stat_rx(nic_dat, count));
	}
}

/**
 * <!--********************************************************************************-->
 * \defgroup mac_functions functions to access the MAC
 * <!--********************************************************************************-->
 * @{
 */


int nic_getReg(struct net_device *ndev, unsigned long address, u32 *data)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	*data = nic_readl(nic_dat, address);
	return 0;
}

/**
 * Function to write to a MAC register
 * \param ndev pointer to network device structure (device to write to)
 * \param address register address to read from
 * \param data to write to register specified by address
 * \param verify if unequal to zero, the written data is read back and checked
 * \return 0 on success, negative values if an error occured
 */
int nic_setReg(struct net_device *ndev, unsigned long address, u32 data,
		   int verify)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;

	dev_dbg(dev, "write to MAC address 0x%lX data: 0x%lX\n", address,
		 (unsigned long)data);

	nic_writel(nic_dat, data, address);

	if (verify && (nic_readl(nic_dat, address) != data)) {
		dev_err(dev,
			 "write verify on MAC address 0x%lX failed (0x%X)!\n",
			 address, nic_readl(nic_dat, address));
		return -EIO;
	}

	return 0;
}

/// @}

static int nic_dev_timestamping_ioctl_set_hw_ts(struct net_device *ndev,
					  struct ifreq *ifr)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct hwtstamp_config hwts_cfg;
	struct hwtstamp_config hwts_cfg_req;

	if (copy_from_user(&hwts_cfg, ifr->ifr_data, sizeof(hwts_cfg)))
		return -EFAULT;

	// /* no flags defined right now, must be zero */
	if (hwts_cfg.flags != 0)
		return -EINVAL;

	hwts_cfg_req = hwts_cfg;

	// check filter
	switch (hwts_cfg.tx_type) {
	case HWTSTAMP_TX_OFF:
		break;
	case HWTSTAMP_TX_ON:
		if( !nic_dat->mac_capabilities.ptp_timestamper &&
		    !nic_dat->mac_capabilities.tx_timestamper_implemented &&
		    !nic_dat->mac_capabilities.second_timestamper
		) {
			return -ERANGE;
		}
		break;
	case HWTSTAMP_TX_ONESTEP_SYNC:
		if( !nic_dat->mac_capabilities.ptp_timestamper ) {
			return -ERANGE;
		}
		break;
	default:
		return -ERANGE;
	}

	switch (hwts_cfg.rx_filter) {
	case HWTSTAMP_FILTER_NONE:
		break;
	case HWTSTAMP_FILTER_ALL:
		if(!nic_dat->mac_capabilities.rx_timestamper_implemented &&
		   !nic_dat->mac_capabilities.second_timestamper
		) {
			return -ERANGE;
		}
		break;
	case HWTSTAMP_FILTER_PTP_V2_L4_EVENT:
	case HWTSTAMP_FILTER_PTP_V2_L4_SYNC:
	case HWTSTAMP_FILTER_PTP_V2_L4_DELAY_REQ:
	case HWTSTAMP_FILTER_PTP_V2_L2_EVENT:
	case HWTSTAMP_FILTER_PTP_V2_L2_SYNC:
	case HWTSTAMP_FILTER_PTP_V2_L2_DELAY_REQ:
	case HWTSTAMP_FILTER_PTP_V2_EVENT:
	case HWTSTAMP_FILTER_PTP_V2_SYNC:
	case HWTSTAMP_FILTER_PTP_V2_DELAY_REQ:
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			hwts_cfg.rx_filter = HWTSTAMP_FILTER_PTP_V2_EVENT;
		} else if(
			nic_dat->mac_capabilities.rx_timestamper_implemented ||
			nic_dat->mac_capabilities.second_timestamper
		) {
			hwts_cfg.rx_filter = HWTSTAMP_FILTER_ALL;
		} else {
			return -ERANGE;
		}
		break;
	default:
		return -EINVAL;
	};

	// apply filter
	switch (hwts_cfg.tx_type) {
	case HWTSTAMP_TX_OFF:
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp & ~((u32)MAC_PTP_TS_CTRL_ENABLE_TX) ,MAC_PTP_TS_CTRL);
		}else if(nic_dat->mac_capabilities.tx_timestamper_implemented) {
			nic_writel(nic_dat, 0, MAC_TS_TXCTRL);
		}else if(nic_dat->mac_capabilities.second_timestamper) {
			nic_writel(nic_dat, 0, MAC_TS2_TXCTRL);
		} else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.tx_type = HWTSTAMP_TX_OFF;
		break;
	case HWTSTAMP_TX_ON:
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp | MAC_PTP_TS_CTRL_ENABLE_TX ,MAC_PTP_TS_CTRL);
		}else if(nic_dat->mac_capabilities.tx_timestamper_implemented) {
			nic_writel(nic_dat, 2, MAC_TS_TXCTRL);
			nic_writel(nic_dat, 0x00000000, MAC_TS_TXMEM_BASE);
		}else if(nic_dat->mac_capabilities.second_timestamper) {
			nic_writel(nic_dat, 2, MAC_TS2_TXCTRL);
			nic_writel(nic_dat, 0x00000000, MAC_TS2_TXMEM_BASE);
		} else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.tx_type = HWTSTAMP_TX_ON;
		break;
	case HWTSTAMP_TX_ONESTEP_SYNC:
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp | MAC_PTP_TS_CTRL_ENABLE_TX | MAC_PTP_TS_CTRL_ENABLE_TX_1STEP, MAC_PTP_TS_CTRL);
		} else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.tx_type = HWTSTAMP_TX_ONESTEP_SYNC;
		break;
	default:
		return -EFAULT;
	}

	switch (hwts_cfg.rx_filter) {
	case HWTSTAMP_FILTER_NONE:
		//reset filter_ptp
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp & ~((u32)MAC_PTP_TS_CTRL_ENABLE_RX), MAC_PTP_TS_CTRL);
		}
		//reset filter_all
		if(nic_dat->mac_capabilities.rx_timestamper_implemented) {
			nic_writel(nic_dat, 0, MAC_TS_RXCTRL);
			nic_writel(nic_dat, 0x0000, MAC_TS_RXMEM_BASE);
		}else if(nic_dat->mac_capabilities.second_timestamper) {
			nic_writel(nic_dat, 0, MAC_TS2_RXCTRL);
			nic_writel(nic_dat, 0x0000, MAC_TS2_RXMEM_BASE);
		}else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.rx_filter = HWTSTAMP_FILTER_NONE;
		break;

	case HWTSTAMP_FILTER_PTP_V2_EVENT:
		// deaktivate other timestamper
		if(nic_dat->mac_capabilities.rx_timestamper_implemented) {
			nic_writel(nic_dat, 0, MAC_TS_RXCTRL);
			nic_writel(nic_dat, 0x0000, MAC_TS_RXMEM_BASE);
		}else if(nic_dat->mac_capabilities.second_timestamper) {
			nic_writel(nic_dat, 0, MAC_TS2_RXCTRL);
			nic_writel(nic_dat, 0x0000, MAC_TS2_RXMEM_BASE);
		}

		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp | MAC_PTP_TS_CTRL_ENABLE_RX ,MAC_PTP_TS_CTRL);
		} else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.rx_filter = HWTSTAMP_FILTER_PTP_V2_EVENT;
		break;
	case HWTSTAMP_FILTER_ALL:
		// deaktivate other timestamper
		if(nic_dat->mac_capabilities.ptp_timestamper) {
			u32 tmp = nic_readl(nic_dat, MAC_PTP_TS_CTRL);
			nic_writel(nic_dat, tmp & ~((u32)MAC_PTP_TS_CTRL_ENABLE_RX), MAC_PTP_TS_CTRL);
		}

		if(nic_dat->mac_capabilities.rx_timestamper_implemented) {
			nic_writel(nic_dat, 2, MAC_TS_RXCTRL);
			nic_writel(nic_dat, 0x00000000, MAC_TS_RXMEM_BASE);
		}else if(nic_dat->mac_capabilities.second_timestamper) {
			nic_writel(nic_dat, 2, MAC_TS2_RXCTRL);
			nic_writel(nic_dat, 0x00000000, MAC_TS2_RXMEM_BASE);
		} else {
			return -EFAULT;
		}
		nic_dat->tstamp_config.rx_filter = HWTSTAMP_FILTER_ALL;
		break;
	default:
		return -EFAULT;
	}

	if(hwts_cfg_req.rx_filter != hwts_cfg.rx_filter) {
		hwts_cfg.rx_filter = HWTSTAMP_FILTER_SOME;
	}

	return copy_to_user(ifr->ifr_data, &hwts_cfg, sizeof(hwts_cfg)) ? -EFAULT : 0;
}

static int syn1588_get_ts_info(struct net_device *ndev,
				   struct ethtool_ts_info *ts_info)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct clock_m_device *clock_m = dev_get_clock_m(nic_dat->clock_m);

	if(clock_m == NULL) {
		//no clock use sw fallback
		return ethtool_op_get_ts_info(ndev, ts_info);
	}

	ts_info->so_timestamping =
		SOF_TIMESTAMPING_TX_SOFTWARE | SOF_TIMESTAMPING_RX_SOFTWARE |
		SOF_TIMESTAMPING_SOFTWARE | SOF_TIMESTAMPING_TX_HARDWARE |
		SOF_TIMESTAMPING_RX_HARDWARE | SOF_TIMESTAMPING_RAW_HARDWARE;
	ts_info->phc_index = -1;
	if (clock_m->ptp_clock)
		ts_info->phc_index = ptp_clock_index(clock_m->ptp_clock);

	ts_info->tx_types = BIT(HWTSTAMP_TX_OFF);
	if(nic_dat->mac_capabilities.tx_timestamper_implemented || nic_dat->mac_capabilities.second_timestamper) {
		ts_info->tx_types |= BIT(HWTSTAMP_TX_ON);
	}
	if(nic_dat->mac_capabilities.ptp_timestamper) {
		ts_info->tx_types |= BIT(HWTSTAMP_TX_ON) | BIT(HWTSTAMP_TX_ONESTEP_SYNC);
	}
	ts_info->rx_filters = BIT(HWTSTAMP_FILTER_NONE);
	if(nic_dat->mac_capabilities.tx_timestamper_implemented || nic_dat->mac_capabilities.second_timestamper) {
		ts_info->rx_filters |= BIT(HWTSTAMP_FILTER_ALL);
	}
	if(nic_dat->mac_capabilities.ptp_timestamper) {
		ts_info->rx_filters |= BIT(HWTSTAMP_FILTER_PTP_V2_EVENT);
	}
	return 0;
}

static int nic_dev_timestamping_ioctl_get_hw_ts(struct net_device *ndev,
				      struct ifreq *ifr)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct hwtstamp_config *config = &nic_dat->tstamp_config;

	return copy_to_user(ifr->ifr_data, config, sizeof(*config)) ?
		-EFAULT : 0;
}

static int nic_dev_ioctl(struct net_device *ndev, struct ifreq *ifr, int cmd)
{
	switch (cmd) {
	case SIOCGHWTSTAMP:
		return nic_dev_timestamping_ioctl_get_hw_ts(ndev, ifr);
	case SIOCSHWTSTAMP:
		return nic_dev_timestamping_ioctl_set_hw_ts(ndev, ifr);
	case SIOCGMIIPHY:
	case SIOCGMIIREG:
	case SIOCSMIIREG:
	default:
		return phy_do_ioctl(ndev, ifr, cmd);
	}
}


static void oregano_mac_phy_handler(struct net_device *ndev) {
	struct device *dev = &ndev->dev;

	if (nic_dev_setconfig(ndev)) {
		dev_err(dev, "phy_handler: Error setting initial MAC mode!\n");
		(void)nic_dev_close(ndev);
	}
}

static void oregano_mac_phy_handler_fiber_quirk(struct net_device *ndev) {
	//struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat =netdev_priv(ndev);
	struct phy_device *phy = nic_phydev(ndev);
	u32 miicommand_status = nic_readl(nic_dat, MAC_MIICOMMAND);
	bool is_fiber = (miicommand_status & MAC_MIICOMMAND_SFP_BASEX) > 0;

	if(is_fiber) {
		linkmode_mod_bit_compat(ETHTOOL_LINK_MODE_1000baseX_Full_BIT, phy->supported, 1);
	} else {
		linkmode_mod_bit_compat(ETHTOOL_LINK_MODE_1000baseX_Full_BIT, phy->supported, 0);
	}

	oregano_mac_phy_handler(ndev);
}

static void nic_dev_watchdog(struct work_struct *work) {

	struct nic_priv_data *nic_dat = container_of(work, struct nic_priv_data, work_watchdog);
	struct net_device *ndev = nic_dat->net_dev;
	struct device *dev = &ndev->dev;
	bool prev_link = netif_carrier_ok(ndev);

	if(nic_dat->mac_capabilities.i2c_interface) {

		u32 line_status = ioread32(nic_dat->iobase + MAC_LINE_STATUS);
		if((line_status & MAC_LINE_STATUS_PRESENT) == 0) {
			dev_dbg(dev, "no sfp module present");
		} else {
			dev_dbg(dev, "sfp module present");
		}
		if((line_status & MAC_LINE_STATUS_PRESENT) == 0 && oregano_mac_sfp_connected(&nic_dat->sfp)) {
			dev_info(dev, "remove sfp module");
			oregano_mac_sfp_exit_phy(&nic_dat->sfp);
		}
		if((line_status & MAC_LINE_STATUS_PRESENT) != 0 && !oregano_mac_sfp_connected(&nic_dat->sfp)) {
			int ret = 0;
			dev_info(dev, "sfp module inserted");
			// probe for sfp
			ret = oregano_mac_sfp_probe_phy(&nic_dat->sfp, oregano_mac_phy_handler);
			if(ret != 0) {
				dev_err(dev, "sfp probe error");
			}
		}

		oregano_mac_sfp_set_phylink(&nic_dat->sfp, (line_status & MAC_LINE_STATUS_LINK) != 0);
	}

	{
		bool cur_link =  ndev->phydev ? ndev->phydev->link : false;

		if (cur_link && !prev_link) {
			phy_print_status(ndev->phydev);
			netif_carrier_on(ndev);
		}
		else if (prev_link && !cur_link) {
			netif_carrier_off(ndev);
		}
	}
}

/**
 * Watchdog function for the MAC, checking media status
 * \param ndev_pt memory address of the netword device structure
 */
#if (KERNEL_VERSION(4, 15, 0) > LINUX_VERSION_CODE)
static void nic_dev_watchdog_timer(unsigned long ndev_pt)
{
	struct net_device *ndev =(struct net_device *)ndev_pt;
	struct nic_priv_data *nic_dat =netdev_priv(ndev);
#else
// timer api has been reworked, starting with kernel 4.15
// see https://lwn.net/Articles/735887/
static void nic_dev_watchdog_timer(struct timer_list *t)
{
	struct nic_priv_data *nic_dat = from_timer(nic_dat, t, watchdog);
	struct net_device *ndev = nic_dat->net_dev;
#endif
	struct device *dev = &ndev->dev;

	dev_dbg(dev, "*** syn1588_mac_watchdog ***\n");
	nic_dat->watchdog.expires = jiffies + LINK_UPDATE;
	add_timer(&nic_dat->watchdog);

	queue_work(nic_dat->workq, &nic_dat->work_watchdog);

}

/**
 * Configure MAC for correct operation parameters (speed, duplex)
 * \param ndev pointer to network device structure to set configuration for
 * \return 0 on success, negative error code otherwise
 * \ingroup network_functions
 */
int nic_dev_setconfig(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	struct phy_device *phy = nic_phydev(ndev);
	u32 mac_mode_val = MAC_MODE_PAD; // default value for MODE register

	dev_dbg(dev, "*** nic_dev_setconfig ***\n");
	mac_mode_val |= MAC_MODE_CRCEN; // enable HW CRC check

	if(IS_ERR_OR_NULL(phy)) {
		// phy not set
		dev_info(dev, "no phy link, disable mac\n");

		if(!nic_writel_check(nic_dat, mac_mode_val,
				MAC_MODE)){
			dev_err(dev, "error writing mac mode register\n");
		}
		return 0;
	}
	phy_print_status(phy);

	// disable RX and TX to be able to change duplex/GbE settings
	if (nic_setReg(ndev, MAC_MODE, mac_mode_val, 1)) {
		dev_err(dev, "error disabling RX/TX in MAC!\n");
		return -EIO;
	}

	// FIXME! stop network subsystem to be save
	// FIXME! This is a patch to compensate for internal MAC reset on disabling
	// RXEN/TXEN
	{
		int bd_initval, ret, count;

		netif_stop_queue(ndev);

		// Update the error counts.
		nic_dat->stats.tx_aborted_errors +=
			((int)nic_dat->tx_dma_bd_wp - nic_dat->tx_dma_bd_rp +
			nic_dat->bd_cnt) % nic_dat->bd_cnt;

		// note RX error for every filled buffer
		for (count = 0; count < nic_dat->bd_cnt; count++) {
			if (unlikely(!(nic_get_bd_stat_rx(nic_dat,count) & MAC_RX_BD_EMPTY)))
				nic_dat->stats.rx_missed_errors++;
		}


		nic_dat->rx_dma_bd_rp = 0; // reset RX ring read pointer
		nic_dat->tx_dma_bd_rp = 0; // reset TX ring read pointer
		nic_dat->tx_dma_bd_wp = 0; // reset TX ring write pointer
		// initialize DMA buffer rings
		dev_dbg(dev,
			 "initializing %u RX DMA buffers of size %u.\n",
			 nic_dat->bd_cnt, DMA_BD_SIZE);
		for (count = 0; count < nic_dat->bd_cnt; count++) {
			// initialize HW buffer descriptor
			bd_initval = MAC_RX_BD_EMPTY | MAC_RX_BD_IRQ;

			// write RX BD initial STATUS
			if (nic_setReg(ndev, RX_BD(count), bd_initval, 1)) {
				dev_err(dev,
					 "error setting initial RX BD status!\n");
				ret = -1;
			}

#ifdef DMA_BD_INIT
			memset(nic_dat->rx_dma_buffer[count].vaddr, DMA_BD_INIT,
				   DMA_BD_SIZE);
#endif
		}

		dev_dbg(dev,
			 "initializing %u TX DMA buffers of size %u.\n",
			 nic_dat->bd_cnt, DMA_BD_SIZE);
		for (count = 0; count < nic_dat->bd_cnt; count++) {
			// initialize HW buffer descriptor
			bd_initval = 0x0;

			// write TX BD initial STATUS
			nic_set_bd_stat_tx(nic_dat, count, bd_initval);
		}
	}

	if (phy->duplex == DUPLEX_FULL)
	{
		mac_mode_val |= MAC_MODE_FULLD;
		dev_dbg(dev, "full duplex mode enable.\n");
	}
	if(linkmode_test_bit_compat(ETHTOOL_LINK_MODE_1000baseT_Full_BIT, phy->supported)) {
		dev_dbg(dev, "baseT mode active\n");
		mac_mode_val |= MAC_MODE_1000BASE_T;
	} else {
		dev_dbg(dev, "baseX mode active\n");
		mac_mode_val |= MAC_MODE_1000BASE_X;
	}

	if (phy->speed == SPEED_1000)
	{
		mac_mode_val |= MAC_MODE_GBIT;
		dev_dbg(dev, "GBIT mode enable.\n");
	}

	// Check if the device has the hardware reset bit
	//
	pr_debug("Performing MAC SOFTWARE RESET\n");
	// Perform MAC SW Reset
	if(!nic_writel_check(nic_dat, mac_mode_val | MAC_MODE_SWRST,
			MAC_MODE)){
		pr_err("error performing software reset");
	}
	mdelay(10);
	if(!nic_writel_check(nic_dat, mac_mode_val,
			MAC_MODE)){
		pr_err("error performing software reset");
	}
	mdelay(10);

	set_phy_delays(ndev, phy);

	{ // if a VLAN is enabled - enable vlan bit in MAC mode register
		u32 regVlan0;
		u32 regVlan1;
		u32 regVlan2;
		u32 regVlan3;
		u32 regVlanGroup;

		if (!nic_dat->mac_capabilities.vlan_group_implemented) {
			nic_getReg(ndev, MAC_VLAN0, &regVlan0);
			nic_getReg(ndev, MAC_VLAN1, &regVlan1);
			nic_getReg(ndev, MAC_VLAN2, &regVlan2);
			nic_getReg(ndev, MAC_VLAN3, &regVlan3);

			if (MAC_VLAN_VLANEN &
				(regVlan3 | regVlan2 | regVlan1 | regVlan0)) {
				mac_mode_val |= MAC_MODE_VLANEN;
				dev_dbg(dev, "VLAN enable.\n");
			}
		} else {
			nic_getReg(ndev, MAC_VLAN_GRP_ADR, &regVlanGroup);

			if (regVlanGroup) {
				mac_mode_val |= MAC_MODE_VLANEN;
				dev_dbg(dev, "VLAN enable.\n");
			}
		}
	}
	// Disbale FCS forwarding in RX part of mac
	mac_mode_val |= MAC_MODE_SUPPRESS_FCS;

	dev_dbg(dev, " write new setting for MAC -> %x", mac_mode_val);
	// write new setting for MAC
	if (nic_setReg(ndev, MAC_MODE, mac_mode_val, 1)) {
		dev_err(dev, "error	MAC mode!\n");
		return -EIO;
	}

	nic_dev_set_multicast_list(ndev);

// enable RX and TX for MAC again
#ifdef JUMBO_PACKET
	if (nic_setReg(ndev, MAC_MODE,
			   mac_mode_val | MAC_MODE_RXEN | MAC_MODE_TXEN |
				   MAC_MODE_HUGEN,
			   1))
#else
	if (nic_setReg(ndev, MAC_MODE,
			   mac_mode_val | MAC_MODE_RXEN | MAC_MODE_TXEN, 1))
#endif
	{
		dev_err(dev, "error enabling RX/TX in MAC!\n");
		return -EIO;
	}
	netif_start_queue(ndev); // FIXME! stop network subsystem to be save

	dev_dbg(dev, "mac configured for phy mode\n");
	return 0;
}

#if (KERNEL_VERSION(3, 3, 0) > LINUX_VERSION_CODE)
static void nic_dev_vlan_rx_add_vid(struct net_device *ndev, unsigned short vid)
{
#elif (KERNEL_VERSION(3, 10, 0) > LINUX_VERSION_CODE)
static int nic_dev_vlan_rx_add_vid(struct net_device *ndev, unsigned short vid)
{
#else
static int nic_dev_vlan_rx_add_vid(struct net_device *ndev, __be16 proto,
				   unsigned short vid)
{
#endif
	u32 data;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;

	if(!nic_dat->mac_capabilities.vlan_group_implemented) {
		dev_err(dev, "vlan_group is not implemented");
#if (KERNEL_VERSION(3, 3, 0) <= LINUX_VERSION_CODE)
		return -ENODEV;
#else
		return;
#endif
	}

	spin_lock(&nic_dat->vlan_grp_bitmap_lock);
	set_bit(vid, nic_dat->vlan_grp_bitmap);
	spin_unlock(&nic_dat->vlan_grp_bitmap_lock);

	update_vlan_grp(ndev);

	// Enable VLAN processing in the MAC
	nic_getReg(ndev, MAC_MODE, &data);
	nic_writel(nic_dat, data | MAC_MODE_VLANEN, MAC_MODE);

	nic_getReg(ndev, MAC_MODE, &data);
	printk("MAC_MODE %x\n", data);
#if (KERNEL_VERSION(3, 3, 0) <= LINUX_VERSION_CODE)
	return 0;
#endif
}

#if (KERNEL_VERSION(3, 3, 0) > LINUX_VERSION_CODE)
static void nic_dev_vlan_rx_kill_vid(struct net_device *ndev,
					 unsigned short vid)
{
#elif (KERNEL_VERSION(3, 10, 0) > LINUX_VERSION_CODE)
static int nic_dev_vlan_rx_kill_vid(struct net_device *ndev, unsigned short vid)
{
#else
static int nic_dev_vlan_rx_kill_vid(struct net_device *ndev, __be16 proto,
					unsigned short vid)
{
#endif
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	u32 regVlanGroup;

	if(!nic_dat->mac_capabilities.vlan_group_implemented) {
		dev_err(&ndev->dev, "vlan_group is not implemented");
#if (KERNEL_VERSION(3, 3, 0) <= LINUX_VERSION_CODE)
		return -ENODEV;
#else
		return;
#endif
	}

	spin_lock(&nic_dat->vlan_grp_bitmap_lock);
	clear_bit(vid, nic_dat->vlan_grp_bitmap);
	spin_unlock(&nic_dat->vlan_grp_bitmap_lock);

	update_vlan_grp(ndev);

	nic_getReg(ndev, MAC_VLAN_GRP_ADR, &regVlanGroup);

	if (regVlanGroup == 0) {
		u32 data;
		// Disable VLAN processing in the MAC
		nic_getReg(ndev, MAC_MODE, &data);
		nic_writel(nic_dat, data & (~MAC_MODE_VLANEN), MAC_MODE);

		nic_getReg(ndev, MAC_MODE, &data);
		printk("MAC_MODE %x\n", data);
	}
#if (KERNEL_VERSION(3, 3, 0) <= LINUX_VERSION_CODE)
	return 0;
#endif
}



// preprocessor makro to extract the 5bit group id from the 12bit vlan id
// subsequently it is necessary to mask any higher nibble
unsigned int vlan_id2group(unsigned int id) {
	return (id >> 7) & 0x1f;
}


// function updates the vlan group register,
// iterates over the vlan group list
// and set the vlan_grp, which are the first 5 bit of the vlanID
// needs network device as parameter to access the NIC data
static void update_vlan_grp(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	u32 vlan_grp = 0;
	unsigned int bit;

	spin_lock(&nic_dat->vlan_grp_bitmap_lock);
	for_each_set_bit(bit, nic_dat->vlan_grp_bitmap, sizeof(nic_dat->vlan_grp_bitmap)*8) {
		vlan_grp |= 1 << vlan_id2group(bit);
	}
	spin_unlock(&nic_dat->vlan_grp_bitmap_lock);
	nic_writel(nic_dat, vlan_grp, MAC_VLAN_GRP_ADR);
	printk("Set MAC_VLAN_GRP_ADR to %08x\n", vlan_grp);
}

/// net device functions
static struct net_device_ops syn1588_net_device_ops = {
	//.ndo_init
	//.ndo_uninit
	.ndo_open = nic_dev_open,
	.ndo_stop = nic_dev_close,
	.ndo_start_xmit = nic_dev_start_xmit,
	//.ndo_select_queue
	//.ndo_change_rx_flags
#if (KERNEL_VERSION(3, 2, 0) <= LINUX_VERSION_CODE)
	.ndo_set_rx_mode = nic_dev_set_multicast_list,
#else
	.ndo_set_multicast_list = nic_dev_set_multicast_list,
#endif
	.ndo_set_mac_address = nic_dev_set_mac_address,
	//.ndo_validate_addr
#if (KERNEL_VERSION(5, 15, 0) <= LINUX_VERSION_CODE) || defined(RHEL9_BACKPORT_NDO_ETH_IOCTL)
	.ndo_eth_ioctl = nic_dev_ioctl,
#else
	.ndo_do_ioctl = nic_dev_ioctl,
#endif
	//.ndo_set_config
	//.ndo_change_mtu
	//.ndo_neigh_setup
	.ndo_tx_timeout = nic_dev_tx_timeout,
	.ndo_get_stats = nic_dev_get_stats,
	//.ndo_vlan_rx_register
	.ndo_vlan_rx_add_vid = nic_dev_vlan_rx_add_vid,
	.ndo_vlan_rx_kill_vid = nic_dev_vlan_rx_kill_vid,
	//.ndo_poll_controller
	//.ndo_fcoe_ddp_setup
	//.ndo_fcio_ddp_done
};

/**
 * Initialise MAC and register network device
 * \param adapter pointer to device structure
 * @param  clock_m nullable
 * \return 0 on success, negatice error code otherwise
 */
struct net_device *nic_dev_alloc(struct device *parent_device, void __iomem *iobase, struct syn1588_dma *dma,int irq, struct oregano_mac_platform_data *pdata, syn1588_phydelay_getter_t phydelay_getter)
{
	struct net_device *net_dev = NULL;
	struct nic_priv_data *nic_dat = NULL;
	struct device *dev = NULL;
	int ret;
	u32 burst_len = BURST_LEN_PCI;
	int startoff = 0;
	u32 bd_initval;
	int count;

	pr_debug("*** nic_dev_init ***\n");

	// allocate private data structure for network device
	net_dev = alloc_etherdev(sizeof(struct nic_priv_data));
	if (net_dev == NULL) {
		pr_err("failed to allocate NIC data structure.\n");
		return ERR_PTR(-ENOMEM);
	}
	nic_dat = netdev_priv(net_dev);

	nic_dat->iobase = iobase;
	nic_dat->clock_m = pdata->clock_m;
	nic_dat->net_dev = net_dev;
	nic_dat->phydelay_getter = phydelay_getter;
	//nic_dat->adapter = adapter; // store pointer to card
	sema_init(&nic_dat->mii_access, 1); // initialize MII access lock
	spin_lock_init(&nic_dat->tx_dma_bdp_lock); // initialize lock
	SET_NETDEV_DEV(net_dev, parent_device);
	nic_dat->bd_cnt = BUFF_CNT; // initialize number of DMA buffers per direction


	dev = &net_dev->dev;

	syn1588_nic_read_capabilities(net_dev);

	syn1588_nic_print_capabilities(net_dev);

	{ // read version information from MAC
		u32 version = nic_readl(nic_dat, MAC_ETH_VERSION);;

		nic_dat->version[0] = (version >> 24) & 0xff;
		nic_dat->version[1] = (version >> 16) & 0xff;
		nic_dat->version[2] = (version >> 8) & 0xff;
		nic_dat->version[3] = (version >> 0) & 0xff;
	}


	// set PCI burst length for DMA transfers
	if (nic_setReg(net_dev, MAC_BURST_LENGTH, burst_len, 1)) {
		pr_err("error setting PCIe DMA burst length!\n");
		nic_dev_exit(net_dev);
		return ERR_PTR(-EIO);
	}
	pr_debug("Using burst lenght: %d\n", burst_len);

	nic_dat->msg_enable = netif_msg_init(/*debug*/ 0, NETIF_MSG_DRV | NETIF_MSG_PROBE);

// setup watchdog timer checking linkstate
#if (KERNEL_VERSION(4, 15, 0) > LINUX_VERSION_CODE)
	setup_timer(&nic_dat->watchdog, nic_dev_watchdog_timer,
			(unsigned long)net_dev);
#else
	// see comments at function nic_dev_watchdog_timer
	timer_setup(&nic_dat->watchdog, nic_dev_watchdog_timer, 0);
#endif

// hw bcast address set by alloc_etherdev
// see netdevice.h for details about features
#if (KERNEL_VERSION(3, 10, 0) > LINUX_VERSION_CODE)
	net_dev->features = NETIF_F_HW_VLAN_FILTER;
#else
	net_dev->features = NETIF_F_HW_VLAN_CTAG_FILTER;
#endif

	net_dev->netdev_ops = &syn1588_net_device_ops;
	net_dev->ethtool_ops = &syn1588_ethtool_ops;

#ifdef RHEL6_ETHTOOL_OPS_EXT_STRUCT
	dev_dbg(dev, "set ethtool_ops_ext");
	set_ethtool_ops_ext(net_dev, &syn1588_ethtool_ops_ext);
#endif

	strncpy(net_dev->name, dev_name(parent_device),
		sizeof(net_dev->name) - 1);
	{
		u32 temp;
		u32 temp_macaddr;
		unsigned char *macaddr_ptr = (unsigned char *)&net_dev->perm_addr;

		// readout first part of HW MAC address
		temp = nic_readl(nic_dat, MAC_ETH_ADDR1);

		if (temp & MAC_ETH_ADDR1_SW) {
			// clear SW overriding flag for MAC address
			temp &= ~MAC_ETH_ADDR1_SW;
			nic_writel(nic_dat, temp, MAC_ETH_ADDR1);
			// reread HW MAC address
			temp = nic_readl(nic_dat, MAC_ETH_ADDR1);
		}

		// readout second part HW MAC address
		temp_macaddr = __be16_to_cpu(temp & 0x0000FFFF);
		memcpy(net_dev->perm_addr, &temp_macaddr, 2);
		temp = nic_readl(nic_dat, MAC_ETH_ADDR0);
		temp_macaddr = __be32_to_cpu(temp);
		memcpy(&net_dev->perm_addr[2], &temp_macaddr, 4);

		if (is_valid_ether_addr(pdata->mac_address)) {
			pr_info("use permanent MAC address %pM from platform settings\n", pdata->mac_address);
			oregano_mac_dev_set_mac_address(net_dev, pdata->mac_address);

		}else if (is_valid_ether_addr(net_dev->perm_addr)) {
			pr_info("read permanent HW MAC address %pM", macaddr_ptr);
			// copy permanent HW MAC address to current device address
			// (initialisation)
#if (KERNEL_VERSION(5, 15, 0) <= LINUX_VERSION_CODE)
			eth_hw_addr_set(net_dev, net_dev->perm_addr);
#else
			memcpy(net_dev->dev_addr, net_dev->perm_addr, ETH_ALEN);
#endif
		}else {
			u8 mac_addr[ETH_ALEN];
			pr_info("could not read mac address, use random generated one\n");
			temp_macaddr = 0xDEAC;
			memcpy(mac_addr, &temp_macaddr, 2);
			temp_macaddr = 0x48 | ((jiffies & 0xFFFFFF) << 8);
			memcpy(&mac_addr[2], &temp_macaddr, 4);
			memset(&net_dev->perm_addr, 0x00, ETH_ALEN);
			dev_dbg(dev, "use auto generated MAC address: %pM\n",mac_addr);
			nic_dev_set_mac_address(net_dev, mac_addr);
		}
	}

	strcpy(net_dev->name, "eth%d"); //TODO(nos)



	// Initialize DMA pool pointer from adapter data structure
	nic_dat->dma = dma;

	// initialize DMA buffer rings
	dev_dbg(dev,
		 "NIC_DEV_INIT: Initializing %u RX DMA buffers of size %u.\n",
		 nic_dat->bd_cnt, DMA_BD_SIZE);


	// Initialize RX BDs
	ret = 0;
	for (count = 0; count < nic_dat->bd_cnt; count++) {
		nic_dat->rx_dma_buffer[count].vaddr =
			dma->dma_vaddr +
			(count * DMA_BD_SIZE);
		nic_dat->rx_dma_buffer[count].addr =
			dma->dma_addr +
			(count * DMA_BD_SIZE);

		// initialize HW buffer descriptor
		bd_initval = MAC_RX_BD_EMPTY | MAC_RX_BD_IRQ;

		// write RX BD initial STATUS
		nic_set_bd_stat_rx(nic_dat, count, bd_initval);

		// write RX BD DMA pointer
		nic_set_bd_addr_rx(nic_dat, count, (u32)nic_dat->rx_dma_buffer[count].addr & ~dma->passthru_mask);

#ifdef DMA_BD_INIT
		memset(nic_dat->rx_dma_buffer[count].vaddr, DMA_BD_INIT,
			   DMA_BD_SIZE);
#endif
	}

	pr_info("NIC_DEV_INIT: Initializing %u TX DMA buffers of size %u and startoff %d.\n",
		 nic_dat->bd_cnt, DMA_BD_SIZE, startoff);

	// Initialize TX BDs
	// Execute only if RX BDs had suzccessfully set up
	for (count = 0; count < nic_dat->bd_cnt; count++) {
		nic_dat->tx_dma_buffer[count].vaddr =
			dma->dma_vaddr +
			( count + nic_dat->bd_cnt ) * DMA_BD_SIZE;
		nic_dat->tx_dma_buffer[count].addr =
			dma->dma_addr +
			( count + nic_dat->bd_cnt ) * DMA_BD_SIZE;

		// initialize HW buffer descriptor
		bd_initval = 0x0;

		// write TX BD initial STATUS
		nic_set_bd_stat_tx(nic_dat, count, bd_initval);

		// write TX BD DMA pointer
		nic_set_bd_addr_tx(nic_dat, count,
				(u32)nic_dat->tx_dma_buffer[count].addr &
					~dma->passthru_mask);
	}

	// initialize DMA buffer pointer
	nic_dat->rx_dma_bd_rp = 0; // reset RX ring read pointer
	nic_dat->tx_dma_bd_rp = 0; // reset TX ring read pointer
	nic_dat->tx_dma_bd_wp = 0; // reset TX ring write pointer

	nic_dat->workq = alloc_workqueue("syn1588", 0, 1);
	if(!nic_dat->workq) {
		pr_err("can't alloc workqueue\n");
		nic_dev_exit(net_dev);
		return ERR_PTR(-ENOMEM);
	}
	INIT_WORK(&nic_dat->work_watchdog, nic_dev_watchdog);

	if (irq != 0) {
		tasklet_init(&nic_dat->tl_rx, syn1588_BHH_RX,
				 (unsigned long)net_dev);
		tasklet_init(&nic_dat->tl_tx, syn1588_BHH_TX,
				 (unsigned long)net_dev);
		ret = request_irq(irq, &nic_interrupt,
			IRQF_DISABLED | IRQF_SHARED, DRV_NAME,
			net_dev);
		if (ret < 0) {
			pr_err("NIC_DEV_INIT: Could not obtain IRQ #%u\n", irq);
			nic_dev_exit(net_dev);
			return ERR_PTR(-EIO);
		}
		nic_dat->irqs[OREGANO_MAC_IRQ_LEGACY] = irq;
	}

	spin_lock_init(&nic_dat->vlan_grp_bitmap_lock);

	return net_dev;
}


static void syn1588_mii_exit(struct net_device *ndev);

/**
 * De-reference network device, remove DMA structure
 * \param adapter pointer to device structure
 * \return 0 on success, negatice error code otherwise
 */
void nic_dev_exit(struct net_device *net_dev)
{
	struct nic_priv_data *nic_dat = netdev_priv(net_dev);
	int i = 0;


	pr_debug("*** nic_dev_exit ***\n");
	nic_dev_close(net_dev);

	if(!IS_ERR_OR_NULL(nic_dat)) {

		for(i = 0; i < OREGANO_MAC_IRQ_AMOUNT; i++) {
			if(nic_dat->irqs[i] > 0 ) {
				free_irq(nic_dat->irqs[i], net_dev);
			}
		}

		if(nic_dat->workq) {
			destroy_workqueue(nic_dat->workq);
		}
		tasklet_kill(&nic_dat->tl_rx); // kill bottom half RX handler
		tasklet_kill(&nic_dat->tl_tx); // kill bottom half TX handler

		pr_info("PCI_DEV_EXIT: Unregistering network device.\n");
	}

	if(nic_dat->mac_capabilities.i2c_interface) {
		oregano_mac_sfp_exit(&nic_dat->sfp);
	}

	if (nic_dat->mii_bus) {
		syn1588_mii_exit(net_dev);
	}

	if(device_is_registered(&net_dev->dev)) {
		unregister_netdev(net_dev);
	}

	free_netdev(net_dev);
	return;
} // nic_dev_exit

/**
 * <!--********************************************************************************-->
 * \defgroup pci_functions functions to handle device operation
 * <!--********************************************************************************-->
 */



/**
 * <!--********************************************************************************-->
 * \defgroup ethtool_functions functions to support ethtool operations
 * <!--********************************************************************************-->
 * @{
 */

// prototype for syn1588_mdio_read
static int syn1588_mdio_read(struct mii_bus *bus, int phy_id, int location);


static void syn1588_get_drvinfo(struct net_device *ndev,
				struct ethtool_drvinfo *info)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;

	dev_dbg(dev, "*** syn1588_get_drvinfo ***\n");
	strncpy(info->driver, DRV_NAME, sizeof(info->driver) - 1);
	// cut off "revision"
	strncpy(info->version, DRV_VERSION, sizeof(info->version));
	strncpy(info->bus_info, dev_name(&ndev->dev),
		sizeof(info->bus_info) - 1);
	snprintf(info->fw_version, sizeof(info->fw_version), "MAC: %s", nic_dat->version);
	info->testinfo_len = SYN1588_TEST_LEN;
}

static int syn1588_get_regs_len(struct net_device *netdev)
{
	struct device *dev = &netdev->dev;

	dev_dbg(dev, "*** syn1588_get_regs_len ***\n");
	return sizeof(mac_regset);
}

static void syn1588_get_regs(struct net_device *ndev, struct ethtool_regs *regs,
				 void *p)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	unsigned char regnr;
	u32 *regs_buff = p;

	dev_dbg(dev, "*** syn1588_get_regs ***\n");

	memset(p, 0, sizeof(mac_regset));

	regs->version = nic_readl(nic_dat, MAC_ETH_VERSION);

	for (regnr = 0; regnr < ARRAY_SIZE(mac_regset) - 1; regnr++)
		regs_buff[regnr] =
			nic_readl(nic_dat, mac_regset[regnr]);
}

static u32 syn1588_get_msglevel(struct net_device *ndev)
{
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	dev_dbg(dev, "*** syn1588_get_msglevel ***\n");
	return nic_dat->msg_enable;
}

static void syn1588_set_msglevel(struct net_device *ndev, u32 level)
{
	struct device *dev = &ndev->dev;
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	dev_dbg(dev, "*** syn1588_set_msglevel ***\n");
	nic_dat->msg_enable = level;
}


static void syn1588_get_strings(struct net_device *ndev, u32 stringset,
				u8 *data)
{
	u8 *p = data;
	int i;

	switch (stringset) {
	case ETH_SS_TEST:
		memcpy(data, *syn1588_gstrings_test,
			   SYN1588_TEST_LEN * ETH_GSTRING_LEN);
		break;
	case ETH_SS_STATS:
		for (i = 0; i < SYN1588_STATS_LEN; i++) {
			memcpy(p, syn1588_gstrings_stats[i].stat_string,
				   ETH_GSTRING_LEN);
			p += ETH_GSTRING_LEN;
		}
		break;
	}
}

/**
 * This function can be used to get the necessary buffer size for calling
 * get_ethtool_stats
 * \param ndev pointer to network device structure
 * \return number of u64 values the get_ethtool_stats function will return
 */
static int syn1588_get_stats_count(struct net_device *ndev)
{
	return SYN1588_STATS_LEN;
}

#if (KERNEL_VERSION(2, 6, 32) < LINUX_VERSION_CODE)
/*
 * This function replaces the following functions for this kernel version:
 * syn1588_self_test_count
 * syn1588_get_stats_count
 * */
static int syn1588_get_sset_count(struct net_device *ndev, int sset)
{
	switch (sset) {
	case ETH_SS_STATS:
		return syn1588_get_stats_count(ndev);
	case ETH_SS_TEST:
		//return syn1588_self_test_count(ndev);
	default:
		return -ENOTSUPP;
	}
}
#endif

static void syn1588_update_stats(struct nic_priv_data *nic_dat) {

	nic_dat->stats.rx_errors =
		nic_readl(nic_dat, MAC_CNT_RXERR);
	nic_dat->stats.rx_crc_errors =
		nic_readl(nic_dat, MAC_CNT_FCSERR);
	nic_dat->extra_stats.mcast_dropped =
		nic_readl(nic_dat, MAC_CNT_MCASTERR);
	if(nic_dat->mac_capabilities.base_1g) {
		nic_dat->extra_stats.rx_late_coll =
			nic_readl(nic_dat, MAC_CNT_LCOLLERR);
	}
	nic_dat->stats.rx_length_errors =
		nic_readl(nic_dat, MAC_CNT_LENERR);
	nic_dat->stats.rx_fifo_errors =
		nic_readl(nic_dat, MAC_CNT_OVERERR);


	if(nic_dat->mac_capabilities.base_1g) {
		nic_dat->stats.rx_frame_errors =
			nic_readl(nic_dat, MAC_CNT_FRAMEERR);
	}

	nic_dat->extra_stats.vlan_err =
		nic_readl(nic_dat, MAC_CNT_VLAN_ERR);


	if(nic_dat->mac_capabilities.rx_timestamper_implemented) {
		nic_dat->extra_stats.ts_rx =
			nic_readl(nic_dat, MAC_CNT_RX_TS);
	}
	if(nic_dat->mac_capabilities.tx_timestamper_implemented) {
		nic_dat->extra_stats.ts_tx =
			nic_readl(nic_dat, MAC_CNT_TX_TS);
	}

	if(nic_dat->mac_capabilities.second_timestamper) {
		nic_dat->extra_stats.ts2_rx =
			nic_readl(nic_dat, MAC_CNT_RX_TS2);
		nic_dat->extra_stats.ts2_tx =
			nic_readl(nic_dat, MAC_CNT_TX_TS2);
	}
	if(nic_dat->mac_capabilities.ptp_timestamper) {
		nic_dat->extra_stats.ts_ptp_rx =
			nic_readl(nic_dat, MAC_CNT_PTP_RXTS);
		nic_dat->extra_stats.ts_ptp_tx =
			nic_readl(nic_dat, MAC_CNT_PTP_TXTS);
	}


	nic_dat->extra_stats.hw_rx_packets =
		nic_readl(nic_dat, MAC_CNT_RX);
	nic_dat->extra_stats.hw_tx_packets =
		nic_readl(nic_dat, MAC_CNT_TX);
	if(nic_dat->mac_capabilities.base_1g) {
		nic_dat->extra_stats.hw_tx_error =
			nic_readl(nic_dat, MAC_CNT_TXERR);
	}
	nic_dat->extra_stats.hw_vlan_packets =
		nic_readl(nic_dat, MAC_CNT_VLAN);
	nic_dat->extra_stats.hw_broadcast_packets =
		nic_readl(nic_dat, MAC_CNT_BROADCAST);
	nic_dat->extra_stats.hw_multicast_packets =
		nic_readl(nic_dat, MAC_CNT_MULTICAST);
}

/**
 * get current network interface ethtool statistics
 * \param ndev pointer to network device structure
 * \param stats pointer to ethtool statistic structure
 * \param data pointer to array of u64 which capable of storing all statistical
 * information
 */
static void syn1588_get_ethtool_stats(struct net_device *ndev,
					  struct ethtool_stats *stats, u64 *data)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	const struct syn1588_stats *stat = syn1588_gstrings_stats;
	int i;
	char *p = NULL;

	syn1588_update_stats(nic_dat);

	for (i = 0; i < SYN1588_STATS_LEN; i++) {
		p = (char *)nic_dat + stat->stat_offset;
		if (stat->sizeof_stat == sizeof(u64))
			data[i] = *(u64 *)p;
		else
			data[i] = *(u32 *)p;

		stat++;
	}
}

/**
 * lock MII management interface access
 * \param ndev pointer to network device of the MII management interface
 * \return 0 on success (MIIM is locked), -EAGAIN if locking failed
 */
static int syn1588_begin(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	if (down_trylock(&nic_dat->mii_access))
		return -EAGAIN;

	return 0;
}

/**
 * release lock on MII management interface
 * \param ndev pointer to network device of the MII management interface
 */
static void syn1588_complete(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);

	up(&nic_dat->mii_access);
}
/// @} ethtool_functions

/**
 * <!--********************************************************************************-->
 * \defgroup mii_functions functions for MII management operation
 * <!--********************************************************************************-->
 * @{
 */

/**
 * read from PHY register
 * \param ndev pointer to network device the PHY is connected to
 * \param phy_id ID of the PHY to be accessed
 * \param location PHY register address
 * \return 16 bit data on success, 0 otherwise
 */
static int syn1588_mdio_read(struct mii_bus *bus, int phy_id, int location)
{
	struct nic_priv_data *nic_dat = bus->priv;
	struct device *dev = &nic_dat->net_dev->dev;
	u32 data;
	short retries = 0;
	unsigned long flags;

	spin_lock_irqsave(&nic_dat->mdio_lock, flags);
	// 100/1000 Mbit MAC found

	// dummy read and dummy write to speed up consecutive PCIe access
	nic_readl(nic_dat, MAC_MIIRX_DATA);
	nic_writel(nic_dat, 0, MAC_MIIADDRESS);

	pr_debug("*** syn1588_mdio_read ***\n");

	if (nic_readl(nic_dat, MAC_MIIRX_DATA) & MAC_MIISTATUS_BUSY) {
		dev_err(dev,
				"MII management read failed: MII link not ready!\n");
		spin_unlock_irqrestore(&nic_dat->mdio_lock, flags);
		udelay(1000);
		return 0;
	}

	// write register address
	if(!nic_writel_check(nic_dat, location << 8 | phy_id, MAC_MIIADDRESS)){
		pr_err("could not write MAC_MIIADDRESS");
	}
	pr_debug("read from MII PHY register 0x%X.\n", location);
	// request register readout
	nic_writel(nic_dat, MAC_MIICOMMAND_READ, MAC_MIICOMMAND);
	udelay(MII_TOUT); // wait till MAC finishes command

	while ((data = (nic_readl(nic_dat, MAC_MIIRX_DATA))) &
			MAC_MIISTATUS_BUSY) {
		if (retries++ > MAX_MII_READ_RETRIES) {
			printk("MII management read timeout!\n");
			data = 0xCAFE;
			break;
		}
	}
	spin_unlock_irqrestore(&nic_dat->mdio_lock, flags);
	return data & 0x0000FFFF;
}
/**
 * write to PHY register
 * \param ndev pointer to network device the PHY is connected to
 * \param phy_id ID of the PHY to be accessed
 * \param location PHY register address
 * \param val value to write to the PHY register
 */
static int syn1588_mdio_write(struct mii_bus *bus, int phy_id, int location,
				u16 val)
{
	struct nic_priv_data *nic_dat = bus->priv;
	struct device *dev = &nic_dat->net_dev->dev;
	u32 data;
	short retries = 0;
	unsigned long flags;

	dev_dbg(dev, "*** syn1588_mdio_write ***\n");
	spin_lock_irqsave(&nic_dat->mdio_lock, flags);
	if (nic_readl(nic_dat, MAC_MIIRX_DATA) & MAC_MIISTATUS_BUSY) {
		dev_err(dev, "MII management write failed: MII link not ready!\n");
		spin_unlock_irqrestore(&nic_dat->mdio_lock, flags);
		return -1;
	}

	// write register address
	nic_writel(nic_dat, location << 8 | phy_id, MAC_MIIADDRESS);
	// write register data
	nic_writel(nic_dat, val & 0x0000FFFF, MAC_MIITX_DATA);
	// request register write
	nic_writel(nic_dat, MAC_MIICOMMAND_WRITE, MAC_MIICOMMAND);
	udelay(MII_TOUT); // wait till MAC finishes command

	while ((data = (nic_readl(nic_dat, MAC_MIIRX_DATA))) &
			MAC_MIISTATUS_BUSY) {
		if (retries++ > MAX_MII_WRITE_RETRIES) {
			dev_err(dev, "MII management write timeout!\n");
			break;
		}
	}
	spin_unlock_irqrestore(&nic_dat->mdio_lock, flags);
	return 0;
}


static void syn1588_mii_exit(struct net_device *ndev) {
	struct phy_device *phydev = nic_phydev(ndev);
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	if(!IS_ERR_OR_NULL(phydev)) {
		phy_disconnect(phydev);
		phy_device_remove(phydev);
		phy_device_free(phydev);
		nic_phydev_set(ndev, NULL);
	}
	if (!IS_ERR_OR_NULL(nic_dat->mii_bus)) {
		mdiobus_unregister(nic_dat->mii_bus);
		mdiobus_free(nic_dat->mii_bus);
		nic_dat->mii_bus = NULL;
	}

}

/**
 * initialise MDIO functions.
 * detect PHY and run initialisation sequence
 * \param ndev pointer to network device the PHY is connected to
 * \return 0 on success, negative error code otherwise (e.g., no PHY detected)
 */
static int syn1588_mii_probe(struct net_device *ndev)
{
	struct nic_priv_data *nic_dat = netdev_priv(ndev);
	struct device *dev = &ndev->dev;
	struct phy_device *phydev = NULL;
	u32 miicommand_status = nic_readl(nic_dat, MAC_MIICOMMAND);
	bool is_sfp_port = (miicommand_status & MAC_MIICOMMAND_SFP_PORT) > 0;
	int phy_addr;
	int ret;
	u32 nic_mac_frequency = 0;
	u32 nic_mdioclk_frequency = 0;

	dev_dbg(dev, "*** syn1588_mii_init ***\n");
	spin_lock_init(&nic_dat->mdio_lock);
	// calculate correck MII clock devider (has to be even)
	// This also determines how long to wait for access to MIIM (mdio_read/write)

	nic_mac_frequency	 = MAC_FRQ;
	nic_mdioclk_frequency = MDC_FRQ;

	nic_dat->mii_clkdiv = (nic_mac_frequency / nic_mdioclk_frequency + 1) & (~1);


	if (nic_dat->mac_capabilities.base_1g) {
		// 100/1000 Mbit MAC
		// write MII clock divider to MAC register for MDIO speed
		if (!nic_writel_check(nic_dat, nic_dat->mii_clkdiv,
				MAC_MIIMODE)) {
			u32 tmp = nic_readl(nic_dat, MAC_MIIMODE);
			dev_err(dev, "error setting MII clock divider! expected: %x - %x\n", nic_dat->mii_clkdiv, tmp);
			return -EIO;
		}
		pr_info("using MII managment data clock %u kHz (div.: %lu).\n",
			nic_mac_frequency / nic_dat->mii_clkdiv / 1000,
			(unsigned long)nic_dat->mii_clkdiv);
	} else if(nic_dat->mac_capabilities.base_10g) {
		// 10Gbit interface
		// no need to set MDIO devider as no MDIO present
	} else {
		pr_err("error not known speed\n");
		return -ENXIO;
	}

	nic_dat->mii_bus = mdiobus_alloc();
	if (nic_dat->mii_bus == NULL) {
		dev_err(dev, "could not allocate onboard mdiobus\n");
		return -ENOMEM;
	}

	snprintf(nic_dat->mii_bus->id, MII_BUS_ID_SIZE, "mdio-%s", ndev->name);
	nic_dat->mii_bus->name = "Oregano MDIO Bus";
	nic_dat->mii_bus->priv = nic_dat;
	nic_dat->mii_bus->parent = &ndev->dev;
	nic_dat->mii_bus->read = &syn1588_mdio_read;
	nic_dat->mii_bus->write = &syn1588_mdio_write;
	nic_dat->mii_bus->phy_mask = ~0u;

	ret = mdiobus_register(nic_dat->mii_bus);
	if (ret)	{
		dev_err(dev, "could not register mdiobus\n");
		mdiobus_free(nic_dat->mii_bus);
		return ret;
	}


	for (phy_addr = 0; phy_addr < PHY_ID_MASK; phy_addr++) {
		phydev = mdiobus_scan_c22(nic_dat->mii_bus, phy_addr);
		if (!IS_ERR_OR_NULL(phydev)) {
			dev_info(dev, "found phy on addess %d\n", phy_addr);
			break;
		}
	}
	if (IS_ERR_OR_NULL(phydev)) {
		dev_err(dev, "error no mdio phy found for base_t\n");
		mdiobus_unregister(nic_dat->mii_bus);
		mdiobus_free(nic_dat->mii_bus);
		return -ENODEV;
	}
	nic_phydev_set(ndev, phydev);
	phy_attached_info(phydev);
	if(is_sfp_port) {
		dev_info(dev, "old sfp implementation, apply workaround for fiber module\n");
		phy_connect_direct(ndev, phydev, oregano_mac_phy_handler_fiber_quirk, PHY_INTERFACE_MODE_SGMII );
	} else {
	phy_connect_direct(ndev, phydev, oregano_mac_phy_handler, PHY_INTERFACE_MODE_SGMII );
	}
	phy_start(phydev);

	return 0;
}
/// @} mii_functions

struct oregano_mac_data {
	void __iomem *iobase;
	struct net_device *netdev;

	struct platform_device *i2c_pdev;

	struct oregano_mac_platform_data pdata;
	struct syn1588_dma dma; ///< depricated dma buffer, will be removed in the future currently needed for dual nic
};


struct nic_priv_data *dev_get_mac(struct device *dev) {
	if(IS_ERR_OR_NULL(dev)) {
		return NULL;
	} else {
		//TODO check driver
		struct oregano_mac_data *data = dev_get_drvdata(dev);
		return netdev_priv(data->netdev);
	}
}
EXPORT_SYMBOL_GPL(dev_get_mac);

/// @{ paramter parsing
static int oregano_mac_probe_pdata(struct platform_device *pdev, struct oregano_mac_platform_data *pdata)
{
	struct device *dev = &pdev->dev;
	struct oregano_mac_platform_data *pdata_org = dev_get_platdata(dev);

	*pdata = *pdata_org;
	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id oregano_mac_dt_ids[] = {
	{ .compatible = DRV_NAME },
	{ .compatible = "oregano,mac" },
	{ .compatible = "oregano,xmac" },
	{ .compatible = "oregano,tse_mac" },
	{}
};
MODULE_DEVICE_TABLE(of, oregano_mac_dt_ids);

static int oregano_mac_probe_dt(struct platform_device *pdev, struct oregano_mac_platform_data *pdata)
{
	struct device *dev = &pdev->dev;

	struct device_node *clock_m_node;
	struct platform_device *clock_m_pdev;
	struct device_node *i2c_node;

	#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
	if(!of_get_mac_address(pdev->dev.of_node, pdata->mac_address)) {
		dev_info(dev, "found mac in device tree\n");
	}
	#endif // end KERNEL_VERSION(5, 13, 0)


	clock_m_node = of_parse_phandle(pdev->dev.of_node, "clock_m", 0);
	clock_m_pdev = of_find_device_by_node(clock_m_node);
	if (!clock_m_pdev) {
		dev_err(dev, "of_find_device_by_node(%pOF) failed\n", clock_m_node);
		return -ENODEV;
	}

	pdata->clock_m = &clock_m_pdev->dev;



	i2c_node = of_parse_phandle(pdev->dev.of_node, "sfp_i2c", 0);
	pdata->i2c_adapter = of_get_i2c_adapter_by_node(i2c_node);


	// data->i2c_adapter = i2c_get_adapter(4);
	if(!IS_ERR_OR_NULL(i2c_node)) {
		if(IS_ERR_OR_NULL(pdata->i2c_adapter)) {
			dev_err(dev, "could not find i2c bus");
			return -ENODEV;
		} else {
			dev_info(dev, "DEBUG: found i2c: %s, adapter: %s\n", i2c_node->name, pdata->i2c_adapter->name);
		}
	}


	return 0;
}
#else
static int oregano_mac_probe_dt(struct platform_device *pdev, struct oregano_mac_platform_data *pdata)
{
	(void)pdev;
	(void)pdata;
	return -1;
}
#endif

/// @} paramter parsing

static struct oregano_i2c_platform_data ocores_platform_data = {
	.big_endian = false,
	.clock_khz = 125000,
	.bus_khz = 100,
	.reg_io_width = 4,
	.reg_shift = 2,
};


static int oregano_mac_remove(struct platform_device *pdev)
{
	struct oregano_mac_data *data = platform_get_drvdata(pdev);
	dev_dbg(&pdev->dev, "remove driver");


	if(IS_ERR_OR_NULL(data->netdev)) {
		pr_err("internal error wrong error handling\n");
		return 0;
	}
	nic_dev_exit(data->netdev);


	if(!IS_ERR_OR_NULL(data->dma.dma_vaddr)) {
		dma_pool_free(data->dma.dma_pool, data->dma.dma_vaddr, data->dma.dma_addr);
		data->dma.dma_vaddr = NULL;
		data->dma.dma_addr = 0;
	}

	if(!IS_ERR_OR_NULL(data->dma.dma_pool)) {
		dma_pool_destroy(data->dma.dma_pool);
		data->dma.dma_pool = NULL;
	}

	if(!IS_ERR_OR_NULL(data->i2c_pdev)){
		//i2c_put_adapter(data->i2c_adapter);
		platform_device_unregister(data->i2c_pdev);
	}

	return 0;
}

static int oregano_mac_probe(struct platform_device *pdev)
{
	int status;
	int ret;
	struct device *dev = &pdev->dev;
	struct device *dev_dma = dev; // device what should do the dma allocation
	struct resource *res;
	struct resource res_i2c;
	int irq = -1;
	struct oregano_mac_data *data = devm_kzalloc(dev, sizeof(struct oregano_mac_data), GFP_KERNEL);
	if(IS_ERR_OR_NULL(data)) {
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, data);

	dev_info(dev, "probe oregano mac driver\n");

	if (dev_of_node(dev)) {
		status = oregano_mac_probe_dt(pdev, &data->pdata);
	} else {
		status = oregano_mac_probe_pdata(pdev, &data->pdata);
		//some kernel need to have non devicetree drivers, to use the pci device for dma handling
		dev_dma = dev->parent;
	}

	if (status) {
		dev_err(dev, "error reading probe data\n");
		return status;
	}


	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "main");
	if(res == NULL) {
		res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
		if(res == NULL) {
			dev_err(dev, "could not request resources\n");
			return -ENODEV;
		}
	}
	// sperate memory into I2C and MAC
	res_i2c = *res;
	res_i2c.end = res_i2c.start + PCI_SIZE_MAC_I2C - 1;
	res->start += PCI_SIZE_MAC_I2C;

	irq = platform_get_irq_byname(pdev, "legacy");
	if(irq < 0) {
		irq = platform_get_irq(pdev, 0);
		if(irq < 0) {
			dev_err(dev, "could not get valid interrupt");
			return -EINVAL;
		}
	}

	data->iobase = devm_ioremap_resource(dev, res);
	if(IS_ERR_OR_NULL(data->iobase)) {
		dev_err(dev, "could not ioremap resource\n");
		return PTR_ERR(data->iobase);
	}

	data->iobase -= PCI_SIZE_MAC_I2C; // workaround to have the first 0x10 bytes as i2c


	{
		data->dma.dma_pool = dma_pool_create(
			"syn1588nic_dma_pool",
			dev_dma,
			DMA_BD_SIZE * BUFF_CNT * 2,
			8,
			0);

		if(IS_ERR_OR_NULL(data->dma.dma_pool)) {
			dev_err(dev, "Error allocating DMA pool!\n");
			return -ENOMEM;
		}

		data->dma.dma_vaddr = dma_pool_alloc(data->dma.dma_pool,
					GFP_KERNEL | GFP_DMA | __GFP_ZERO, &data->dma.dma_addr);

		if(IS_ERR_OR_NULL(data->dma.dma_vaddr)) {
			dev_err(dev, "Error allocating DMA buffer memory! errno: %ld\n", PTR_ERR(data->dma.dma_vaddr));
			return -ENOMEM;
		}

		if(! IS_ERR_OR_NULL(data->pdata.pci_bridge)) {
			data->dma.passthru_mask = altera_pci_bridge_get_address_mask(data->pdata.pci_bridge);
			altera_pci_bridge_set_address_translation(data->pdata.pci_bridge, data->dma.dma_addr);
		}
	}


	data->netdev = nic_dev_alloc(
		dev,
		data->iobase,
		&data->dma,
		irq,
		&data->pdata,
		get_pci_nic_2_sfp_hw_delay); //TODO phy delay
	if(IS_ERR_OR_NULL(data->netdev)) {
		dev_err(dev, "could not init nic\n");
		return -ENODEV;
	}
	ret = register_netdev(data->netdev);
	if (ret != 0) {
		dev_err(dev, "NIC_DEV_INIT: Failed to register NIC!\n");
		nic_dev_exit(data->netdev);
		return ret;
	}

	if(((struct nic_priv_data *)netdev_priv(data->netdev))->mac_capabilities.i2c_interface && data->pdata.i2c_adapter == NULL) {
		data->i2c_pdev = platform_device_register_resndata(
			dev, "oregano-i2c", PLATFORM_DEVID_AUTO, &res_i2c, 1,
			&ocores_platform_data, sizeof(ocores_platform_data)
		);
		if(IS_ERR_OR_NULL(data->i2c_pdev)) {
			dev_err(dev, "could not request i2c resources, err: %ld\n", PTR_ERR(data->i2c_pdev));
			oregano_mac_remove(pdev);
			return -ENODEV;
		}

		data->pdata.i2c_adapter = dev_get_oregano_i2c(&data->i2c_pdev->dev);
		if(IS_ERR_OR_NULL(data->pdata.i2c_adapter)) {
			dev_err(dev, "i2c_addapter is NULL\n");
			oregano_mac_remove(pdev);
			return -ENODEV;
		}
	}
	if(((struct nic_priv_data *)netdev_priv(data->netdev))->mac_capabilities.i2c_interface) {
		struct nic_priv_data *nic_dat = netdev_priv(data->netdev);
		phy_interface_t interface = PHY_INTERFACE_MODE_SGMII;
		if(IS_ERR_OR_NULL(data->pdata.i2c_adapter)) {
			dev_err(dev, "i2c_addapter is NULL\n");
			oregano_mac_remove(pdev);
			return -ENODEV;
		}
		if(nic_dat->mac_capabilities.base_10g) {
			interface = PHY_INTERFACE_MODE_XGMII;
		}

		ret = oregano_mac_sfp_probe(&pdev->dev, &nic_dat->sfp, nic_dat->net_dev, data->pdata.i2c_adapter, interface);
		if(ret) {
			dev_err(dev, "could not init sfp part");
			oregano_mac_remove(pdev);
			return ret;
		}
	} else {
		int ret = syn1588_mii_probe(data->netdev);
		if(ret) {
			dev_err(dev,"could not init mdio driver, err: %d\n", ret);
			nic_dev_exit(data->netdev);
			return ret;
		}
	}





	return 0;
}






MODULE_ALIAS("platform:" DRV_NAME);

static struct platform_driver oregano_mac_driver = {
	.driver = {
		.name	= DRV_NAME,
#ifdef CONFIG_OF
		.of_match_table = of_match_ptr(oregano_mac_dt_ids),
#endif // CONFIG_OF
	},
	.probe = oregano_mac_probe,
	.remove = oregano_mac_remove,
};
module_platform_driver(oregano_mac_driver);

MODULE_DESCRIPTION("mac driver for oregano");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
