// SPDX-License-Identifier: GPL-2.0
/*
 *------------------------------------------------------------------------------
 *
 * syn1588(R) NIC
 *
 *------------------------------------------------------------------------------
 *
 * Oregano Systems - Design & Consulting GesmbH
 *
 * Web: http://oregano.at	@: support@oregano.at
 *
 *------------------------------------------------------------------------------
 *
 * Copyright (c) 2006-2020
 *
 *------------------------------------------------------------------------------
 *
 * Confidential data.
 *
 * This source code, netlists, documentation is confidential and must not be
 * published nor made available to third parties. Usage of this source code,
 * netlists, documentation is subject to restrictions and conditions of a
 * license agreement.
 *
 * ALL RIGHTS RESERVED.
 * Oregano Systems does not assume any liability
 * arising out of the application or use of any product described or shown
 * herein nor does it convey any license under its patents, copyrights, or any
 * rights of others.
 *
 * Oregano Systems reserves the right to make changes, at any time, in
 * order to improve reliability, function or design. Oregano Systems will not
 * assume responsibility for the use of any circuitry described herein.
 *
 * All trademarks used are the property of their respective owners.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA	02110-1301, USA.
 *
 *------------------------------------------------------------------------------
 *
 *  Author: Norbert Summer
 *
 *  Orginal Author: Patrick Loschmidt
 *	Austrian Academy of Science
 *	Institute for Integrated Sensor Systems
 *  Date of Creation:			 Fri 2006-06-16
 *  Description: the synchronisation driver char module
 */

/**
 * \todo
 * virtdev access key is unnecessary, only filps are unique!
 */

#include <asm/types.h>
#include <linux/atomic.h>
#include <linux/device.h>
#include <linux/fcntl.h> // O_ACCMODE
#include <linux/fs.h> // everything...
#include <linux/init.h> // module_init, module_exit
#include <linux/interrupt.h> // required for 2.6.16 < Kernel < 2.6.19
#include <linux/kernel.h> // printk()
#include <linux/miscdevice.h> //
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/poll.h> // for syncD_poll() values
#include <linux/proc_fs.h>
#include <linux/sched.h> // TASK_NORMAL and similar macros
#include <linux/slab.h> // kmalloc()
#include <linux/types.h> // size_t
#include <linux/uaccess.h>
#include <linux/version.h>

#if defined(CONFIG_OF)
#include <linux/of.h>
#include <linux/of_device.h>
#endif


#include "SyncDriver.h" // local definitions
#include "oregano_tse_mac.h"
#include "version.h" // version
#include "clock_m.h"
#include "oregano_updater_altera.h"

#include "include/syn1588_sys.h"


#ifndef RHEL_BACKPORT_IDR2
#include <linux/idr.h>
#else
#include <linux/idr2.h>
#endif



#define DRV_NAME "oregano-syncd"



#define DEV_NAME_TPL "syncD%d" ///< template for generating device file

#ifndef RHEL_BACKPORT_IDR2
DEFINE_IDR(syncd_dev_idr); // keep track of minor -> device relation
DEFINE_IDA(syncd_dev_ida); // keep trak of free ids
#else
DEFINE_IDR2(syncd_dev_idr); // keep track of minor -> device relation
DEFINE_IDA(syncd_dev_ida); // keep trak of free ids
#endif
spinlock_t idr_lock; // lock register idr, enumeration id


static void syncd_device_free_minor(int minor)
{
	spin_lock(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	idr_remove(&syncd_dev_idr, minor);
#else
	idr2_remove(&syncd_dev_idr, minor);
#endif
	spin_unlock(&idr_lock);
}

static void syncD_device_alloc_minor(struct syncD_device *device)
{
	struct device *dev = device->dev;
	int minor = device->miscdev.minor;
	int alloc_id;
	spin_lock(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	alloc_id = idr_alloc(&syncd_dev_idr, device, minor, MINORMASK + 1, GFP_KERNEL);
#else
	alloc_id = idr2_alloc(&syncd_dev_idr, device, minor, MINORMASK + 1, GFP_KERNEL);
#endif
	spin_unlock(&idr_lock);
	if(minor != alloc_id) {
		syncd_device_free_minor(alloc_id);
		dev_err(dev, "can't alloc minor id for #%d \n", minor);
	}
}


static struct syncD_device *syncD_device_get_from_minor(int minor)
{
	struct syncD_device *device;

	spin_lock(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	device = idr_find(&syncd_dev_idr, minor);
#else
	device = idr2_find(&syncd_dev_idr, minor);
#endif
	if (!device) {
		spin_unlock(&idr_lock);
		return NULL;
	}
	spin_unlock(&idr_lock);

	return device;
}

static int syncd_device_alloc_id(void)
{
	int ret;
	spin_lock(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	ret = ida_simple_get(&syncd_dev_ida, 0, 0xff, GFP_KERNEL);
#else
	ret = ida_simple_get(&syncd_dev_ida, 0, 0xff, GFP_KERNEL);
#endif
	spin_unlock(&idr_lock);
	return ret;
}

static void syncD_device_free_id(int id)
{
	spin_lock(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	ida_simple_remove(&syncd_dev_ida, id);
#else
	ida_simple_remove(&syncd_dev_ida, id);
#endif
	spin_unlock(&idr_lock);
}


static struct clock_m_device *get_clock_m(struct syncD_device *syncd) {
	if(IS_ERR_OR_NULL(syncd->syn1588_dev.clock_m_dev)) {
		return NULL;
	}
	return dev_get_clock_m(&syncd->syn1588_dev.clock_m_dev->dev);
}
static struct nic_priv_data *get_mac_by_index(struct syncD_device *syncd, int index) {
	if(index < 0 || index >= MAX_AMOUNT_MACS) {
		return NULL;
	}
	if(syncd->syn1588_dev.mac_devs[index] == NULL) {
		return NULL;
	}
	return dev_get_mac(&syncd->syn1588_dev.mac_devs[index]->dev);
}

static struct oregano_updater_altera *get_updater(struct syncD_device *syncd) {
	return dev_get_updater_altera(syncd->updater_dev);
}


// forward deklarations
static unsigned int syncD_poll(struct file *filp, poll_table * wait);
static loff_t syncD_llseek(struct file *filp, loff_t off, int whence);
static ssize_t syncD_read(struct file *filp, char *buf, size_t count,
			  loff_t *f_pos);
static ssize_t syncD_write(struct file *filp, const char *buf, size_t count,
			   loff_t *f_pos);
static int syncD_open(struct inode *inode, struct file *filp);
static int syncD_release(struct inode *inode, struct file *filp);

static long syncD_ioctl(struct file *, unsigned int, unsigned long);

/// structure for synchonisation driver file operation functions
static struct file_operations syncD_fops = {
	.poll = syncD_poll,
#if (KERNEL_VERSION(2, 6, 34) < LINUX_VERSION_CODE)
	.compat_ioctl = NULL,
	.unlocked_ioctl = syncD_ioctl,
#else
	.ioctl = syncD_ioctl,
#endif
	.mmap = NULL,
	.open = syncD_open,
	.flush = NULL,
	.release = syncD_release,
	.lock = NULL,
	// depricated
	.fsync = NULL,
	.fasync = syncD_fasync,
	.owner = THIS_MODULE,
	.llseek = syncD_llseek,
	.read = syncD_read,
	.write = syncD_write,
};

static void syncD_Event(const struct clock_m_device *clock_m, const struct oregano_clock_event_data *event, void *data)
{
	struct syncD_device *syncD_dev = (struct syncD_device *)data;
	struct device *dev = syncD_dev->dev;
	u32 irq_src = event->irs;
	//	unsigned char count;
	struct syncD_virtdev *entry;

	(void) clock_m;

	// modify virtual IRSRC registers
	spin_lock(&syncD_dev->virtdev_lock); // lock virtual device list
	list_for_each_entry(entry, &syncD_dev->virtdev_list, list) {
		if (0 != (irq_src & entry->iren)) {
			if (unlikely(0 != (entry->irsrc &
					   (irq_src & entry->iren)))) {
				// mask just contains the interrupt which was lost != lost interrupt(s)!
				dev_warn(dev,
					 "PID %u lost interrupt (mask: 0x%X).\n",
					 entry->pid,
					 (entry->irsrc & (irq_src & entry->iren)));
				entry->irq_fail = -EIO;
			}
			entry->irsrc |= (irq_src & entry->iren);
			wake_up(&(
				entry->irq_waitq)); // wake up process with blocking read on IRSRC
		}
	}
	spin_unlock(&syncD_dev->virtdev_lock); // unlock virtual device list

	if (syncD_dev->async_queue)
		kill_fasync(&syncD_dev->async_queue, SIGIO,
				POLL_IN); // notify asynchronous readers

}

/**
 * returns the virtual device identified by the access_key
 * searches for a device belonging to access_key and returns the pointer to the
 * device structure.
 * if no appropriate device is found in the list of active virtual devices and
 * the flag O_CREAT is
 * set, a device is created and linked to the list.
 * \param dev pointer to SyncD device structure representing a physical device
 * \param access_key pointer to the file structure identifing the opened
 * instance
 * \return pointer to valid SyncD_virtdev structure, NULL in case of an error
 * \note the call of this procedure has to be secured by mutual exclusion!
 */
static struct syncD_virtdev *syncD_create_virtdev(struct syncD_device *syncD_dev,
						  struct file *access_key)
{
	struct device *dev = syncD_dev->miscdev.this_device;
	struct syncD_virtdev *entry;

	spin_lock_bh(&syncD_dev->virtdev_lock); // lock linked list change
	list_for_each_entry(entry, &syncD_dev->virtdev_list, list)
		if (entry->access_key == access_key) {
			dev_notice(dev, "multiple open on virtual device 0x%p\n",
				 access_key);
			spin_unlock_bh(&syncD_dev->virtdev_lock);
			return entry; // device with given access key was found
		}
	// no device was found
	// malloc has to be atomic because the function should be protected by
	// semaphores
	entry = kzalloc(sizeof(struct syncD_virtdev), GFP_ATOMIC);
	if (!entry) {
		dev_warn(dev,
			 "unable to allocate memory for virtual device structure.\n");
		spin_unlock_bh(&syncD_dev->virtdev_lock);
		return NULL;
	}

	// initialize virtual device
	entry->access_key = access_key; // assign device to access_key
	entry->pid = current->pid; // store current process id
	entry->syncd_dev = syncD_dev; // store pointer to main device
	entry->event_id = -1;
	init_rwsem(&(entry->access_sem)); // init virtual device semaphore
	init_waitqueue_head(
		&(entry->irq_waitq)); // init virtual device IRQ wait queue
	if (list_empty(&syncD_dev->virtdev_list) && atomic_read(&syncD_dev->virtdev_count) != 0)
		dev_warn(dev,
			 "creating first virtual device, but device count was not zero.\n");
	list_add(&entry->list, &syncD_dev->virtdev_list);
	atomic_add(1, &syncD_dev->virtdev_count);
	dev_info(dev, "created new virtual device for PID %u.\n", entry->pid);
	spin_unlock_bh(&syncD_dev->virtdev_lock);

	return entry;
}

/**
 * unlinks the device matching access_key from the device list and frees the
 * memory
 * \param dev pointer to SyncD device structure representing a physical device
 * \param access_key pointer to the file structure identifing the opened
 * instance
 * \return 0 on success or -ENODEV if device is not found
 * \note the call of this procedure has to be secured by mutual exclusion!
 */
static int syncD_release_virtdev(struct syncD_device *syncd_dev,
				 struct file *access_key)
{
	struct syncD_virtdev *entry, *next;
	struct device *dev = syncd_dev->miscdev.this_device;

	spin_lock_bh(&syncd_dev->virtdev_lock);
	list_for_each_entry_safe(entry, next, &syncd_dev->virtdev_list, list)
		if (entry->access_key == access_key) {
			list_del(
				&entry->list); // unlink current the device found
			kfree(entry); // free used memory
			dev_info(dev, "unlinked virtual device for PID %u.\n",
				 entry->pid);

			atomic_sub(1, &syncd_dev->virtdev_count);
			if (atomic_read(&syncd_dev->virtdev_count) < 0)
				dev_warn(dev,
					 "virtual device count decrement failed.\n");

			if (list_empty(&syncd_dev->virtdev_list) &&
				atomic_read(&syncd_dev->virtdev_count) != 0)
				dev_err(dev, "virtual device count mismatch.\n");
			spin_unlock_bh(&syncd_dev->virtdev_lock);
			return 0;
		}

	dev_notice(dev, "no virtual device with access_key 0x%p to unlink.\n",
		 access_key);
	spin_unlock_bh(&syncd_dev->virtdev_lock);
	return -ENODEV; // device with key was not found
}

/**
 * <!--********************************************************************************-->
 * \defgroup syncD_fops synchronization device file operations
 * <!--********************************************************************************-->
 */

/**
 * function to handle file open request from user-space
 * \param inode pointer to inode structure the file is referenced in
 * \param filp pointer to file structure (file descriptor)
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
static int syncD_open(struct inode *inode, struct file *filp)
{
	struct syncD_virtdev *virtdev = NULL; // virtual device
	struct syncD_device *entry;

	pr_debug("driver device file open.\n");


	entry = syncD_device_get_from_minor(iminor(inode));
	if (entry == NULL) {
		pr_err("no device to minor id found");
		return -ENODEV;
	}
	virtdev = syncD_create_virtdev(entry, filp);
	filp->private_data = virtdev;
	return 0;
}

/**
 * function to handle file close request from user-space
 * \param inode pointer to inode structure the file is referenced in
 * \param filp pointer to file structure (file descriptor)
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
static int syncD_release(struct inode *inode, struct file *filp)
{
	struct syncD_virtdev *virtdev = filp->private_data; // virtual device
	struct syncD_device *syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->miscdev.this_device;
	struct clock_m_device *clock_m = get_clock_m(syncd_dev);
	int ret;

	if (!virtdev) {
		dev_err(dev, "illegal release on virtual device 0x%p.\n", filp);
		return -EBADF;
	}

	if (filp->f_flags & O_NONBLOCK) {
		if (down_write_trylock(&syncd_dev->virtdev_list_sem) == 0)
			return -EWOULDBLOCK;
	} else
		down_write(&syncd_dev->virtdev_list_sem);

	if (filp->f_flags & O_NONBLOCK) {
		if (down_write_trylock(&virtdev->access_sem) == 0) {
			up_write(&syncd_dev->virtdev_list_sem);
			return -EWOULDBLOCK;
		}
	} else
		down_write(&virtdev->access_sem);

	// remove asynchronous reader from queue
	if (filp->f_flags & FASYNC)
		syncD_fasync(-1, filp, 0);

	virtdev->syncd_dev =
		NULL; // block reading or writing to this virtual device from now on
	up_write(&virtdev->access_sem);


	if(!IS_ERR_OR_NULL(clock_m)) {
		int period_id;
		int trigger_id;
		int io_port_id;
		for_each_set_bit(period_id, virtdev->used_period_ids, __SYN1588_IOCTL_PERIOD_NBITS) {
			clock_m_period_clear(clock_m, period_id);
			clear_bit(period_id, virtdev->used_period_ids);
		}
		for_each_set_bit(trigger_id, virtdev->used_trigger_ids, __SYN1588_IOCTL_TRIGGER_NBITS) {
			clock_m_trigger_clear(clock_m, trigger_id);
			clear_bit(trigger_id, virtdev->used_trigger_ids);
		}
		for_each_set_bit(io_port_id, virtdev->used_io_port_ids, __SYN1588_IOCTL_IO_PORT_NBITS) {
			struct syn1588_ioctl_sink sink = {
				.type = SYN1588_IOCTL_SINK_IO_PORT,
				.id = io_port_id,
			};
			clock_m_io_matrix_remove_sink(clock_m, sink);
			clock_m_io_port_clear(clock_m, io_port_id);
			clear_bit(io_port_id, virtdev->used_io_port_ids);
		}
		if(virtdev->event_id >=0) {
			struct syn1588_ioctl_sink sink = {.type = SYN1588_IOCTL_SINK_EVENT};
			sink.id = virtdev->event_id ;
			clock_m_io_matrix_remove_sink(clock_m, sink);
			clock_m_event_free(clock_m, virtdev->event_id);
			virtdev->event_id = -1;
		}
	}

	// release current device
	ret = syncD_release_virtdev(syncd_dev, filp);

	up_write(&syncd_dev->virtdev_list_sem);

	if (list_empty(&syncd_dev->virtdev_list)) {
		// last opened virtual device closed
		dev_info(dev, "disabling all IRQs.\n");
		if (clock_m_setreg(get_clock_m(syncd_dev), IREN, 0x0) != 0)
			dev_warn(dev, "disabling interrupts failed.\n");
		dev_info(dev, "unregistered IRQ callback function.\n");
	}
	return ret;
}

static int ioctl_read_build_id(struct syncD_virtdev *virtdev, uint32_t __user *arg) {
	u32 value = virtdev->syncd_dev->syn1588_dev.build_id;

	if( copy_to_user(arg, &value, sizeof(value)) )
	{
		pr_err("Data Read : Err!\n");
		return -1;
	}
	return 0;
}


static int ioctl_read_get_phc_id(struct syncD_virtdev *virtdev, u32 __user *arg) {
	struct clock_m_device * clock_m = get_clock_m(virtdev->syncd_dev);
	u32 clock_id;
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	clock_id = ptp_clock_index(clock_m->ptp_clock);
	return copy_to_user(arg, &clock_id, sizeof(u32)) ? -EFAULT : 0;
}

static int ioctl_read_clock_id(struct syncD_virtdev *virtdev, uint8_t __user arg[8]) {
	u8 buf[8];

	struct clock_m_device * clock_m = get_clock_m(virtdev->syncd_dev);
	u32 lower = 0;
	u32 upper = 0;
	clock_m_getreg(clock_m, CLOCKID_L, &lower);
	clock_m_getreg(clock_m, CLOCKID_H, &upper);
	buf[7] = lower & 0xFF;
	buf[6] = (lower >> 8) & 0xFF;
	buf[5] = (lower >> 16) & 0xFF;
	buf[4] = (lower >> 24) & 0xFF;
	buf[3] = upper & 0xFF;
	buf[2] = (upper >> 8) & 0xFF;
	buf[1] = (upper >> 16) & 0xFF;
	buf[0] = (upper >> 24) & 0xFF;

	return  copy_to_user(arg, buf, sizeof(buf)) ? -EFAULT : 0;
}

static int ioctl_read_product_id(struct syncD_virtdev *virtdev, struct syn1588_ioctl_buffer  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	const char *product_id = "unknown";
	struct syn1588_ioctl_buffer tmp;
	if(copy_from_user(&tmp, arg, sizeof(struct syn1588_ioctl_buffer))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	product_id = syn1588_dev_get_product_id(syncd_dev->syn1588_dev.product_id);

	if(strlen(product_id) >= tmp.size) {
		return -ENOMEM;
	}
	return  copy_to_user(tmp.data, product_id, strlen(product_id)+1) ? -EFAULT : 0;
}

static int ioctl_flash_and_verify(struct syncD_virtdev *virtdev, struct syn1588_ioctl_buffer_const  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct syn1588_ioctl_buffer tmp;
	u32 *buffer;
	const size_t max_update_size = 50*1024*1024;
	int result = 0;

	dev_warn(dev, "start update of PCIE card\n");
	if(IS_ERR_OR_NULL(get_updater(syncd_dev))) {
		return -ENODEV;
	}

	if(copy_from_user(&tmp, arg, sizeof(struct syn1588_ioctl_buffer))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	if(tmp.size == 0 && tmp.size >= max_update_size) {
		return -EINVAL;
	}

	buffer = vmalloc(DIV_ROUND_UP(tmp.size, sizeof(buffer[0])) * sizeof(buffer[0]));
	if(IS_ERR_OR_NULL(buffer)) {
		return -ENOMEM;
	}
	if(copy_from_user(buffer, tmp.data, tmp.size) !=0) {
		dev_err(dev, "could not copy buffer from userspace");
		vfree(buffer);
		return -EINVAL;
	}
	result = oregano_updater_altera_update_firmware(get_updater(syncd_dev), tmp.size, buffer);
	if(result) {
		dev_err(dev, "could not start with write firmware\n");
		vfree(buffer);
		return -result;
	}

	return result;
}

static int ioctl_update_status(struct syncD_virtdev *virtdev, struct syn1588_ioctl_update_status_reponse  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct syn1588_ioctl_update_status_reponse response;

	struct oregano_updater_altera *updater = get_updater(syncd_dev);

	dev_warn(dev, "start update of PCIE card\n");
	if(IS_ERR_OR_NULL(updater)) {
		return -ENODEV;
	}

	response = oregano_updater_altera_update_status(updater);

	if(copy_to_user(arg, &response, sizeof(response)) !=0) {
		dev_err(dev, "could not copy buffer to userspace");
		return -EINVAL;
	}

	return 0;
}

static int ioctl_read_serial_nr(struct syncD_virtdev *virtdev, struct syn1588_ioctl_buffer  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct syn1588_ioctl_buffer tmp;
	if(copy_from_user(&tmp, arg, sizeof(struct syn1588_ioctl_buffer))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	if(strlen(syncd_dev->syn1588_dev.serial_nr) >= tmp.size) {
		return -ENOMEM;
	}
	return  copy_to_user(tmp.data, syncd_dev->syn1588_dev.serial_nr, strlen(syncd_dev->syn1588_dev.serial_nr)+1) ? -EFAULT : 0;
}

static int ioctl_period_alloc(struct syncD_virtdev *virtdev, struct syn1588_ioctl_period_config  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_period_config tmp;
	int error = 0;
	if(copy_from_user(&tmp, arg, sizeof(struct syn1588_ioctl_period_config))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}


	error = clock_m_period_setup(clock_m, tmp.id, tmp.high_time, tmp.low_time, tmp.is_rising_edge, tmp.start_condition);
	if(error) {
		return error;
	}

	set_bit(tmp.id, virtdev->used_period_ids);

	return 0;
}


static int ioctl_period_free(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_source source = {.type = SYN1588_IOCTL_SOURCE_PERIOD};
	u32 id;
	if(get_user(id, arg)){
		return -EFAULT;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	if(!test_bit(id,virtdev->used_period_ids)) {
		dev_err(dev, "period not owned by file_descriptor\n");
		return -EINVAL;
	}
	source.id = id;
	clock_m_io_matrix_remove_source(clock_m, source);

	clock_m_period_clear(clock_m, id);

	clear_bit(id, virtdev->used_period_ids);
	return 0;

}

static int ioctl_period_set_enable(struct syncD_virtdev *virtdev, struct syn1588_ioctl_period_set_enable  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_period_set_enable tmp;
	if(copy_from_user(&tmp, arg, sizeof(tmp))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	if(!test_bit(tmp.id,virtdev->used_period_ids)) {
		dev_err(dev, "period not owned by file_descptor\n");
		return -EINVAL;
	}
	clock_m_period_set_enable(clock_m, tmp.id, tmp.is_enabled);

	return 0;
}

static int ioctl_period_get_duration(struct syncD_virtdev *virtdev, struct syn1588_ioctl_period_get_duration  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_period_get_duration tmp;

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	if(copy_from_user(&tmp, arg, sizeof(tmp))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	if(tmp.id >= __SYN1588_IOCTL_PERIOD_NBITS) {
		return -EINVAL;
	}

	if(!test_bit(tmp.id,virtdev->used_period_ids)) {
		dev_err(dev, "period not owned by file_descptor\n");
		return -EINVAL;
	}
	clock_m_period_get_durations(clock_m, tmp.id, &tmp.high_time, &tmp.low_time);
	if(copy_to_user(arg, &tmp, sizeof(tmp))!=0) {
		pr_err("could not copy arg to userspace");
		return -EINVAL;
	}

	return 0;
}

static int ioctl_event_alloc(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	u32 event_id;
	int error = 0;
	if(copy_from_user(&event_id, arg, sizeof(event_id))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}


	error = clock_m_event_setup(clock_m, event_id, &virtdev->irq_waitq);
	if(error) {
		return error;
	}

	virtdev->event_id = event_id;

	return 0;
}


static int ioctl_event_free(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_sink sink = {.type = SYN1588_IOCTL_SINK_EVENT};
	u32 id;
	int ret = 0;
	if(get_user(id, arg)){
		return -EFAULT;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	if(id != virtdev->event_id) {
		dev_err(dev, "event not owned by file_descptor\n");
		return -EINVAL;
	}
	sink.id = id;
	(void)clock_m_io_matrix_remove_sink(clock_m, sink);


	ret = clock_m_event_free(clock_m, id);
	if(ret != 0) {
		dev_err(dev, "could not clear event\n");
		return ret;
	}

	virtdev->event_id = -1;
	return 0;

}


static int ioctl_event_get_ts(struct syncD_virtdev *virtdev, struct syn1588_ioctl_event_read_action  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_event_read_action tmp;
	u32 read_ts_count = 0;
	syn1588_ioctl_timepoint *timestamp_buff;
	if(copy_from_user(&tmp, arg, sizeof(tmp))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	if(tmp.id != virtdev->event_id) {
		return -EINVAL;
	}

	timestamp_buff = vmalloc(sizeof(*timestamp_buff) * tmp.bufsize);
	if(timestamp_buff == NULL) {
		return -ENOMEM;
	}
	while(tmp.bufsize > read_ts_count) {
		syn1588_ioctl_timepoint * timestamp = &timestamp_buff[read_ts_count];
		int error = clock_m_event_get_ts(clock_m, tmp.id, timestamp);

		if(error == -EAGAIN) {
			break;
		}
		if(error) {
			vfree(timestamp_buff);
			return error;
		}
		read_ts_count++;
	}
	// copy leftover
	if(read_ts_count != 0) {
		if(copy_to_user(
			tmp.buf,
			timestamp_buff,
			sizeof(*timestamp_buff) * read_ts_count)
		) {
			pr_err("could not save timestamp in user buffer\n");
			vfree(timestamp_buff);
			return -EINVAL;
		}
	}
	vfree(timestamp_buff);

	if(copy_to_user(&arg->bufsize, &read_ts_count, sizeof(read_ts_count))) {
		pr_err("could not write userspace read_ts_count\n");
		return -EINVAL;
	}

	return 0;
}

static int ioctl_event_clear(struct syncD_virtdev *virtdev, u32 __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct clock_m_device *clock_m = NULL;
	u32 event_id;
	if(copy_from_user(&event_id, arg, sizeof(event_id))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	if(event_id != virtdev->event_id) {
		return -EINVAL;
	}

	return clock_m_event_clear(clock_m, (enum syn1588_ioctl_event_id)event_id);
}

static int ioctl_trigger_alloc(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	u32 trigger_id;
	int error = 0;
	if(copy_from_user(&trigger_id, arg, sizeof(trigger_id))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}


	error = clock_m_trigger_setup(clock_m, trigger_id);
	if(error) {
		return error;
	}

	set_bit(trigger_id, virtdev->used_trigger_ids);

	return 0;
}


static int ioctl_trigger_free(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_source source = {.type = SYN1588_IOCTL_SOURCE_TRIGGER};
	u32 id;
	if(get_user(id, arg)){
		return -EFAULT;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	if(!test_bit(id,virtdev->used_trigger_ids)) {
		dev_err(dev, "trigger not owned by file_descptor\n");
		return -EINVAL;
	}

	source.id = id;
	clock_m_io_matrix_remove_source(clock_m, source);

	clock_m_trigger_clear(clock_m, id);

	clear_bit(id, virtdev->used_trigger_ids);
	return 0;

}


static int ioctl_trigger_add_actions(struct syncD_virtdev *virtdev, struct syn1588_ioctl_trigger_add_actions  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_trigger_add_actions tmp;
	u32 wrote_ts_count = 0;
	u32 checked_ts_count = 0;
	struct syn1588_ioctl_trigger_entry former_entry = {};

	if(copy_from_user(&tmp, arg, sizeof(tmp))!=0) {
		pr_err("could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	if(!test_bit(tmp.id,virtdev->used_trigger_ids)) {
		dev_err(dev, "trigger not owned by file_descriptor\n");
		return -EINVAL;
	}

	while (tmp.bufsize > checked_ts_count) {
		struct syn1588_ioctl_trigger_entry entry;
		if(copy_from_user(&entry, &tmp.buf[checked_ts_count], sizeof(entry))) {
			pr_err("could not save timestamp from user buffer\n");
			return -EINVAL;
		}
		if(entry.ts.nsec >= (u32)1e9) {
            		pr_err("nanoseconds are bigger than 1e9 - 1\n");
			return -EINVAL;
		}
		if(checked_ts_count != 0 &&
		  (former_entry.ts.sec > entry.ts.sec ||
		  (former_entry.ts.sec == entry.ts.sec &&
		   former_entry.ts.nsec >= entry.ts.nsec)))
		{
			pr_err("timestamps arent unique or chronologically ordered\n");
			return -EINVAL;
		}
		former_entry = entry;
		++checked_ts_count;
	}

	while(tmp.bufsize > wrote_ts_count) {
		struct syn1588_ioctl_trigger_entry entry;
		int error;
		if(copy_from_user(&entry, &tmp.buf[wrote_ts_count], sizeof(entry))) {
			pr_err("could not save timestamp from user buffer\n");
			return -EINVAL;
		}
		error = clock_m_trigger_add_action(clock_m, tmp.id, &entry);
		if(error == -EBUSY) {
			break;
		}
		if(error) {
			return error;
		}

		wrote_ts_count++;
	}

	if(copy_to_user(&arg->bufsize, &wrote_ts_count, sizeof(wrote_ts_count))) {
		pr_err("could not write userspace wrote_ts_count\n");
		return -EINVAL;
	}

	return 0;
}

static int ioctl_io_port_alloc(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	u32 io_port_id;
	int error = 0;
	if(copy_from_user(&io_port_id, arg, sizeof(io_port_id))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}


	error = clock_m_io_port_setup(clock_m, io_port_id);
	if(error) {
		return error;
	}

	set_bit(io_port_id, virtdev->used_io_port_ids);

	return 0;
}

static int ioctl_io_port_free(struct syncD_virtdev *virtdev, u32  __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_source source = {.type = SYN1588_IOCTL_SOURCE_IO_PORT};
	struct syn1588_ioctl_sink sink = {.type = SYN1588_IOCTL_SINK_IO_PORT};
	u32 id;
	if(get_user(id, arg)){
		return -EFAULT;
	}
	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}
	if(!test_bit(id,virtdev->used_io_port_ids)) {
		dev_err(dev, "io_port not owned by file_descptor\n");
		return -EINVAL;
	}

	source.id = id;
	clock_m_io_matrix_remove_source(clock_m, source);
	sink.id = id;
	clock_m_io_matrix_remove_sink(clock_m, sink);

	clock_m_io_port_clear(clock_m, id);

	clear_bit(id, virtdev->used_io_port_ids);
	return 0;

}

static bool source_owned_by_process(struct syncD_virtdev *virtdev, struct syn1588_ioctl_source source) {

	// pr_err("DEBUG: %d\n", current->pid);
	//TODO
	return true;
}
static bool sink_owned_by_process(struct syncD_virtdev *virtdev, struct syn1588_ioctl_sink sink) {
	//TODO
	return true;
}

static int ioctl_io_matrix_connect(struct syncD_virtdev *virtdev, struct syn1588_ioctl_connection __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_connection connection;
	if(copy_from_user(&connection, arg, sizeof(connection))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	if(!source_owned_by_process(virtdev, connection.source) || !sink_owned_by_process(virtdev, connection.sink)) {
		return -EPERM;
	}

	return clock_m_io_matrix_connect(clock_m, connection.source, connection.sink);
}

static int ioctl_io_matrix_disconnect(struct syncD_virtdev *virtdev, struct syn1588_ioctl_sink __user *arg) {
	struct syncD_device * syncd_dev = virtdev->syncd_dev;
	struct device *dev = syncd_dev->dev;
	struct clock_m_device *clock_m = NULL;
	struct syn1588_ioctl_sink sink;
	if(copy_from_user(&sink, arg, sizeof(sink))!=0) {
		dev_err(dev, "could not copy arg from userspace");
		return -EINVAL;
	}

	clock_m = get_clock_m(syncd_dev);
	if(IS_ERR_OR_NULL(clock_m)) {
		return -ENODEV;
	}

	return clock_m_io_matrix_remove_sink(clock_m, sink);
}

static long syncD_ioctl(struct file *filp, unsigned int cmd, unsigned long arg) {
	struct syncD_virtdev *virtdev = filp->private_data; // virtual device

	/* Check type and command number */
	if (_IOC_TYPE(cmd) != SYN1588_IOC_MAGIC)
		return -ENODEV;

	switch(cmd) {
	// general
	case SYN1588_IOC_RD_BUILD_ID:
		return ioctl_read_build_id(virtdev, (u32 __user *)arg);
	case SYN1588_IOC_RD_PRODUCT_ID:
		return ioctl_read_product_id(virtdev, (struct syn1588_ioctl_buffer  __user *)arg);
	// clock_m
	case SYN1588_CLOCK_IOC_RD_PHC_ID:
		return ioctl_read_get_phc_id(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_RD_CLOCK_ID:
		return ioctl_read_clock_id(virtdev, (u8 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_PERIOD_ALLOC:
		return ioctl_period_alloc(virtdev, (struct syn1588_ioctl_period_config __user *)arg);
	case SYN1588_CLOCK_IOC_WR_PERIOD_FREE:
		return ioctl_period_free(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_PERIOD_SET_ENABLE:
		return ioctl_period_set_enable(virtdev, (struct syn1588_ioctl_period_set_enable __user *)arg);
	case SYN1588_CLOCK_IOC_WR_PERIOD_GET_DURATION:
		return ioctl_period_get_duration(virtdev, (struct syn1588_ioctl_period_get_duration __user *)arg);
	case SYN1588_CLOCK_IOC_WR_EVENT_ALLOC:
		return ioctl_event_alloc(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_EVENT_FREE:
		return ioctl_event_free(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_EVENT_READ_TS:
		return ioctl_event_get_ts(virtdev, (struct syn1588_ioctl_event_read_action __user *)arg);
	case SYN1588_CLOCK_IOC_WR_EVENT_CLEAR:
		return ioctl_event_clear(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_TRIGGER_ALLOC:
		return ioctl_trigger_alloc(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_TRIGGER_FREE:
		return ioctl_trigger_free(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_TRIGGER_ADD_ACTIONS:
		return ioctl_trigger_add_actions(virtdev, (struct syn1588_ioctl_trigger_add_actions __user *)arg);
	// io matrix
	case SYN1588_CLOCK_IOC_WR_IO_PORT_ALLOC:
		return ioctl_io_port_alloc(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_IO_PORT_FREE:
		return ioctl_io_port_free(virtdev, (u32 __user *)arg);
	case SYN1588_CLOCK_IOC_WR_IO_MATRIX_CONNECT:
		return ioctl_io_matrix_connect(virtdev, (struct syn1588_ioctl_connection __user *)arg);
	case SYN1588_CLOCK_IOC_WR_IO_MATRIX_DISCONNECT:
		return ioctl_io_matrix_disconnect(virtdev, (struct syn1588_ioctl_sink __user *)arg);
	// pcie_nicl
	case SYN1588_PCI_IOC_WR_FLASH_AND_VERIFY:
		return ioctl_flash_and_verify(virtdev, (struct syn1588_ioctl_buffer_const __user *)arg);
	case SYN1588_PCI_IOC_RD_UPDATE_STATUS:
		return ioctl_update_status(virtdev, (struct syn1588_ioctl_update_status_reponse __user *)arg);

	case SYN1588_IOC_RD_SERIAL_NR:
		return ioctl_read_serial_nr(virtdev, (struct syn1588_ioctl_buffer  __user *)arg);
	}


	return -ENOIOCTLCMD;
}


/**
 * checks, whether the requested address range in the device file is valid
 * \param count number of bytes to be accessed from the current position on
 * \param f_pos current position within the device file
 * \return 0 on success, negative error code otherwise
 */
static int syncD_fpos_chk(size_t *count, loff_t *f_pos)
{
	if (0 != *f_pos % sizeof(u32)) {
		pr_err("access to file position 0x%X not 4Byte aligned.\n",
			 (unsigned int)*f_pos);
		return -EINVAL;
	}

	return 0;
}



static u32 syncD_read_nic(struct syncD_device *syncD_dev, loff_t addr) {
	// always use first mac for legacy interface so singl nic works fine
	struct nic_priv_data *nic_dat = get_mac_by_index(syncD_dev, 0);

	if(nic_dat != NULL) {
		return oregano_mac_readl(nic_dat, addr);

	}
	pr_warn("read faild, nic not existing %llx\n", addr);
	return 0;
}

/**
 * function to handle file read requests from user-space
 * \param filp pointer to file structure (file descriptor)
 * \param buf pointer to buffer, the read data shall be stored in
 * \param count number of bytes to be read
 * \param f_pos start position in the device file (register address)
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
static ssize_t syncD_read(struct file *filp, char *buf, size_t count,
			  loff_t *f_pos)
{
	struct syncD_virtdev *virtdev = filp->private_data;
	struct syncD_device *syncD_dev = virtdev->syncd_dev;
	struct device *dev = syncD_dev->miscdev.this_device;
	struct clock_m_device *clock_m = get_clock_m(syncD_dev);
	loff_t address;
	u32 data = 0;
	int ret;
	// pointer to clock_m for extime functions
	struct clock_m_device *extime_clock_m = clock_m;

	if(clock_m->extime_clock_m != NULL) {
		extime_clock_m = clock_m->extime_clock_m;
	}

	if (unlikely(!virtdev || !syncD_dev)) { // device was not opened
		dev_err(dev, "write to non-existent virtual device 0x%p.\n", filp);
		return -EBADF;
	}

	if (O_WRONLY == (filp->f_flags & O_ACCMODE)) {
		dev_dbg(dev, "read to write-only virtual device.\n");
		return -EACCES;
	}

	// verify that f_pos and count are in address range
	if (0 > (ret = syncD_fpos_chk(&count, f_pos)))
		return ret;

	if (filp->f_flags & O_NONBLOCK) {
		if (down_read_trylock(&virtdev->access_sem) == 0)
			return -EWOULDBLOCK;
	} else
		down_read(&virtdev->access_sem);

	for (address = *f_pos; address < *f_pos + count;
		 address += sizeof(u32)) {
		switch (address) {
		case IRSRC:
			if (!(filp->f_flags & O_NONBLOCK)) {
				dev_dbg(dev, "IRSRC blocking read for PID %u.\n",
					 virtdev->pid);
				// only block if IRSRC is empty (read is consuming!)
				if (virtdev->irsrc == 0) {
					ret = wait_event_interruptible(
						virtdev->irq_waitq,
						virtdev->irsrc);
					if (ret < 0) {
						dev_info(dev,
							 "IRSRC blocking read for PID %u interrupted.\n",
							 virtdev->pid);
						up_read(&virtdev->access_sem);
						return ret;
					}
				}
			}
			spin_lock_bh(&syncD_dev->virtdev_lock);
			data = virtdev->irsrc; // save IRSRC register
			virtdev->irsrc = 0x0; // reading virt. IRSRC consumes
			ret = virtdev->irq_fail; // query if multiple interrupted occured
			virtdev->irq_fail =
				0x0; // reset multiple interrupt flag
			spin_unlock_bh(&syncD_dev->virtdev_lock);
			dev_dbg(dev, "read access to IRSRC (data 0x%X).\n",
				 data);
			break;

		case IREN:
			ret = clock_m_getreg(get_clock_m(syncD_dev), IREN, &data);
			if (ret < 0) {
				dev_err(dev, "read access to IREN failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			spin_lock_bh(&syncD_dev->virtdev_lock);
			{ // modify register readout to virtdevice enabled IRQs and unused
				// (not by other virtual devices and not by driver) ones
				u32 mask = 0;
				struct syncD_virtdev *entry;

				list_for_each_entry(
					entry, &syncD_dev->virtdev_list, list)
					mask |= entry->iren;
				data &= virtdev->iren | (~mask);
			}
			spin_unlock_bh(&syncD_dev->virtdev_lock);
			dev_dbg(dev, "read access to IREN (data 0x%X).\n", data);
			break;

		case TIMECTRL:
			if (0 > (ret = clock_m_getreg(extime_clock_m,
							 TIMECTRL, &data))) {
				dev_err(dev, "read access to TIMECTRL failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			//			 spin_lock(&syncD_dev->virtreg_lock);
			//			 data |= syncD_dev->timectrl; // add virtual time control
			//			 register (only virtual bits are set)
			//			 spin_unlock(&syncD_dev->virtreg_lock);
			dev_dbg(dev, "read access to TIMECTRL (data 0x%X).\n",
				 data);
			break;
		case SHDWSTEPPU_L:
		case SHDWSTEPPU_H:
		case SHDWTIME_NS:
		case SHDWTIME_SEC:
			if (0 > (ret = clock_m_getreg(extime_clock_m,
							 address, &data))) {
				pr_err("read access to %x failed.\n", (u32)address);
				up_read(&virtdev->access_sem);
				return ret;
			}
			dev_dbg(dev, "read access to %x (data 0x%X).\n",
				 (u32)address, data);
			break;

#ifdef STATUS
		case STATUS:
			if (0 > (ret = clock_m_getreg(get_clock_m(syncd_dev),
							 STATUS, &data))) {
				dev_err(dev, "read access to STATUS failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			/*					// set TISB_FCLK bit if synchronous amortisation is running
							data |= (0x1 && ((off != syncD_dev->sAmortState)
																&& (finished !=
					 syncD_dev->sAmortState)))<<TISB_FCLK; */
			dev_dbg(dev, "read access to STATUS (data 0x%X).\n",
				 data);
			break;
#endif

		case EVENTCTRL:
			ret = clock_m_getreg(get_clock_m(syncD_dev), EVENTCTRL,
						&data);
			if (ret < 0) {
				dev_err(dev, "read access to EVENTCTRL failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			dev_dbg(dev, "read access to EVENTCTRL (data 0x%X).\n",
				 data);
			break;
		case TIME_L:
			// deprecated register, kept for downwards compaibility
			data = 0;
			dev_dbg(dev, "read access to TIME_L (data 0x%X).\n",
				 data);
			break;

		case TIME_NS:
			spin_lock_bh(&syncD_dev->virtdev_lock);
			ret = clock_m_getreg(get_clock_m(syncD_dev), TIME_NS,
						&data);
			ret |= clock_m_getreg(get_clock_m(syncD_dev), TIME_SEC,
						 &virtdev->time_h);
			spin_unlock_bh(&syncD_dev->virtdev_lock);

			if (ret < 0) {
				dev_err(dev, "read access to TIME failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			dev_dbg(dev, "read access to TIME_NS (data 0x%X).\n",
				 data);
			break;

		case TIME_SEC:
			data = virtdev->time_h;
			dev_dbg(dev, "read access to TIME_SEC (data 0x%X).\n",
				 data);
			break;

		case TIME2_M:
			spin_lock_bh(&syncD_dev->virtdev_lock);
			ret = clock_m_getreg(get_clock_m(syncD_dev), TIME2_M,
						&data);
			ret |= clock_m_getreg(get_clock_m(syncD_dev), TIME2_H,
						 &virtdev->time2_h);
			spin_unlock_bh(&syncD_dev->virtdev_lock);

			if (ret < 0) {
				dev_err(dev, "read access to TIME2 failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
			dev_dbg(dev, "read access to TIME2_M (data 0x%X).\n",
				 data);
			break;

		case TIME2_H:
			data = virtdev->time2_h;
			dev_dbg(dev, "read access to TIME2_H (data 0x%X).\n",
				 data);
			break;

		case VR_SYSTS_L: {
			unsigned long long delta_sys, delta_hw;
			syn1588_timespec sys_ts, hw_ts;

			// we aquire a spinlock here
			// this also implicitly disables preemption
			spin_lock_bh(&syncD_dev->virtdev_lock);
			/* read system, read hardware, check TSC delta*/

			/* read system read hw
			 * store values for SYSV_H, TIMEV_M, TIMEV_H
			 *
			 */

			clock_m_getTimestamps(get_clock_m(syncD_dev), &sys_ts,
					  &hw_ts, &delta_sys, &delta_hw);

			virtdev->systs_l = sys_ts.tv_nsec;
			virtdev->systs_h = sys_ts.tv_sec;

			virtdev->hwts_l = hw_ts.tv_nsec;
			virtdev->hwts_h = hw_ts.tv_sec;

			// return values for TIMEV2_M and TIMEV2_h

			// convert 64bit deltas to 32bit values
			// should be fine for the deltas. they should be small enough
			virtdev->delta_sys = delta_sys & 0xFFFFFFFF;
			virtdev->delta_hw = delta_hw & 0xFFFFFFFF;

			// return value of SYSTS_L register
			data = virtdev->systs_l;
			spin_unlock_bh(&syncD_dev->virtdev_lock);

			if (ret < 0) {
				dev_err(dev, "read access to SYSV_M failed.\n");
				up_read(&virtdev->access_sem);
				return ret;
			}
		} break;
		case CLK_FREQ:
		case VR_CLK_FREQ:
			data = clock_m->clk_freq;
			break;
		case VR_ETH_TSFIFORX:
			// FAKEFIFO is depricated
			data = 0;
			break;
		case VR_ETH_TSFIFOTX:
			// FAKEFIFO is depricated
			data = 0;
			break;
		case VR_ETH_TSFIFO_STATUS:
			// FAKEFIFO is depricated
			data = 1 << 16 | 1; // return allways empty
			break;
		case VR_SYSTS_H:
			data = virtdev->systs_h;
			break;

		case VR_HWTS_L:
			data = virtdev->hwts_l;
			break;

		case VR_HWTS_H:
			data = virtdev->hwts_h;
			break;

		case VR_DELTASYS:
			data = virtdev->delta_sys;
			break;

		case VR_DELTAHW:
			data = virtdev->delta_hw;
			break;

		case HWBUILD:
		case VR_ETH_BUILD:
			// Eth_Build id is no longer existing
			// The build id itself got arbitary per device
			// use the new ioctl api to get the "new" build id
			if ( syncD_dev->syn1588_dev.build_id > 5000 ) {
				//new rev2.3 nic fake the id for the userspace
				data = 999;
			} else {
				data = syncD_dev->syn1588_dev.build_id;
			}
			break;

		case VR_ETH_VERSION:
			// since the PTP-Stack has no access to the MAC-Register-Space to get the
			// MAC_ETH_ETHERNET register, we map that register here.
			data = syncD_read_nic(syncD_dev, MAC_ETH_VERSION);
			break;

		case VR_ETH_TS_TXCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TS TXCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS_TXCTRL);
			break;
		case VR_ETH_TS2_TXCTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register TS TXCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS2_TXCTRL);
			break;
		case VR_ETH_TS_RXCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TS RXCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS_RXCTRL);
			break;
		case VR_ETH_TS2_RXCTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register TS RXCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS2_RXCTRL);
			break;
		case VR_ETH_PHY_DELAY_TX:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register PHY DELAY TX to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS_PHY_DLY_TX);
			break;
		case VR_ETH_PHY_DELAY_RX:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register PHY DELAY RX to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS_PHY_DLY_RX);
			break;
		case VR_ETH_TIMECTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TIMECTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS_CTRL);
			break;
		case VR_ETH_1STEPCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register MAC_1STEPCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_1STEPCTRL);
			break;
		case VR_ETH_TS2_1STEPCTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register MAC_TS2_1STEPCTRL to the MAC
			data = syncD_read_nic(syncD_dev, MAC_TS2_1STEPCTRL);
			break;
		case VR_ETH_RXERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_RXERR);
			break;

		case VR_ETH_FCSERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_FCSERR);
			break;

		case VR_ETH_MCASTERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_MCASTERR);
			break;

		case VR_ETH_LCOLLERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_LCOLLERR);
			break;

		case VR_ETH_LENERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_LENERR);
			break;

		case VR_ETH_OVERERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_OVERERR);
			break;

		case VR_ETH_FRAMEERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_FRAMEERR);
			break;

		case VR_ETH_RX:
			data = syncD_read_nic(syncD_dev, MAC_CNT_RX);
			break;

		case VR_ETH_TX:
			data = syncD_read_nic(syncD_dev, MAC_CNT_TX);
			break;

		case VR_ETH_CNT_VLANERR:
			data = syncD_read_nic(syncD_dev, MAC_CNT_VLAN_ERR);
			break;
		case VR_ETH_CNT_VLAN:
			data = syncD_read_nic(syncD_dev, MAC_CNT_VLAN);
			break;
		case VR_ETH_CNT_BROADCAST:
			data = syncD_read_nic(syncD_dev, MAC_CNT_BROADCAST);
			break;
		case VR_ETH_CNT_MULTICAST:
			data = syncD_read_nic(syncD_dev, MAC_CNT_MULTICAST);
			break;
		case VR_ETH_CNT_RX_TS:
			data = syncD_read_nic(syncD_dev, MAC_CNT_RX_TS);
			break;
		case VR_ETH_CNT_TX_TS:
			data = syncD_read_nic(syncD_dev, MAC_CNT_TX_TS);
			break;
		case VR_ETH_CNT_RX_TS2:
			data = syncD_read_nic(syncD_dev, MAC_CNT_RX_TS2);
			break;
		case VR_ETH_CNT_TX_TS2:
			data = syncD_read_nic(syncD_dev, MAC_CNT_TX_TS2);
			break;
	#ifdef USE_MBG_HPS
		case VR_TSU_PHY_INFO:
			data = mbg_get_tsu_phy_info(clock_m_dev);
			break;
	#endif
		case VR_ETH_CAPABILITIES_0:
			data = syncD_read_nic(syncD_dev, MAC_CAPABILITY_0);
			break;
		case VR_ETH_MIICOMMAND:
			data = syncD_read_nic(syncD_dev, MAC_MIICOMMAND);
			break;
		case VR_ETH_SFP_EEPROM:
		 	// removed as not needed anymore and new useage register address
			data = 0;
			break;
		case VR_ETH_CAPABILITIES_1:
			data = syncD_read_nic(syncD_dev, MAC_CAPABILITY_1);
			break;
		case VR_ETH_CAPABILITIES_2:
			data = syncD_read_nic(syncD_dev, MAC_CAPABILITY_2);
			break;
		case VR_ETH_CAPABILITIES_3:
			data = syncD_read_nic(syncD_dev, MAC_CAPABILITY_3);
			break;
		case VR_ETH_IFG_CTRL:
			data = syncD_read_nic(syncD_dev, MAC_IFG_CTRL);
			break;
		default:
			if(address >= 0x3000 && address <= 0x303F) {
				// old spi interface
				u32 relative_address = address - 0x3000;
				if(!IS_ERR_OR_NULL(syncD_dev->iobase_spi)) {
					data = ioread32(syncD_dev->iobase_spi + relative_address);
				}
			}else if(address >= 0x30C0 && address <= 0x3100) {
				// old i2c interface
				u32 relative_address = address - 0x30C0;
				if(!IS_ERR_OR_NULL(syncD_dev->iobase_i2c_video)) {
					data = ioread32(syncD_dev->iobase_i2c_video + relative_address);
				}
			}else if (address >= CLOCK_M_MAC_SER_NUM_ADDR_RANGE_L &&
				address < CLOCK_M_MAC_SER_NUM_ADDR_RANGE_H) {
				u32 relative_address = address - CLOCK_M_MAC_SER_NUM_ADDR_RANGE_L;
				if(!IS_ERR_OR_NULL(syncD_dev->iobase_eeprom)) {
					data = ioread32(syncD_dev->iobase_eeprom + relative_address);
				}
			} else if (0 > (ret = clock_m_getreg(get_clock_m(syncD_dev),
						address, &data))) {
				dev_err(dev, "read from address 0x%llX failed.\n",
				address);
				up_read(&virtdev->access_sem);
				return ret;
			}
			dev_dbg(dev, "read data (0x%08X) from register 0x%08X\n",
			(unsigned int)data, (unsigned int)address);
		}
		// TODO: verify byte count or limit byte count to double words
		if (0 !=
			copy_to_user(buf + address - *f_pos, &data, sizeof(data))) {
			dev_warn(dev,
				 "writing data to user address 0x%lX failed.\n",
				 (unsigned long)buf);
			up_read(&virtdev->access_sem);
			return -EFAULT;
		}
		//		spin_lock(&syncD_dev->virtreg_lock);
		if ((address == IRSRC) && (ret != 0)) {
			//			spin_unlock(&syncD_dev->virtreg_lock);
			up_read(&virtdev->access_sem);
			return -EIO; // fail if multiple interrupts occured since last readout
		}
		//		spin_unlock(&syncD_dev->virtreg_lock);
	}
	dev_dbg(dev, "read addr: 0x%04lx data: 0x%08lx\n", (unsigned long)address, (unsigned long)data);
	*f_pos += count;
	up_read(&virtdev->access_sem);

	return count;
}


void syncD_write_nic(struct syncD_device *syncD_dev, u32 data, loff_t addr) {
	// always use first mac for legacy interface so singl nic works fine
	struct nic_priv_data *nic_dat = get_mac_by_index(syncD_dev, 0);

	if(nic_dat != NULL) {
		oregano_mac_writel(nic_dat, data, addr);
	}else {
		pr_warn("write faild, nic not existing %llx %x\n", addr, data);
	}
}

/**
 * function to handle file write requests from user-space
 * \param filp pointer to file structure (file descriptor)
 * \param buf pointer to constant buffer, containing the data to write
 * \param count number of bytes to be write
 * \param f_pos start position in the device file (register address)
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
static ssize_t syncD_write(struct file *filp, const char *buf, size_t count,
			   loff_t *f_pos)
{
	struct syncD_virtdev *virtdev = filp->private_data;
	struct syncD_virtdev *entry;
	struct syncD_device *syncD_dev = virtdev->syncd_dev;
	struct device *dev = syncD_dev->dev;
	struct clock_m_device *clock_m = get_clock_m(syncD_dev);
	unsigned long address;
	u32 data;
	int ret;
	// pointer to clock_m for extime functions, by default itself
	struct clock_m_device *extime_clock_m = clock_m;

	if(clock_m->extime_clock_m != NULL) {
		extime_clock_m = clock_m->extime_clock_m;
	}

	if (unlikely(!virtdev || !syncD_dev)) { // device was not opened
		dev_err(dev, "write to non-existent virtual device.\n");
		return -EBADF;
	}

	if (!(filp->f_mode & FMODE_WRITE)) {
		dev_dbg(dev, "write to read-only virtual device.\n");
		return -EACCES;
	}

	// verify that f_pos and count are in address range
	if (0 != (ret = syncD_fpos_chk(&count, f_pos)))
		return ret;

	if (0 != count % sizeof(u32)) {
		dev_info(dev, "write byte count not double word aligned!\n");
		return -EINVAL;
	}

	if (filp->f_flags & O_NONBLOCK) {
		if (down_read_trylock(&virtdev->access_sem) == 0)
			return -EWOULDBLOCK;
	} else
		down_read(&virtdev->access_sem);

	// run through specified file position in sizeof(u32) intervals and write data
	for (address = (unsigned long)*f_pos; address < *f_pos + count;
		 address += sizeof(u32)) {
		if (0 != copy_from_user(&data, buf - *f_pos + address,
					sizeof(data))) {
			dev_warn(dev,
				 "fetching data from user address 0x%lX failed.\n",
				 (unsigned long)buf);
			up_read(&virtdev->access_sem);
			return -EFAULT;
		}
		switch (address) {
		case IRSRC:
			dev_dbg(dev, "write access to IRSRC (data 0x%X).\n",
				 data);
			spin_lock_bh(&syncD_dev->virtdev_lock);
			virtdev->irsrc &=
				data; // writing a 1 doesn't influence register content
			spin_unlock_bh(&syncD_dev->virtdev_lock);
			break;

		case IREN:
			// note that TIMECTRL and IRQ handler also write to IREN
			dev_dbg(dev, "write access to IREN (data 0x%X).\n",
				 data);

			spin_lock_bh(&syncD_dev->virtdev_lock);
			virtdev->iren = data;

			// generate logical or of all virtual IREN registers
			list_for_each_entry(entry, &syncD_dev->virtdev_list,
						 list)
				data |= entry->iren;
			spin_unlock_bh(&syncD_dev->virtdev_lock);

			// write to device register
			if (clock_m_setreg(clock_m, IREN, data) != 0) {
				dev_err(dev, "write to IREN (data 0x%X) failed.\n",
					 data);
				up_read(&virtdev->access_sem);
				return -EINVAL;
			}
			dev_dbg(dev, "writen IREN (data 0x%X).\n", data);
			break;

		case TIMECTRL:
			dev_dbg(dev, "write access to TIMECTRL (data 0x%X).\n",
				 data);

			// write to device register
			if (clock_m_setreg(extime_clock_m, TIMECTRL, data) != 0) {
				pr_err("write to TIMECTRL (data 0x%X) failed.\n",
					 data);
				//				spin_unlock(&syncD_dev->virtreg_lock);
				up_read(&virtdev->access_sem);
				return -EINVAL;
			}

			//			spin_unlock(&syncD_dev->virtreg_lock);
			break;

		case SHDWSTEPPU_L:
		case SHDWSTEPPU_H:
		case SHDWTIME_NS:
		case SHDWTIME_SEC:
			if (0 > (ret = clock_m_setreg(extime_clock_m,
							 address, data))) {
				pr_err("read access to %x failed.\n", (u32)address);
				up_read(&virtdev->access_sem);
				return ret;
			}
			pr_debug("write access to %x (data 0x%X).\n",
				 (u32)address, data);
			break;

		case EVENTCTRL:
			dev_dbg(dev, "write access to EVENTCTRL (data 0x%X).\n",
				 data);
			//			 spin_lock(&syncD_dev->virtreg_lock);
			//			 syncD_dev->eventctrl = data;
			//			 spin_unlock(&syncD_dev->virtreg_lock);
			if (clock_m_setreg(clock_m, EVENTCTRL, data) != 0) {
				dev_err(dev,
					 "write to EVENTCTRL (data 0x%X) failed.\n",
					 data);
				up_read(&virtdev->access_sem);
				return -EINVAL;
			}
			break;

		case VR_ETH_TS_TXCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TS TXCTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS_TXCTRL);
			break;
		case VR_ETH_TS2_TXCTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register TS TXCTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS2_TXCTRL);
			// no support for FAKEFIFO
			break;
		case VR_ETH_TS_RXCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TS RXCTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS_RXCTRL);
			break;
		case VR_ETH_TS2_RXCTRL:
			syncD_write_nic(syncD_dev, data, MAC_TS2_RXCTRL);
			// no support for FAKEFIFO
			break;
		case VR_ETH_PHY_DELAY_TX:
			// Since the user space has no access to the MAC register space, redirect
			syncD_write_nic(syncD_dev, data, MAC_TS_PHY_DLY_TX);
			break;
		case VR_ETH_PHY_DELAY_RX:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register PHY DELAY RX to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS_PHY_DLY_RX);
			break;
		case VR_ETH_TIMECTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TIMECTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS_CTRL);
			break;
		case VR_ETH_1STEPCTRL:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register TIMECTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_1STEPCTRL);
			break;
		case VR_ETH_TS2_1STEPCTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register TIMECTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS2_1STEPCTRL);
			break;
		case VR_ETH_IFG_CTRL:
			// Since the user space has no access to the MAC register space, redirect the
			// virtual register TIMECTRL to the MAC
			syncD_write_nic(syncD_dev, data, MAC_TS2_1STEPCTRL);
			break;
		case VR_ETH_MIICOMMAND:
			// Since the user space has no access to the MAC register space, redirect
			// the virtual register MIICOMMAND to the MAC
			syncD_write_nic(syncD_dev, data, MAC_MIICOMMAND);
			break;
		case VR_ETH_SFP_EEPROM:
			// removed as not needed anymore and new useage register address
			return -EINVAL;
		default:
			if(address >= 0x3000 && address <= 0x303F) {
				// old spi interface
				u32 relative_address = address - 0x3000;
				if(!IS_ERR_OR_NULL(syncD_dev->iobase_spi)) {
					iowrite32(data, syncD_dev->iobase_spi + relative_address);
				}
			}else if(address >= 0x30C0 && address <= 0x3100) {
				// old i2c interface
				u32 relative_address = address - 0x30C0;
				if(!IS_ERR_OR_NULL(syncD_dev->iobase_i2c_video)) {
					iowrite32(data, syncD_dev->iobase_i2c_video + relative_address);
				}
			}else if (address >= VR_ETH_TS_TXMEM_BASE &&
				address < (VR_ETH_TS_TXMEM_BASE + MAC_TS_MEM_SIZE)) {
				// MAC TS memory TX
				syncD_write_nic(syncD_dev, data,
					MAC_TS_TXMEM_BASE +
						(address - VR_ETH_TS_TXMEM_BASE));
			}  else if (address >= VR_ETH_TS2_TXMEM_BASE &&
				address <
					(VR_ETH_TS2_TXMEM_BASE + MAC_TS_MEM_SIZE)) {
				// MAC 2nd TS memory TX
				syncD_write_nic(syncD_dev, data,
					MAC_TS2_TXMEM_BASE +
						(address - VR_ETH_TS2_TXMEM_BASE));
			} else if (address >= VR_ETH_TS_RXMEM_BASE &&
				address < (VR_ETH_TS_RXMEM_BASE + MAC_TS_MEM_SIZE)) {
				// MAC TS memory RX
				syncD_write_nic(syncD_dev, data,
					MAC_TS_RXMEM_BASE +
						(address - VR_ETH_TS_RXMEM_BASE));
			} else if (address >= VR_ETH_TS2_RXMEM_BASE &&
				address <
					(VR_ETH_TS2_RXMEM_BASE + MAC_TS_MEM_SIZE)) {
				// MAC 2nd TS memory RX
				syncD_write_nic(syncD_dev, data,
					MAC_TS2_RXMEM_BASE +
						(address - VR_ETH_TS2_RXMEM_BASE));
			} else if (clock_m_setreg(clock_m, address,
						 data) != 0) {
				dev_err(dev,
					 "write to address 0x%lX (data 0x%X) failed.\n",
					 address, data);
				up_read(&virtdev->access_sem);
				return -EINVAL;
			}
			dev_dbg(dev,
				 "write access to address 0x%lX (data 0x%X).\n",
				 address, data);
		}
	}

	*f_pos += count;
	up_read(&virtdev->access_sem);

	return count;
}

/**
 * This function is called whenever the user-space program performs a poll,
 * select, or epoll systemcall
 * \param filp file pointer assotiated with this driver
 * \param wait waitqueue the user-process should be put into
 * \return POLLERR in case of an error, appropriate bitmask describing available
 * operations otherwise
 * \ingroup syncD_fops
 */
static unsigned int syncD_poll(struct file *filp, poll_table *wait)
{
	struct syncD_virtdev *virtdev = filp->private_data; // virtual device
	struct syncD_device *syncD_dev = virtdev->syncd_dev; //
	struct device *dev = syncD_dev->miscdev.this_device;
	unsigned int mask =
		POLLOUT | POLLWRNORM; // device is writeable (always)

	if (!virtdev) {
		// device wasnt opened
		dev_err(dev, "poll on non-existent virtual device 0x%p.\n", filp);
		return POLLERR;
	}
	if(virtdev->event_id>=0) {
		struct clock_m_device *clock_m = get_clock_m(syncD_dev);
		int available = clock_m_event_ts_available(clock_m, virtdev->event_id);
		if(available>0) {
			return mask | POLLIN | POLLRDNORM;
		}
		poll_wait(filp, &(virtdev->irq_waitq), wait);
		available = clock_m_event_ts_available(clock_m, virtdev->event_id);
		if(available>0) {
			return mask | POLLIN | POLLRDNORM;
		}
		return mask;
	}

	if (down_read_trylock(&syncD_dev->virtdev_list_sem) != 0) {
		poll_wait(filp, &(virtdev->irq_waitq), wait);

		spin_lock_bh(&syncD_dev->virtdev_lock); // FIXME! remove lock?!
		if (virtdev->irsrc != 0)
			mask |= POLLIN | POLLRDNORM; // device is readable
		spin_unlock_bh(&syncD_dev->virtdev_lock);

		up_read(&syncD_dev->virtdev_list_sem);
	} else{
		pr_warn("could not down_read_trylock");
	}

	dev_dbg(dev, "poll returns: 0x%X.\n", mask);
	return mask;
}

/**
 * function to handle file seek request from user-space (change current position
 * within the file)
 * \param filp pointer to file structure (file descriptor)
 * \param off offset within the file
 * \param whence type of see (set position to offst, increment position by
 * offset, behind the end)
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
static loff_t syncD_llseek(struct file *filp, loff_t off, int whence)
{
	struct syncD_virtdev *virtdev = filp->private_data;
	struct syncD_device *syncD_dev = virtdev->syncd_dev;
	struct device *dev = syncD_dev->miscdev.this_device;
	loff_t newpos;

	if (!virtdev) { // device is was not opened
		dev_err(dev, "seek on non-existent virtual device 0x%p.\n", filp);
		return -EBADF;
	}

	if (filp->f_flags & O_NONBLOCK) {
		if (down_read_trylock(&syncD_dev->virtdev_list_sem) == 0)
			return -EWOULDBLOCK;
	} else
		down_read(&syncD_dev->virtdev_list_sem);

	switch (whence) {
	case 0: // SEEK_SET
		newpos = off;
		break;

	case 1: // SEEK_CUR
		newpos = filp->f_pos + off;
		break;

	case 2: // SEEK_END
		newpos = CLOCK_M_ADDR_RANGE + off;
		break;

	default: // can't happen
		dev_notice(dev, "seek with illegal whence.\n");
		up_read(&syncD_dev->virtdev_list_sem);
		return -EINVAL;
	}
	if (newpos < 0) {
		dev_notice(dev, "seek to negative position.\n");
		up_read(&syncD_dev->virtdev_list_sem);
		return -ESPIPE;
	}
	dev_dbg(dev, "seek to position 0x%lX.\n", (unsigned long)newpos);
	filp->f_pos = newpos;
	up_read(&syncD_dev->virtdev_list_sem);
	return newpos;
}

/**
 * helper function to register asynchronous notification request from user-space
 * \param fd file descriptor
 * \param filp pointer to file structure (file descriptor)
 * \param mode internal kernel mode for fasync
 * \return 0 on success, negative error code otherwise
 * \ingroup syncD_fops
 */
int syncD_fasync(int fd, struct file *filp, int mode)
{
	struct syncD_virtdev *virtdev = filp->private_data;
	struct syncD_device *syncD_dev = virtdev->syncd_dev;
	struct device *dev = syncD_dev->miscdev.this_device;
	int ret;

	dev_dbg(dev, "registering signal notification.\n");
	if (0 >
		(ret = fasync_helper(fd, filp, mode, &(syncD_dev->async_queue))))
		dev_warn(dev, "signal notification register failed.\n");

	return ret;
}

/******************************************************************************
 * Module functions
 *****************************************************************************/


/**
 * create misc device file (one for each adatpter) to enable access from
 * user-space
 * and initialize virtual device list to allow multiple processes per device
 * file
 * \param work pointer to working task structure
 */
static void syncD_dev_init(struct syncD_device *newdev)
{
	struct device *dev = newdev->dev;

	dev_info(dev, "preparing device file.\n");
	newdev->dev_nr = syncd_device_alloc_id();
	snprintf(newdev->name, sizeof(newdev->name), DEV_NAME_TPL, newdev->dev_nr);

	newdev->miscdev.minor = MISC_DYNAMIC_MINOR;
	newdev->miscdev.name = newdev->name;
	newdev->miscdev.fops = &syncD_fops;

	// give read/write permissions on new device
	// for root and group
	newdev->miscdev.mode = 0664;

	if (misc_register(&newdev->miscdev)) {
		pr_err("cannot register miscdev on minor #%d.\n",
			 newdev->miscdev.minor);
		return;
	}
	syncD_device_alloc_minor(newdev);


	dev_dbg(dev, "device file %s setup (minor #%d) for device handle 0x%p complete.\n",
		 newdev->miscdev.name, newdev->miscdev.minor, get_clock_m(newdev));
}



/// @{ paramter parsing
static int syn1588_syncd_probe_pdata(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct oregano_syncd_platform_data *pdata = dev_get_platdata(dev);
	struct syncD_device *syncd = platform_get_drvdata(pdev);

	syncd->syn1588_dev = *pdata->syn1588_dev;
	syncd->updater_dev = &pdata->syn1588_dev->updater->dev;

	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id syn1588_syncd_dt_ids[] = {
	{ .compatible = DRV_NAME },
	{ .compatible = "oregano,syncd" },
	{}
};
MODULE_DEVICE_TABLE(of, syn1588_syncd_dt_ids);

static int syn1588_syncd_probe_dt(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct syncD_device *syncd = platform_get_drvdata(pdev);

	struct device_node *clock_m_node;
	struct platform_device *clock_m_pdev, *mac_pdev;
	int i;


	clock_m_node = of_parse_phandle(pdev->dev.of_node, "clock_m", 0);
	clock_m_pdev = of_find_device_by_node(clock_m_node);
	if (!clock_m_pdev) {
		dev_err(dev, "of_find_device_by_node(%pOF) failed\n", clock_m_node);
		return -ENODEV;
	}

	for(i =0; true; i++) {
		struct device_node *mac_node = of_parse_phandle(pdev->dev.of_node, "mac", i);
		if(IS_ERR_OR_NULL(mac_node)) {
			break;
		}
		mac_pdev = of_find_device_by_node(mac_node);
		if (!mac_pdev) {
			dev_err(dev, "of_find_device_by_node(%pOF) failed\n", mac_node);
			return -ENODEV;
		}
		if(i >= MAX_AMOUNT_MACS) {
			dev_err(dev, "defined more than `MAX_AMOUNT_MACS` connected nics\n");
			break;
		}
		syncd->syn1588_dev.mac_devs[i] = mac_pdev;
	}

	syncd->syn1588_dev.clock_m_dev = clock_m_pdev;
	syncd->syn1588_dev.sync_d = pdev;

	return 0;
}
#else
static inline int syn1588_syncd_probe_dt(struct platform_device *pdev)
{
	return 0;
}
#endif

/// @} paramter parsing


static int syn1588_syncd_remove(struct platform_device *pdev)
{
	struct syncD_device *syncd = platform_get_drvdata(pdev);
	struct device *dev = &pdev->dev;

	dev_info(dev, "exiting driver");
	if(IS_ERR_OR_NULL(syncd)) {
		return 0;
	}


	{
		struct syncD_virtdev *virt_entry;

		// unregister virtual devices
		down_write(&syncd->virtdev_list_sem); // lock device list
		list_for_each_entry(virt_entry, &syncd->virtdev_list, list)
			syncD_release_virtdev(syncd, virt_entry->access_key);

		if (atomic_read(&syncd->virtdev_count) != 0) // should not happen
			pr_warn("unregistering device handle 0x%p with active virtual devices.\n",
			get_clock_m(syncd));
		atomic_sub(1,&syncd->virtdev_count); // block new open requests
		up_write(&syncd->virtdev_list_sem); // unlock device list
	}

	syncd_device_free_minor(syncd->miscdev.minor);
	misc_deregister(&syncd->miscdev); // remove misc device from system

	if(syncd->dev_nr >= 0) {
		syncD_device_free_id(syncd->dev_nr);
	}

	platform_set_drvdata(pdev, NULL);

	return 0;
}

static int syn1588_syncd_probe(struct platform_device *pdev)
{
	int				status;
	struct device			*dev = &pdev->dev;
	struct syncD_device *syncd = devm_kzalloc(dev, sizeof(struct syncD_device), GFP_KERNEL);
	platform_set_drvdata(pdev, syncd);
	syncd->dev_nr = -1;

	dev_info(dev, "probe oregano clock driver\n");

	if (dev_of_node(&pdev->dev))
		status = syn1588_syncd_probe_dt(pdev);
	else
		status = syn1588_syncd_probe_pdata(pdev);

	if (status)
		return status;

	syncd->dev = &pdev->dev;

	if(get_clock_m(syncd) != NULL) {
		struct clock_m_device *clock_m = get_clock_m(syncd);
		syncd->clock_m_callback.fn = syncD_Event;
		syncd->clock_m_callback.data = syncd;

		clock_m_register_callback(clock_m, &syncd->clock_m_callback);
	}


	syncd->iobase_spi = devm_platform_ioremap_resource_byname(pdev,  "legacy-spi");
	syncd->iobase_i2c_video = devm_platform_ioremap_resource_byname(pdev,  "legacy-i2c-video");
	syncd->iobase_eeprom = devm_platform_ioremap_resource_byname(pdev,  "legacy-eeprom");

	// initialize main device
	init_rwsem(&syncd->virtdev_list_sem); // initialize virtual device list semaphore
	spin_lock_init(&syncd->virtdev_lock); // initialize virtual device list spin lock
	atomic_set(&syncd->virtdev_count, 0); // driver start-up => no active virt. devices
	spin_lock_init(&syncd->virtreg_lock); // used to lock virtual registers
	INIT_LIST_HEAD(&syncd->virtdev_list); // init virtual device list

	syncD_dev_init(syncd);
	return 0;
}





MODULE_ALIAS("platform:" DRV_NAME);

static struct platform_driver syn1588_syncd_driver = {
	.driver = {
		.name	= DRV_NAME,
#ifdef CONFIG_OF
		.of_match_table = of_match_ptr(syn1588_syncd_dt_ids),
#endif // CONFIG_OF
	},
	.probe		= syn1588_syncd_probe,
	.remove		= syn1588_syncd_remove,
};


int syncD_dev_init_loadmodul(void)
{
	spin_lock_init(&idr_lock);
#ifndef RHEL_BACKPORT_IDR2
	idr_init(&syncd_dev_idr);
#else
	idr2_init(&syncd_dev_idr);
#endif
	return platform_driver_register(&syn1588_syncd_driver);
}
module_init(syncD_dev_init_loadmodul);

void syncD_dev_exit_unloadmodul(void)
{
#ifndef RHEL_BACKPORT_IDR2
	idr_destroy(&syncd_dev_idr);
#else
	idr2_destroy(&syncd_dev_idr);
#endif
	platform_driver_unregister(&syn1588_syncd_driver);
}
module_exit(syncD_dev_exit_unloadmodul);

MODULE_DESCRIPTION("userspace driver for oregano");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRV_VERSION);
