
#include "clock_m_phc.h"
#include "compat/compat.h"
#include "compat/compat_rhel.h"
#include "syn1588nic_local.h"

#include <linux/math64.h>
#include <linux/miscdevice.h>
#include <linux/ptp_clock_kernel.h>


static u64 div_rem(u64 n, u32 base, u32 *remainder){
	*remainder = do_div(n, base);
	return n;
}


/**
 * Hardware wants a 40bit fractal of ns
 */
static u64 calculate_step_register(u32 frequency) {
	u32 remainder;
	u64 step_ns = div_rem(NSEC_PER_SEC, frequency,  &remainder);

	u64 tmp = div_rem(((u64)remainder) << 32, frequency,  &remainder);
	u64 tmp2 = div_rem(((u64)remainder) << 8, frequency,  &remainder);
	u64 step_fractal = (step_ns << 40) + (tmp << 8) + tmp2;

	return step_fractal;
}


static int syn1588nic_phc_adjfine(struct ptp_clock_info *ptp, long scaled_ppm)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	struct device *dev = clock_m_dev->dev;
	u32 timectrl_value;
	u32 hw_register_l = 0x0;
	u32 hw_register_h = 0x0;
	u64 hw_register = 0x0;

	u64 hw_step_orginal = calculate_step_register(clock_m_dev->clk_freq);

	dev_dbg(dev, "syn1588nic_phc_adjfine: %ld\n", scaled_ppm);
	if(clock_m_dev->si_time_pll && false) { // deactivate pll adjust for now
		return si_time_pll_clock_adjfine(clock_m_dev->si_time_pll, scaled_ppm);
	}

	hw_register = hw_step_orginal;
	if(scaled_ppm < 0) {
		hw_register -= mul_u64_u64_div_u64(hw_step_orginal, -scaled_ppm, 1000000ULL << 16);
	} else {
		hw_register += mul_u64_u64_div_u64(hw_step_orginal, +scaled_ppm, 1000000ULL << 16);
	}

	hw_register_l = hw_register;
	hw_register_h = hw_register >> 32;


	clock_m_setreg(clock_m_dev, SHDWSTEPPU_L, hw_register_l);
	clock_m_setreg(clock_m_dev, SHDWSTEPPU_H, hw_register_h);

	clock_m_getreg(clock_m_dev, TIMECTRL, &timectrl_value);
	clock_m_setreg(clock_m_dev, TIMECTRL, timectrl_value | TIMECTL_LOADSTEP);
	return 0;
}

#if KERNEL_VERSION(6, 2, 0) > LINUX_VERSION_CODE
static int syn1588nic_phc_adjfreq(struct ptp_clock_info *ptp, s32 delta)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	struct device *dev = clock_m_dev->dev;
	s32 _tmp;
	long scaled_ppm = div_s64_rem((s64)delta * (1u<<16) , 1000, &_tmp );
	dev_dbg(dev, "syn1588nic_phc_adjfreq %d\n", delta);
	return syn1588nic_phc_adjfine(ptp, scaled_ppm);
}
#endif


static int syn1588nic_phc_adjtime(struct ptp_clock_info *ptp, s64 delta_ns)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	struct device *dev = clock_m_dev->dev;
	u32 timectrl_value;
	const u32 max_shift = NSEC_PER_SEC / 2;
	dev_dbg(dev, "syn1588nic_phc_adjtime: %lld ns\n", delta_ns);

	if (delta_ns > max_shift) {
		pr_err("max_shift");
		return -EINVAL;
	}
	clock_m_getreg(clock_m_dev, TIMECTRL, &timectrl_value);
	if (delta_ns < 0) {
		s32 _tmp;
		uint32_t clock_ticks = div_s64_rem((-delta_ns) * clock_m_dev->clk_freq , NSEC_PER_SEC, &_tmp);
		if (clock_ticks >= (1u<<26)) {
			pr_err("max clock_ticks");
			return -EINVAL;
		}
		clock_m_setreg(clock_m_dev, SHDWTIME_NS, clock_ticks);
		clock_m_setreg(clock_m_dev, TIMECTRL, timectrl_value | TIMECTL_STOPCLOPCK);
	}else {

		clock_m_setreg(clock_m_dev, SHDWTIME_NS, (u32)delta_ns);
		clock_m_setreg(clock_m_dev, TIMECTRL, timectrl_value | TIMECTL_ADDTIME);
	}
	pr_err("syn1588nic_phc_adjtime ok");
	return 0;
}

static int syn1588nic_phc_settime64(struct ptp_clock_info *ptp, const struct timespec64 *ts)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	u32 timectrl_value;

	clock_m_setreg(clock_m_dev, SHDWTIME_NS, ts->tv_nsec);
	clock_m_setreg(clock_m_dev, SHDWTIME_SEC, ts->tv_sec);

	clock_m_getreg(clock_m_dev, TIMECTRL, &timectrl_value);
	clock_m_setreg(clock_m_dev, TIMECTRL, timectrl_value | TIMECTL_LOADTIME);
	return 0;
}

#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
static int syn1588nic_phc_gettimex64(struct ptp_clock_info *ptp, struct timespec64 *ts, struct ptp_system_timestamp *sts)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	u32 tmp = 0;

	spin_lock(&clock_m_dev->time1_lock);

	ptp_read_system_prets(sts);
	clock_m_getreg(clock_m_dev, TIME_NS, &tmp);
	ptp_read_system_postts(sts);
	ts->tv_nsec = tmp;
	clock_m_getreg(clock_m_dev, TIME_SEC, &tmp);
	ts->tv_sec = tmp;

	spin_unlock(&clock_m_dev->time1_lock);

	return 0;
}
#else
static int syn1588nic_phc_gettime64(struct ptp_clock_info *ptp, struct timespec64 *ts)
{
	struct clock_m_device *clock_m_dev = container_of(ptp, struct clock_m_device, ptp_clock_info);
	u32 tmp = 0;

	spin_lock(&clock_m_dev->time1_lock);

	clock_m_getreg(clock_m_dev, TIME_NS, &tmp);
	ts->tv_nsec = tmp;
	clock_m_getreg(clock_m_dev, TIME_SEC, &tmp);
	ts->tv_sec = tmp;

	spin_unlock(&clock_m_dev->time1_lock);

	return 0;
}
#endif  // KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE

void clock_m_gettime(struct clock_m_device *device, struct timespec64 *ts) {

#if KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
	syn1588nic_phc_gettimex64(&device->ptp_clock_info, ts, NULL);
#else
	syn1588nic_phc_gettime64(&device->ptp_clock_info, ts);
#endif  // KERNEL_VERSION(5, 0, 0) <= LINUX_VERSION_CODE
}

static int syn1588nic_phc_enable(struct ptp_clock_info *ptp, struct ptp_clock_request *request, int on)
{
	return -EOPNOTSUPP;
}

static const struct ptp_clock_info syn1588_ptp_clock_info = {
	.owner		= THIS_MODULE,
	.name		= "syn1588nic",
	.n_alarm	= 0,
	.n_ext_ts	= 0,
	.n_per_out	= 0,
	.n_pins		= 0,
	.pps		= 0,
	.max_adj        = 1000000,
#if KERNEL_VERSION(6, 2, 0) > LINUX_VERSION_CODE
	.adjfreq	= syn1588nic_phc_adjfreq,
#endif
	.adjfine	= syn1588nic_phc_adjfine,
	.adjtime	= syn1588nic_phc_adjtime,
	.enable		= syn1588nic_phc_enable,

#if KERNEL_VERSION(4, 20, 0) <= LINUX_VERSION_CODE
	.gettimex64	= syn1588nic_phc_gettimex64,
#elif KERNEL_VERSION(4, 1, 0) <= LINUX_VERSION_CODE || defined(RHEL7_BACKPORT_PTP_SETTIME64)
	.gettime64      = syn1588nic_phc_gettime64,
#else
#endif  //  KERNEL_VERSION(4, 20, 0) <= LINUX_VERSION_CODE

#if KERNEL_VERSION(4, 1, 0) <= LINUX_VERSION_CODE || defined(RHEL7_BACKPORT_PTP_SETTIME64)
	.settime64	= syn1588nic_phc_settime64,
#else
	#error not supportet kernel settime64 is required
#endif
};


int syn1588nic_phc_init(struct clock_m_device *clock_m_dev)
{
	clock_m_dev->ptp_clock_info = syn1588_ptp_clock_info;
		clock_m_dev->ptp_clock = ptp_clock_register(&clock_m_dev->ptp_clock_info, clock_m_dev->dev);
		if (IS_ERR(clock_m_dev->ptp_clock)) {
			clock_m_dev->ptp_clock = NULL;
			pr_err("ptp_clock_unregister failed");
			return -ENODEV;
		}
	return 0;
}

void syn1588nic_phc_exit(struct clock_m_device *clock_m_dev)
{
	if (!IS_ERR_OR_NULL(clock_m_dev->ptp_clock)) {
		ptp_clock_unregister(clock_m_dev->ptp_clock);
		clock_m_dev->ptp_clock = NULL;
	}
}
