
// SPDX-License-Identifier: GPL-2.0-or-later
#include "oregano_spi.h"

#include <linux/debugfs.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/jiffies.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/spi/spi.h>
#include <linux/version.h>



#ifndef DRIVER_NAME
#define DRIVER_NAME	"oregano-spi"
#endif


struct spi_oregano {
	void __iomem *iobase; // memory base
	struct spi_master *master; // back reference

	unsigned int clock_hz;

	struct dentry *debugfs;
};

enum spi_oregano_regs {
	OREGANO_SPI_RXDATA = 0x0,
	OREGANO_SPI_TXDATA = 0x4,
	OREGANO_SPI_STATUS = 0x8,
	OREGANO_SPI_CONTROL = 0xc, // currently not implemented in HW
	OREGANO_SPI_SLAVE_SEL = 0x14,
};

enum spi_oregano_reg_status {
	OREGANO_SPI_STATUS_ROE = BIT(3),
	OREGANO_SPI_STATUS_TOE = BIT(4),
	OREGANO_SPI_STATUS_TMT = BIT(5),
	OREGANO_SPI_STATUS_TRDY = BIT(6),
	OREGANO_SPI_STATUS_RRDY = BIT(7),
	OREGANO_SPI_STATUS_E = BIT(8),
};

enum spi_oregano_reg_control {
	OREGANO_SPI_CONTROL_SSO = BIT(10),
};

/*----------------------------------------------------------------------*/

static inline u32 spi_oregano_readl(struct spi_oregano *spi_dev, u32 offset)
{
	struct device *dev = &spi_dev->master->dev;
	u32 val = ioread32(spi_dev->iobase + offset);
	dev_dbg(dev, "spi read addr: 0x%02x data: 0x%08x\n", offset, val);
	return val;
}

static inline void spi_oregano_writel(struct spi_oregano *spi_dev, u32 offset, u32 val)
{
	struct device *dev = &spi_dev->master->dev;
	dev_dbg(dev, "spi write addr: 0x%02x data: 0x%08x\n", offset, val);
	iowrite32(val, spi_dev->iobase + offset);
}

static int spi_oregano_wait(struct spi_oregano *spi_oregano,
			   int reg, u32 mask, u32 val,
			   const unsigned long timeout)
{
	unsigned long j;

	j = jiffies + timeout;
	while (1) {
		u32 status = spi_oregano_readl(spi_oregano, reg);

		if ((status & mask) == val)
			break;

		if (time_after(jiffies, j))
			return -ETIMEDOUT;
	}
	return 0;
}

static void spi_oregano_chipselect(struct spi_device *spi, bool is_high)
{
	struct device *dev = &spi->master->dev;
	struct spi_oregano *spi_oregano =  spi_master_get_devdata(spi->master);
	u32 slave_select = 0x0;

	dev_dbg(dev, "set chipselect to %s\n", is_high? "inactiv":"active");

	/* set initial clock line level */
	if (!is_high) {
		slave_select = BIT(spi->chip_select);
	}

	//spi_oregano_writel(spi_oregano, OREGANO_SPI_SLAVE_SEL, slave_select );

	if(is_high) {
		spi_oregano_writel(spi_oregano, OREGANO_SPI_CONTROL, 0x0 );
		spi_oregano_writel(spi_oregano, OREGANO_SPI_SLAVE_SEL, slave_select );
	}else {
		spi_oregano_writel(spi_oregano, OREGANO_SPI_SLAVE_SEL, slave_select );
		spi_oregano_writel(spi_oregano, OREGANO_SPI_CONTROL, OREGANO_SPI_CONTROL_SSO );
	}

}

static int spi_oregano_setup(struct spi_device *spi)
{
	struct device *dev = &spi->master->dev;
	dev_dbg(dev, "setup spi controller\n");

	return 0;
}

#ifdef CONFIG_OF
static const struct of_device_id spi_oregano_dt_ids[] = {
	{ .compatible = "oregano,spi" },
	{}
};
MODULE_DEVICE_TABLE(of, spi_oregano_dt_ids);

static int spi_oregano_probe_dt(struct platform_device *pdev,
				 struct spi_master *master)
{
	master->dev.of_node = pdev->dev.of_node;
	master->use_gpio_descriptors = true;

	return 0;
}
#else
static inline int spi_oregano_probe_dt(struct platform_device *pdev,
					struct spi_master *master)
{
	return 0;
}
#endif

static int oregano_spi_probe_pdata(struct platform_device *pdev,
				struct spi_master *master)
{
	struct device *dev = &pdev->dev;
	struct oregano_spi_platform_data *pdata = dev_get_platdata(dev);
	struct spi_oregano *spi_oregano = spi_master_get_devdata(master);

	if(master->num_chipselect > 32) {
		return -EINVAL;
	}
	/*
	 * The master needs to think there is a chipselect even if not
	 * connected
	 */
	if(pdata->num_chipselect == 0)
		pdata->num_chipselect = 1;

	master->num_chipselect = pdata->num_chipselect ;
	master->mode_bits = pdata->mode_bits;
	spi_oregano->clock_hz = pdata->clock_hz;

	return 0;
}

static void spi_oregano_put(void *data)
{
	struct spi_master *master = data;
	spi_master_put(master);
}

static int spi_oregano_transfer_one(struct spi_controller *master,
		struct spi_device *spi, struct spi_transfer *transfer)
{
	struct spi_oregano *spi_oregano = dev_get_drvdata(&master->dev);
	struct device *dev = &master->dev;
	int j;
	int err = 0;

	dev_dbg(dev, "spi transfer one, len: %u\n", transfer->len);

#if (KERNEL_VERSION(5, 3, 0) <= LINUX_VERSION_CODE)
	transfer->effective_speed_hz = spi_oregano->clock_hz; // static value as can not be adjusted
#endif

	for(j=0; j<transfer->len;j++) {
		u32 txdata = 0x0;
		u32 rxdata = 0x0;
		if(transfer->tx_buf) {
			txdata = ((u8 *)transfer->tx_buf)[j];
		}
		spi_oregano_writel(spi_oregano, OREGANO_SPI_TXDATA, txdata);
		err = spi_oregano_wait(spi_oregano, OREGANO_SPI_STATUS, OREGANO_SPI_STATUS_RRDY, OREGANO_SPI_STATUS_RRDY, msecs_to_jiffies(1));
		if(err){
			dev_err(dev, "timeout while waiting for spi to do transfer\n");
			return -EIO;
		}
		rxdata = spi_oregano_readl(spi_oregano, OREGANO_SPI_RXDATA);
		if(transfer->rx_buf) {
			((u8 *)transfer->rx_buf)[j] = rxdata;
		}
	}

	dev_dbg(dev, "transfer done\n");

	return 0;
}

static int spi_oregano_probe(struct platform_device *pdev)
{
	int				status;
	struct spi_master		*master;
	struct spi_oregano		*spi_oregano;
	struct device			*dev = &pdev->dev;
	struct resource *res;
	u32 status_reg;

	dev_info(dev, "probe oregano spi driver\n");

	master = spi_alloc_master(dev, sizeof(*spi_oregano));
	if (!master)
		return -ENOMEM;

	status = devm_add_action_or_reset(&pdev->dev, spi_oregano_put, master);
	if (status) {
		spi_master_put(master);
		return status;
	}

	if (pdev->dev.of_node)
		status = spi_oregano_probe_dt(pdev, master);
	else
		status = oregano_spi_probe_pdata(pdev, master);

	if (status)
		return status;


	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(res == NULL) {
		dev_err(dev, "could not request resources\n");
		return -ENODEV;
	}

	spi_oregano = spi_master_get_devdata(master);
	spi_oregano->master = master;
	spi_oregano->iobase = devm_ioremap_resource(dev, pdev->resource);
	if(IS_ERR_OR_NULL(spi_oregano->iobase)) {
		dev_err(dev, "could not ioremap resource\n");
		return PTR_ERR(spi_oregano->iobase);
	}

	master->bits_per_word_mask = SPI_BPW_MASK(8);
	master->mode_bits = 0; //SPI_CPHA | SPI_CPOL | SPI_CS_HIGH;

	master->bus_num = -1;
	master->setup = spi_oregano_setup;
	//master->cleanup = spi_oregano_cleanup;

	master->set_cs = spi_oregano_chipselect;

	master->transfer_one = spi_oregano_transfer_one;
	//master->handle_err = spi_oregano_handle_err;

	master->max_speed_hz = spi_oregano->clock_hz;
	master->min_speed_hz = spi_oregano->clock_hz;
	master->dev.fwnode = dev->fwnode;

	master->flags = SPI_MASTER_GPIO_SS;
	master->num_chipselect = 1;
	//of_property_read_s32

	// always operate in SSO mode
	spi_oregano_writel(spi_oregano, OREGANO_SPI_CONTROL, OREGANO_SPI_CONTROL_SSO );
	//spi_oregano_writel(spi_oregano, OREGANO_SPI_CONTROL, 0x00);
	//reset status
	spi_oregano_writel(spi_oregano, OREGANO_SPI_STATUS, 0x0 );

	status_reg = spi_oregano_readl(spi_oregano, OREGANO_SPI_STATUS);
	if (status_reg & OREGANO_SPI_STATUS_RRDY)
		(void)spi_oregano_readl(spi_oregano, OREGANO_SPI_RXDATA); /* flush rxdata */

	status = devm_spi_register_master(&pdev->dev, spi_master_get(master));
	if(status) {
		dev_err(dev, "could not register spi master device\n");
		return status;
	}

	if (!pdev->dev.of_node) {
		struct device *dev = &pdev->dev;
		struct oregano_spi_platform_data *pdata = dev_get_platdata(dev);
		int i;
		if (pdata) {
			for (i = 0; i < pdata->num_devices; i++) {
				if (!spi_new_device(master, pdata->devices + i))
					dev_warn(&pdev->dev,
						"unable to create SPI device: %s\n",
						pdata->devices[i].modalias);
			}
		}
	}
	dev_info(dev, "%s(): SPI init OK\n", __func__);
	return 0;
}

MODULE_ALIAS("platform:" DRIVER_NAME);

static struct platform_driver spi_oregano_driver = {
	.driver = {
		.name	= DRIVER_NAME,
		.of_match_table = of_match_ptr(spi_oregano_dt_ids),
	},
	.probe		= spi_oregano_probe,
};
module_platform_driver(spi_oregano_driver);

MODULE_DESCRIPTION("SPI master driver for oregano");
MODULE_AUTHOR("Norbert Summer");
MODULE_LICENSE("GPL");
