



#ifndef _SYN1588NIC_LOCAL_H_
#define _SYN1588NIC_LOCAL_H_

#include "clock_m.h"
#include "syn1588nic_reg.h"

#include <linux/types.h>
#include <linux/spinlock_types.h>
#include <linux/version.h>



/// LOG Format
#ifdef pr_fmt
#undef pr_fmt
#endif
#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

/**
 * \name General driver settings
 *@{
 */
#define CLOCK_M_BASE 0x00001000 ///< PCI NIC syn1588 CLOCK_M base address
#define CLOCK_M_SIZE 0x00001000 ///< PCI NIC syn1588 CLOCK_M base address
#define CRA_BASE 0x8000 ///< CRA base address
#define CRA_ADDR_TRANS0_L (CRA_BASE + 0x1000)
#define CRA_ADDR_TRANS0_H (CRA_ADDR_TRANS0_L + 4)

// Burst length for reading/writing DMA buffers
#define BURST_LEN_PCI 16 // PCI NICs
#define BURST_LEN_ARM 32 /// ARM AXI BUS
//#define DMA_BD_INIT	 0xA				///< value for initializing buffer DMA
// descriptor (comment to disable)
//#define DBG_PKTDATA							///< print beginning of received packet data
////FIXME! remove if solved!
#define SELF_TEST_DUR 5 ///< self test duration in seconds
#define DEBUG_LVL KERN_INFO ///< driver debug level (verbosity)
#define ACCEPT_VLAN0 ///< accept VLAN0 as a valid VLAN identifier
#define NUMBER_OF_NIC 2 ///< Number of NIC interface
/*@}*/

/**
 * \name PTP related settings
 *@{
 */
// PHY delays in ns. For NTP mode, the values have to converted to fractional
// seconds
#define RX_PHY_DELAY_1GBIT (213UL << 16)
#define RX_PHY_DELAY_100MBIT 315UL

#define TX_PHY_DELAY_1GBIT (132UL << 16)
#define TX_PHY_DELAY_100MBIT 190UL

#define TX_PHY_DELAY (TX_PHY_DELAY_100MBIT | TX_PHY_DELAY_1GBIT)
#define RX_PHY_DELAY (RX_PHY_DELAY_100MBIT | RX_PHY_DELAY_1GBIT)

#define NTP_PTP_OFFSET 0x83AA7E80
/*@}*/

// defines the minimum build number necessary for the PCIe NIC to use the VLAN
// group register
#ifndef MIN_BUILD_VLAN_GRP_REG
#define MIN_BUILD_VLAN_GRP_REG 759
#endif



/// list of validated PHY chips (PHY_OUI (low), PHY_OUI (high), PHY_ID (address
/// of the PHY))
static const u16 phy_detect_dat[] = {
	0x0022, 0x1622,
	0x00, // syn1588(R) PCIe NIC Rev 2.x (Micrel KSZ9031)
	0x0141, 0x0CC2,
	0x00, // syn1588(R) PCIe NIC Rev 1.x (Marvell 88E1111)
	0x0141, 0x0CC2,
	0x12, // Altera Stratix-II GX eval. card (Marvell 88E1111)
	0x0007, 0xC0C3,
	0x10, // syn1588(R) PCI NIC	(SMSC 8700)
	0x0000, 0x0000,
	0x00 // SFP FIBRE with emulated PHY
};

//---------------------------------------------------------------------------
// Emulated PHY registers contents
//---------------------------------------------------------------------------
#define c_phy_register0 0x1140
#define c_phy_register1_link 0x796D
#define c_phy_register1_nolink 0x7949
// Marvell OUI 0x005043
#define c_phy_register2 0x0141
// Marvell Model 88E1111, revision 1
#define c_phy_register3 0x0CC2
//autoneg advertise
#define c_phy_register4 0x0DE1
#define c_phy_register5_link 0xC1E1
#define c_phy_register5_nolink 0x0000
#define c_phy_register6_link 0x000D
#define c_phy_register6_nolink 0x0004
#define c_phy_register7 0x2801
#define c_phy_register8_link 0x4E62
#define c_phy_register8_nolink 0x0000
//1000Base-T control
#define c_phy_register9 0x0F00
//1000Base-T status
#define c_phy_register10_nolink 0x4000
#define c_phy_register10_link 0x3800
#define c_phy_register15 0xF000
#define c_phy_register17_link 0xAC00
#define c_phy_register17_nolink 0x8100

/**************************************************************************************************
 * general definitions for driver (don't change!)
 **************************************************************************************************/

/// length of Ethernet frame check sequence in bytes
#define ETH_FCSLEN 4

/// time till network layer considers transmission to be failed
// wait 500 ms max
#define TX_TIMEOUT (HZ >> 2)
/// time between two consecutive link checks by the watchdog
#define LINK_UPDATE ((5 * HZ) >> 2)
/// desired waiting time for one MDIO access in us
#define MII_TOUT 10*39
/// self test result size in u64
#define SELF_TEST_RES 1
#define MAX_MII_READ_RETRIES 10
#define MAX_MII_WRITE_RETRIES 10


/// define register access modes
enum access_mode {
	ACC_RD = 1, ///< register is readable
	ACC_WR, ///< register is writable
	ACC_RDWR, ///< register is read and writeable (equals ACC_RD | ACC_WR)
	ACC_NOVERIFY =
		4 ///< do not run a write/verify cycle (register is volatile)
};

/// structure to bild list of PCI adapters the driver currently handles
struct adapter_node {
	struct syn1588_adapter
		*adapter; ///< data structure containing information about
	/// the specific adapter
	struct list_head list; ///< structure to keep linked list
};

/// global driver data
struct syn1588_drvdata {
	/// of CLOCK_M devices
	spinlock_t
		adapter_list_lock; ///< spinlock for locking access to adapter_list
	struct list_head
		adapter_list; ///< list of PCI adapters the driver currently handles
};

/*
struct syn1588_iomatrix {
	u16 SMA_src;
	u8 SMA_inv;
	u8 EVENT_src;
}; */

// define register set for register readout
static const u32 mac_regset[] = MAC_REGSET;

/// structure to generate a list of VLANs (if the VLAN group register is in use)
struct vlan_grp_list {
	struct list_head
		list; ///< to link each element of type struct vlan_grp_list to others
	unsigned short vid; ///< save the VLAN ID in a list
};



struct syn1588_debufs;

struct syn1588_dma {
	dma_addr_t dma_addr; // bus memory address
	void *dma_vaddr; // virtual memory address
	struct dma_pool *dma_pool;
	u32 passthru_mask;
};
/// Struct keeping all neccessary information about one adapter for the driver
struct syn1588_adapter {
	void __iomem *iobase; ///< pointer to mapped PCI I/O memory
	struct resource *physical_memory; ///< bar0 memory resource
	struct clock_m_device *clock_m; ///< CLOCK_M data structure
	struct net_device *nic; ///< NIC data structure
	struct device *dev; ///< PCI data structure
	struct syn1588_dma dma;

	struct syn1588_debugfs *debugfs; // filestructure for debug reasons
};

void syn1588_BHH_TX(unsigned long);
void syn1588_BHH_RX(unsigned long);

/*@}*/




#endif // _SYN1588NIC_LOCAL_H_
